Package: bluster
Version: 1.22.0
Date: 2026-03-05
Title: Clustering Algorithms for Bioconductor
Description: Wraps common clustering algorithms in an easily extended
        S4 framework. Backends are implemented for hierarchical,
        k-means and graph-based clustering. Several utilities are also
        provided to compare and evaluate clustering results.
Authors@R: 
    c(
        person("Aaron", "Lun", role = c("aut", "cre"), email = "infinite.monkeys.with.keyboards@gmail.com"),
        person("Stephanie", "Hicks", role="ctb"),
        person("Basil", "Courbayre", role="ctb"),
        person("Tuomas", "Borman", role="ctb"),
        person("Leo", "Lahti", role="ctb")
    )
Imports: stats, methods, utils, cluster, Matrix, Rcpp, igraph,
        S4Vectors, BiocParallel, BiocNeighbors
Suggests: knitr, rmarkdown, testthat, BiocStyle, dynamicTreeCut,
        scRNAseq, scuttle, scater, scran, pheatmap, viridis, mbkmeans,
        kohonen, apcluster, DirichletMultinomial, vegan, fastcluster
biocViews: ImmunoOncology, Software, GeneExpression, Transcriptomics,
        SingleCell, Clustering
LinkingTo: Rcpp, assorthead
Collate: AllClasses.R AllGenerics.R AgnesParam.R approxSilhouette.R
        bluster-package.R DbscanParam.R DianaParam.R AffinityParam.R
        BlusterParam.R bootstrapStability.R ClaraParam.R clusterRMSD.R
        clusterSweep.R compareClusterings.R DmmParam.R
        FixedNumberParam.R HclustParam.R HierarchicalParam.R
        KmeansParam.R linkClusters.R makeSNNGraph.R MbkmeansParam.R
        mergeCommunities.R neighborPurity.R nestedClusters.R
        NNGraphParam.R pairwiseModularity.R pairwiseRand.R PamParam.R
        RcppExports.R SomParam.R TwoStepParam.R utils.R
License: GPL-3
NeedsCompilation: yes
VignetteBuilder: knitr
SystemRequirements: C++17
RoxygenNote: 7.3.3
Encoding: UTF-8
Config/pak/sysreqs: libglpk-dev libxml2-dev
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 12:53:10 UTC
RemoteUrl: https://github.com/bioc/bluster
RemoteRef: RELEASE_3_23
RemoteSha: c564e528da63e928b8a22e01bf80c148ce8042a3
Packaged: 2026-04-29 15:00:03 UTC; root
Author: Aaron Lun [aut, cre],
  Stephanie Hicks [ctb],
  Basil Courbayre [ctb],
  Tuomas Borman [ctb],
  Leo Lahti [ctb]
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-29 15:06:31 UTC; unix

# Generated by roxygen2: do not edit by hand

export("centers<-")
export(.defaultScalarArguments)
export(.extractScalarArguments)
export(.showScalarArguments)
export(AffinityParam)
export(AgnesParam)
export(ClaraParam)
export(DbscanParam)
export(DianaParam)
export(DmmParam)
export(HclustParam)
export(KNNGraphParam)
export(KmeansParam)
export(MbkmeansParam)
export(NNGraphParam)
export(PamParam)
export(SNNGraphParam)
export(SomParam)
export(TwoStepParam)
export(approxSilhouette)
export(bootstrapStability)
export(centers)
export(clusterRMSD)
export(clusterRows)
export(clusterSweep)
export(compareClusterings)
export(linkClusters)
export(linkClustersMatrix)
export(makeKNNGraph)
export(makeSNNGraph)
export(mergeCommunities)
export(neighborPurity)
export(neighborsToKNNGraph)
export(neighborsToSNNGraph)
export(nestedClusters)
export(pairwiseModularity)
export(pairwiseRand)
exportClasses(AffinityParam)
exportClasses(AgnesParam)
exportClasses(BlusterParam)
exportClasses(ClaraParam)
exportClasses(DbscanParam)
exportClasses(DianaParam)
exportClasses(DmmParam)
exportClasses(FixedNumberParam)
exportClasses(HclustParam)
exportClasses(HierarchicalParam)
exportClasses(KNNGraphParam)
exportClasses(KmeansParam)
exportClasses(MbkmeansParam)
exportClasses(NNGraphParam)
exportClasses(PamParam)
exportClasses(SNNGraphParam)
exportClasses(SomParam)
exportClasses(TwoStepParam)
exportMethods("[[")
exportMethods("[[<-")
exportMethods("centers<-")
exportMethods(.defaultScalarArguments)
exportMethods(centers)
exportMethods(clusterRows)
exportMethods(show)
exportMethods(updateObject)
import(methods)
importFrom(BiocNeighbors,KmknnParam)
importFrom(BiocNeighbors,buildIndex)
importFrom(BiocNeighbors,findDistance)
importFrom(BiocNeighbors,findKNN)
importFrom(BiocNeighbors,findNeighbors)
importFrom(BiocNeighbors,queryKNN)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bpisup)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpstart)
importFrom(BiocParallel,bpstop)
importFrom(Matrix,"diag<-")
importFrom(Matrix,diag)
importFrom(Rcpp,sourceCpp)
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,I)
importFrom(S4Vectors,List)
importFrom(S4Vectors,coolcat)
importFrom(S4Vectors,setValidity2)
importFrom(S4Vectors,updateObject)
importFrom(cluster,agnes)
importFrom(cluster,clara)
importFrom(cluster,diana)
importFrom(cluster,pam)
importFrom(igraph,"E<-")
importFrom(igraph,E)
importFrom(igraph,cluster_walktrap)
importFrom(igraph,components)
importFrom(igraph,graph_from_adjacency_matrix)
importFrom(igraph,is.directed)
importFrom(igraph,make_graph)
importFrom(igraph,membership)
importFrom(igraph,modularity)
importFrom(igraph,simplify)
importFrom(stats,as.hclust)
importFrom(stats,cutree)
importFrom(stats,dist)
importFrom(stats,hclust)
importFrom(stats,kmeans)
importFrom(stats,median)
importFrom(stats,quantile)
importFrom(stats,runif)
importFrom(stats,var)
importFrom(utils,capture.output)
importFrom(utils,getFromNamespace)
importFrom(utils,head)
importFrom(utils,tail)
useDynLib(bluster)

# Clustering methods for Bioconductor

|Build|Status|
|-----|----|
| Bioc-release | [![](http://bioconductor.org/shields/build/release/bioc/bluster.svg)](http://bioconductor.org/checkResults/release/bioc-LATEST/bluster) |
| Bioc-devel   | [![](http://bioconductor.org/shields/build/devel/bioc/bluster.svg)](http://bioconductor.org/checkResults/devel/bioc-LATEST/bluster) | 

This package provides a consistent interface to a variety of clustering algorithms for convenient use in the rest of the Bioconductor ecosystem.
Check out the user guide on the [Bioconductor landing page](https://bioconductor.org/packages/devel/bioc/html/bluster.html) for more details.

## ----echo=FALSE---------------------------------------------------------------
knitr::opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE)
library(BiocStyle)

## -----------------------------------------------------------------------------
library(scRNAseq)
sce <- GrunPancreasData()

# Quality control to remove bad cells.
library(scuttle)
qcstats <- perCellQCMetrics(sce)
qcfilter <- quickPerCellQC(qcstats, sub.fields="altexps_ERCC_percent")
sce <- sce[,!qcfilter$discard]

# Normalization by library size.
sce <- logNormCounts(sce)

# Feature selection.
library(scran)
dec <- modelGeneVar(sce)
hvgs <- getTopHVGs(dec, n=1000)

# Dimensionality reduction.
set.seed(1000)
library(scater)
sce <- runPCA(sce, ncomponents=20, subset_row=hvgs)

# Clustering.
library(bluster)
mat <- reducedDim(sce)
clust.info <- clusterRows(mat, NNGraphParam(), full=TRUE)
clusters <- clust.info$clusters
table(clusters)

## -----------------------------------------------------------------------------
sil <- approxSilhouette(mat, clusters)
sil
boxplot(split(sil$width, clusters))

## -----------------------------------------------------------------------------
best.choice <- ifelse(sil$width > 0, clusters, sil$other)
table(Assigned=clusters, Closest=best.choice)

## -----------------------------------------------------------------------------
pure <- neighborPurity(mat, clusters)
pure
boxplot(split(pure$purity, clusters))

## -----------------------------------------------------------------------------
table(Assigned=clusters, Max=pure$maximum)

## -----------------------------------------------------------------------------
rmsd <- clusterRMSD(mat, clusters)
barplot(rmsd)

## -----------------------------------------------------------------------------
clusterRMSD(mat, clusters, sum=TRUE)

## -----------------------------------------------------------------------------
g <- clust.info$objects$graph
ratio <- pairwiseModularity(g, clusters, as.ratio=TRUE)

library(pheatmap)
pheatmap(log10(ratio+1), cluster_cols=FALSE, cluster_rows=FALSE,
    col=rev(heat.colors(100)))

## -----------------------------------------------------------------------------
cluster.gr <- igraph::graph_from_adjacency_matrix(log2(ratio+1), 
    mode="upper", weighted=TRUE, diag=FALSE)

# Increasing the weight to increase the visibility of the lines.
set.seed(1100101)
plot(cluster.gr, edge.width=igraph::E(cluster.gr)$weight*5,
    layout=igraph::layout_with_lgl)

## -----------------------------------------------------------------------------
merged <- mergeCommunities(g, clusters, number=10)
table(merged)

## -----------------------------------------------------------------------------
hclusters <- clusterRows(mat, HclustParam(cut.dynamic=TRUE))
pairwiseRand(clusters, hclusters, mode="index")

## -----------------------------------------------------------------------------
ratio <- pairwiseRand(clusters, hclusters, mode="ratio")

library(pheatmap)
pheatmap(ratio, cluster_cols=FALSE, cluster_rows=FALSE,
    col=viridis::viridis(100), breaks=seq(-1, 1, length=101))

## -----------------------------------------------------------------------------
set.seed(1001010)
ari <-bootstrapStability(mat, clusters=clusters, 
    mode="index", BLUSPARAM=NNGraphParam())
ari

## -----------------------------------------------------------------------------
set.seed(1001010)
ratio <-bootstrapStability(mat, clusters=clusters,
    mode="ratio", BLUSPARAM=NNGraphParam())

library(pheatmap)
pheatmap(ratio, cluster_cols=FALSE, cluster_rows=FALSE,
    col=viridis::viridis(100), breaks=seq(-1, 1, length=101))

## -----------------------------------------------------------------------------
combinations <- clusterSweep(mat, BLUSPARAM=SNNGraphParam(),
    k=c(5L, 10L, 15L, 20L), cluster.fun=c("walktrap", "louvain", "infomap"))

## -----------------------------------------------------------------------------
colnames(combinations$clusters)
combinations$parameters

## -----------------------------------------------------------------------------
set.seed(10)
nclusters <- 3:25
kcombos <- clusterSweep(mat, BLUSPARAM=KmeansParam(centers=5), centers=nclusters)

sil <- vapply(as.list(kcombos$clusters), function(x) mean(approxSilhouette(mat, x)$width), 0)
plot(nclusters, sil, xlab="Number of clusters", ylab="Average silhouette width")

pur <- vapply(as.list(kcombos$clusters), function(x) mean(neighborPurity(mat, x)$purity), 0)
plot(nclusters, pur, xlab="Number of clusters", ylab="Average purity")

wcss <- vapply(as.list(kcombos$clusters), function(x) sum(clusterRMSD(mat, x, sum=TRUE)), 0)
plot(nclusters, wcss, xlab="Number of clusters", ylab="Within-cluster sum of squares")

## -----------------------------------------------------------------------------
linked <- linkClusters(
    list(
        walktrap=combinations$clusters$k.10_cluster.fun.walktrap,
        louvain=combinations$clusters$k.10_cluster.fun.louvain,
        infomap=combinations$clusters$k.10_cluster.fun.infomap
    )
)
linked

## -----------------------------------------------------------------------------
meta <- igraph::cluster_walktrap(linked)
plot(linked, mark.groups=meta)

## -----------------------------------------------------------------------------
aris <- compareClusterings(combinations$clusters)
g <- igraph::graph.adjacency(aris, mode="undirected", weighted=TRUE)
meta2 <- igraph::cluster_walktrap(g)
plot(g, mark.groups=meta2)

## -----------------------------------------------------------------------------
sessionInfo()

#' Merge communities from graph-based clustering
#'
#' Adjust the resolution of a graph-based community detection algorithm by greedily merging clusters together.
#' At each step, the pair of clusters that yield the highest modularity are merged.
#'
#' @inheritParams pairwiseModularity
#' @param number Integer scalar specifying the number of clusters to obtain.
#' Ignored if \code{steps} is specified.
#' @param steps Integer scalar specifying the number of merge steps.
#'
#' @return A vector or factor of the same length as \code{clusters},
#' containing the desired number of merged clusters.
#'
#' @details
#' This function is similar to the \code{\link{cut_at}} function from the \pkg{igraph} package,
#' but works on clusters that were not generated by a hierarchical algorithm.
#' The aim is to facilitate rapid adjustment of the number of clusters without having to repeat the clustering -
#' or, even worse, repeating the graph construction, e.g., in \code{\link{makeSNNGraph}}.
#'
#' @author Aaron Lun
#'
#' @examples
#' output <- clusterRows(iris[,1:4], NNGraphParam(k=5), full=TRUE)
#' table(output$clusters)
#'
#' merged <- mergeCommunities(output$objects$graph, output$clusters, number=3)
#' table(merged)
#' 
#' @seealso
#' \code{\link{cut_at}}, for a faster and more natural adjustment when using a hierarchical community detection algorithm.
#'
#' \linkS4class{NNGraphParam}, for a one-liner to generate graph-based clusters.
#' @export 
#' @importFrom igraph modularity E
mergeCommunities <- function(graph, clusters, number=NULL, steps=NULL) {
    nclusters <- length(unique(clusters))

    if (is.null(steps)) {
        if (is.null(number)) {
            stop("either 'number' or 'steps' must be specified")
        }
        steps <- max(0, nclusters - number)
    }

    if (steps >= nclusters) {
        clusters <- rep(clusters[1], length(clusters))

    } else if (steps > 0) {
        for (i in seq_len(steps)) {
            labels <- unique(clusters)

            max.m <- 0
            max.label <- clusters

            # Picking the pair to merge that yields the greatest modularity.
            for (lab in labels) {
                to.merge <- clusters==lab
        
                for (other in labels) {
                    if (other==lab) { 
                        next 
                    }

                    next.label <- clusters
                    next.label[to.merge] <- other
                    next.m <- modularity(graph, next.label, weights=E(graph)$weight)
                    if (max.m < next.m) {
                        max.m <- next.m 
                        max.label <- next.label
                    }
                }
            }

            clusters <- max.label
        }
    }

    if (is.factor(clusters)) {
        droplevels(clusters)
    } else {
        clusters
    }
}

# This tests the mergeCommunities function.
# library(testthat); library(bluster); source("test-merge-comm.R")

set.seed(100010)
test_that("mergeCommunities works as expected", {
    m <- matrix(runif(1000), ncol=10)
    output <- clusterRows(m, NNGraphParam(k=5), full=TRUE)
    merged <- mergeCommunities(output$objects$graph, output$clusters, number=3)

    expect_identical(nlevels(merged), 3L)
    expect_identical(length(unique(merged)), 3L)
    expect_true(length(unique(merged)) < length(unique(output$clusters)))

    # Check for nesting.
    freq <- table(paste(merged, output$clusters))
    expect_identical(length(freq), length(unique(output$clusters)))

    merged2 <- mergeCommunities(output$objects$graph, output$clusters, number=2)
    expect_identical(length(unique(merged2)), 2L)
})

set.seed(100011)
test_that("mergeCommunities does, in fact, minimize the modularity", {
    library(igraph)
    groups <- gl(4, 25)
    m <- matrix(runif(1000), ncol=10) + as.numeric(as.integer(groups) %% 2==0 )

    g <- makeSNNGraph(m, k=10)
    ref <- modularity(g, groups, weights=E(g)$weight)

    merged <- mergeCommunities(g, groups, number=2)
    after <- modularity(g, merged, weights=E(g)$weight)
    expect_true(ref < after)

    silly <- modularity(g, gl(2, 50), weights=E(g)$weight)
    expect_true(silly < after)
})

set.seed(100012)
test_that("mergeCommunities behaves with silly inputs", {
    m <- matrix(runif(1000), ncol=10)
    output <- clusterRows(m, NNGraphParam(k=5), full=TRUE)
    expect_error(mergeCommunities(output$objects$graph, output$clusters), "either")
    
    expect_identical(mergeCommunities(output$objects$graph, output$clusters, step=-1), output$clusters)
    expect_identical(mergeCommunities(output$objects$graph, output$clusters, number=100), output$clusters)

    together <- mergeCommunities(output$objects$graph, output$clusters, step=100)
    expect_identical(nlevels(together), 1L)

    expect_identical(mergeCommunities(output$objects$graph, as.integer(output$clusters), step=0), as.integer(output$clusters))
})

#' Compute pairwise Rand indices
#'
#' Breaks down the Rand index calculation to report values for each cluster and pair of clusters
#' in a reference clustering compared to an alternative clustering.
#'
#' @param ref A character vector or factor containing one set of groupings, considered to be the reference.
#' @param alt A character vector or factor containing another set of groupings, to be compared to \code{alt}.
#' @param mode String indicating whether to return the ratio, the number of pairs or the Rand index.
#' @param adjusted Logical scalar indicating whether the adjusted Rand index should be returned.
#'
#' @details
#' Recall that the Rand index calculation consists of four numbers:
#' \describe{
#' \item{\eqn{a}}{The number of pairs of cells in the same cluster in \code{ref} and the same cluster in \code{alt}.}
#' \item{\eqn{b}}{The number of pairs of cells in different clusters in \code{ref} and different clusters in \code{alt}.}
#' \item{\eqn{c}}{The number of pairs of cells in the same cluster in \code{ref} and different clusters in \code{alt}.}
#' \item{\eqn{d}}{The number of pairs of cells in different clusters in \code{ref} but the same cluster in \code{alt}.}
#' }
#' The Rand index is then computed as \eqn{a + b} divided by \eqn{a + b + c + d}, i.e., the total number of pairs.
#'
#' We can break these numbers down into values for each cluster or pair of clusters in \code{ref}.
#' For each cluster, we compute its value of \eqn{a}, 
#' i.e., the number of pairs of cells in \emph{that} cluster that are also in the same cluster in \code{alt}.
#' Similarly, for each pair of clusters in \code{ref}, we compute its value of \eqn{b},
#' i.e., the number of pairs of cells that have one cell in each of those clusters 
#' and also belong in different clusters in \code{alt}.
#'
#' This process provides more information about the specific similarities or differences between \code{ref} and \code{alt},
#' rather than coalescing all the values into a single statistic. 
#' For example, it is now possible to see which specific clusters from \code{ref} are not reproducible in \code{alt},
#' or which specific partitions between pairs of clusters are not reproducible.
#' In the default output, such events can be diagnosed by looking for low entries in the ratio matrix;
#' on the other hand, values close to 1 indicate that \code{ref} is almost perfectly recapitulated by \code{alt}.
#'
#' If \code{adjusted=TRUE}, we adjust all counts by subtracting their expected values under a model of random permutations.
#' This accounts for differences in the number and sizes of clusters within and between \code{ref} and \code{alt},
#' in a manner that mimics the calculation of adjusted Rand index (ARI).
#' We subtract expectations on a per-cluster or per-cluster-pair basis for \eqn{a} and \eqn{b}, respectively;
#' we also redefine the \dQuote{total} number of cell pairs for each cluster or cluster pair based on the denominator of the ARI. 
#' 
#' @return
#' If \code{mode="ratio"}, a square numeric matrix is returned with number of rows equal to the number of unique levels in \code{ref}.
#' Each diagonal entry is the ratio of the per-cluster \eqn{a} to the total number of pairs of cells in that cluster.
#' Each off-diagonal entry is the ratio of the per-cluster-pair \eqn{b} to the total number of pairs of cells for that pair of clusters.
#' Lower-triangular entries are set to \code{NA}.
#' If \code{adjusted=TRUE}, counts and totals are both adjusted prior to computing the ratio.
#'
#' If \code{mode="pairs"}, a list is returned containing \code{correct} and \code{total},
#' both of which are square numeric matrices of the same arrangement as described above.
#' However, \code{correct} contains the actual numbers \eqn{a} (diagonal) and \eqn{b} (off-diagonal) rather than the ratios,
#' while \code{total} contains the total number of cell pairs in each cluster or pair of clusters.
#' If \code{adjusted=TRUE}, both matrices are adjusted by subtracting the random expectations from the counts.
#'
#' If \code{mode="index"}, a numeric scalar is returned containing the Rand index (or ARI, if \code{adjusted=TRUE}).
#' 
#' @author Aaron Lun
#' @examples
#' m <- matrix(runif(10000), ncol=10)
#' 
#' clust1 <- kmeans(m,3)$cluster
#' clust2 <- kmeans(m,5)$cluster
#'
#' ratio <- pairwiseRand(clust1, clust2)
#' ratio
#'
#' # Getting the raw counts:
#' pairwiseRand(clust1, clust2, mode="pairs")
#' 
#' # Computing the original Rand index.
#' pairwiseRand(clust1, clust2, mode="index")
#'
#' @seealso
#' \code{\link{pairwiseModularity}}, which applies the same breakdown to the cluster modularity.
#'
#' @export
pairwiseRand <- function(ref, alt, mode=c("ratio", "pairs", "index"), adjusted=TRUE) {
    ref <- as.factor(ref)
    alt <- as.factor(alt)
    all.lev <- levels(ref)

    correct <- total <- matrix(NA_real_, length(all.lev), length(all.lev), 
        dimnames=list(all.lev, all.lev))

    for (i in seq_along(all.lev)) {
        left <- alt[ref==all.lev[i]]
        tab.left <- table(left)

        for (j in seq_len(i)) {
            right <- alt[ref==all.lev[j]]
            tab.right <- table(right)

            if (i!=j) {
                total.pairs <- sum(tab.left) * sum(tab.right) 
                correct.pairs <- total.pairs - sum(tab.left * tab.right)
            } else {
                correct.pairs <- sum(choose(tab.left, 2))
                total.pairs <- choose(sum(tab.left), 2)
            }

            correct[j,i] <- correct.pairs
            total[j,i] <- total.pairs
        }
    }

    if (adjusted) {
        refn <- table(ref)
        altn <- table(alt)

        # Computing the probability that two randomly chosen cells end up with
        # the same 'alt' label. This is used to compute the expected value for
        # each entry of 'correct'.
        same.alt.n <- sum(choose(altn, 2))
        total.n <- choose(length(alt), 2)
        same.alt.p <- same.alt.n/total.n

        # The LHS of the ARI's denominator is an average of the maximum number
        # of same-cluster pairs in 'ref' and 'alt'. We distribute this number
        # to each cluster in 'ref' according to its number of same-cluster cell
        # pairs. We generalize this for cluster pairs, even though this is not
        # technically used for the calculation of the ARI. 
        same.ref.n <- sum(choose(refn, 2))
        same.alt.mult <- same.alt.n/same.ref.n
        diff.alt.mult <- (total.n - same.alt.n)/(total.n - same.ref.n)

        for (i in seq_along(all.lev)) {
            leftn <- refn[all.lev[i]]
            for (j in seq_len(i)) {
                rightn <- refn[all.lev[j]]
                ref.total <- total[j, i]

                if (i!=j) {
                    expected <- leftn * rightn * (1  - same.alt.p)
                    alt.total <- diff.alt.mult * ref.total
                } else {
                    expected <- choose(leftn, 2) * same.alt.p
                    alt.total <- same.alt.mult * ref.total
                }

                correct[j,i] <- correct[j,i] - expected
                total[j,i] <- 0.5 * (ref.total + alt.total) - expected
            }
        }
    }

    mode <- match.arg(mode)
    if (mode=="ratio") {
        correct/total
    } else if (mode=="index") {
        if (!adjusted) {
            sum(correct, na.rm=TRUE)/sum(total, na.rm=TRUE)
        } else {
            sum(diag(correct))/sum(diag(total))
        }
    } else {
        list(correct=correct, total=total)
    }
}

# Tests the pairwiseRand() function.
# library(testthat); library(bluster); source('test-rand.R')

m <- matrix(rnorm(1000), ncol=10)

set.seed(8000)
clust1 <- kmeans(m,3)$cluster
clust2 <- kmeans(m,5)$cluster

test_that("pairwiseRand gives the expected output", {
    ratio <- pairwiseRand(clust1, clust2, adjusted=FALSE)

    expect_identical(dim(ratio), c(3L, 3L))
    expect_identical(rownames(ratio), as.character(1:3))
    expect_identical(colnames(ratio), as.character(1:3))

    expect_true(all(is.na(ratio[lower.tri(ratio)])))

    vals <- ratio[upper.tri(ratio, diag=TRUE)]
    expect_true(all(!is.na(vals)))
    expect_true(all(vals >= 0))
    expect_true(all(vals <= 1))

    # Correct values emitted.
    ratio <- pairwiseRand(clust1, clust1, adjusted=FALSE)
    expect_true(all(ratio[upper.tri(ratio, diag=TRUE)]==1))
})

test_that("pairwiseRand mimics the original rand index", {
    full <- pairwiseRand(clust1, clust2, mode="pairs", adjusted=FALSE)

    # Applying a reference calculation.
    status1 <- outer(clust1, clust1, "==")
    status2 <- outer(clust2, clust2, "==")

    keep <- upper.tri(status1)
    status1 <- status1[keep]
    status2 <- status2[keep]

    a <- sum(status1 & status2)
    b <- sum(!status2 & !status1)

    expect_equal(a, sum(diag(full$correct)))
    expect_equal(b, sum(full$correct[upper.tri(full$correct)]))

    rand <- sum(full$correct, na.rm=TRUE)/sum(full$total, na.rm=TRUE)
    expect_identical(rand, (a+b)/choose(length(clust1), 2))
    expect_identical(rand, pairwiseRand(clust1, clust2, mode="index", adjusted=FALSE))
})

test_that("pairwiseRand computes the adjusted Rand index", {
    tab <- table(clust1, clust2)
    same1 <- sum(choose(table(clust1), 2))
    same2 <- sum(choose(table(clust2), 2))
    total <- choose(length(clust1), 2)

    expected <-  same1 * same2 / total
    ref <- (sum(choose(tab, 2)) - expected) / (0.5 * (same1 + same2) - expected)

    ari <- pairwiseRand(clust1, clust2, mode="index", adjusted=TRUE)
    expect_equal(ari, ref)

    # Not entirely sure why this is true... but whatever!
    paired <- pairwiseRand(clust1, clust2, mode="pairs", adjusted=TRUE)
    expect_equal(ari, sum(paired$correct, na.rm=TRUE)/sum(paired$total, na.rm=TRUE))

    ari <- pairwiseRand(clust1, clust1, mode="index", adjusted=TRUE)
    expect_equal(ari, 1)
})

test_that("pairwiseRand handles silly inputs", {
    ref <- pairwiseRand(clust1, clust2)

    # Respects empty levels in 'ref'.
    clustX <- factor(clust1, levels=1:5)
    out <- pairwiseRand(clustX, clust2)

    expect_identical(dim(out), c(5L, 5L))
    expect_identical(rownames(out), as.character(1:5))
    expect_identical(colnames(out), as.character(1:5))
    expect_identical(out[1:3,1:3], ref)

    # Ignores empty levels in 'alt'.
    clustY <- factor(clust2, levels=1:5)
    out <- pairwiseRand(clust1, clustY)
    expect_identical(ref, out)

    # Behaves sensibly with zero-length inputs.
    out <- pairwiseRand(clust1[0], clust2[0])
    expect_identical(dim(out), c(0L, 0L))
})

#' Build a nearest-neighbor graph
#'
#' Build a shared or k-nearest-neighbors graph of observations for downstream community detection.
#'
#' @param x A matrix-like object containing expression values for each observation (row) and dimension (column).
#' @param k An integer scalar specifying the number of nearest neighbors to consider during graph construction.
#' @param type A string specifying the type of weighting scheme to use for shared neighbors.
#' @param directed A logical scalar indicating whether the output of \code{buildKNNGraph} should be a directed graph.
#' @param BNPARAM A \linkS4class{BiocNeighborParam} object specifying the nearest neighbor algorithm.
#' @param BPPARAM A \linkS4class{BiocParallelParam} object to use for parallel processing.
#' @param indices An integer matrix where each row corresponds to an observation
#' and contains the indices of the \code{k} nearest neighbors (by increasing distance and excluding self) from that observation.
#' 
#' @details
#' The \code{makeSNNGraph} method builds a shared nearest-neighbour graph using observations as nodes.
#' For each observation, its \code{k} nearest neighbours are identified using the \code{\link{findKNN}} function,
#' based on distances between their expression profiles (Euclidean by default).
#' An edge is drawn between all pairs of observations that share at least one neighbour,
#' weighted by the characteristics of the shared nearest neighbors:
#' \itemize{
#' \item If \code{type="rank"}, the weighting scheme defined by Xu and Su (2015) is used.
#' The weight between two nodes is \eqn{k - r/2} where \eqn{r} is the smallest sum of ranks for any shared neighboring node.
#' For example, if one node was the closest neighbor of each of two nodes, the weight between the two latter nodes would be \eqn{k - 1}.
#' For the purposes of this ranking, each node has a rank of zero in its own nearest-neighbor set. 
#' More shared neighbors, or shared neighbors that are close to both observations, will generally yield larger weights.
#' \item If \code{type="number"}, the weight between two nodes is simply the number of shared nearest neighbors between them.
#' The weight can range from zero to \eqn{k + 1}, as the node itself is included in its own nearest-neighbor set.
#' This is a simpler scheme that is also slightly faster but does not account for the ranking of neighbors within each set.
#' \item If \code{type="jaccard"}, the weight between two nodes is the Jaccard similarity between the two sets of neighbors for those nodes.
#' This weight can range from zero to 1, and is a monotonic transformation of the weight used by \code{type="number"}.
#' It is provided for consistency with other clustering algorithms such as those in \pkg{seurat}.
#' }
#' 
#' The aim is to use the SNN graph to perform clustering of observations via community detection algorithms in the \pkg{igraph} package.
#' This is faster and more memory efficient than hierarchical clustering for large numbers of observations.
#' In particular, it avoids the need to construct a distance matrix for all pairs of observations.
#' Only the identities of nearest neighbours are required, which can be obtained quickly with methods in the \pkg{BiocNeighbors} package.
#' 
#' The choice of \code{k} controls the connectivity of the graph and the resolution of community detection algorithms.
#' Smaller values of \code{k} will generally yield smaller, finer clusters, while increasing \code{k} will increase the connectivity of the graph and make it more difficult to resolve different communities.
#' The value of \code{k} can be roughly interpreted as the anticipated size of the smallest subpopulation.
#' If a subpopulation in the data has fewer than \code{k+1} observations, \code{buildSNNGraph} and \code{buildKNNGraph} will forcibly construct edges between observations in that subpopulation and observations in other subpopulations. 
#' This increases the risk that the subpopulation will not form its own cluster as it is more interconnected with the rest of the observations in the dataset.
#' 
#' Note that the setting of \code{k} here is slightly different from that used in SNN-Cliq.
#' The original implementation considers each observation to be its first nearest neighbor that contributes to \code{k}.
#' In \code{buildSNNGraph}, the \code{k} nearest neighbours refers to the number of \emph{other} observations.
#' 
#' The \code{makeKNNGraph} method builds a simpler k-nearest neighbour graph.
#' Observations are again nodes, and edges are drawn between each observation and its k-nearest neighbours.
#' No weighting of the edges is performed.
#' In theory, these graphs are directed as nearest neighour relationships may not be reciprocal.
#' However, by default, \code{directed=FALSE} such that an undirected graph is returned.
#'
#' The \code{neighborsToSNNGraph} and \code{neighborsToKNNGraph} functions operate directly on a matrix of nearest neighbor indices,
#' obtained using functions like \code{\link{findKNN}}.
#' This may be useful for constructing a graph from precomputed nearest-neighbor search results.
#' Note that the user is responsible for ensuring that the indices are valid (i.e., \code{range(indices)} is positive and no greater than \code{max(indices)}).
#' 
#' @return
#' A \link{graph} where nodes are cells and edges represent connections between nearest neighbors.
#' For \code{buildSNNGraph}, these edges are weighted by the number of shared nearest neighbors.
#' For \code{buildKNNGraph}, edges are not weighted but may be directed if \code{directed=TRUE}.
#' 
#' @author 
#' Aaron Lun, with KNN code contributed by Jonathan Griffiths.
#' 
#' @seealso
#' See \code{\link{make_graph}} for details on the graph output object.
#' 
#' See \code{\link{cluster_walktrap}}, \code{\link{cluster_louvain}} and related functions in \pkg{igraph} for clustering based on the produced graph.
#' 
#' Also see \code{\link{findKNN}} for specifics of the nearest-neighbor search.
#' 
#' @references
#' Xu C and Su Z (2015).
#' Identification of cell types from single-cell transcriptomes using a novel clustering method.
#' \emph{Bioinformatics} 31:1974-80
#' 
#' @examples
#' m <- matrix(rnorm(10000), ncol=10)
#'
#' g <- makeSNNGraph(m)
#' clusters <- igraph::cluster_fast_greedy(g)$membership
#' table(clusters)
#'
#' # Any clustering method from igraph can be used:
#' clusters <- igraph::cluster_walktrap(g)$membership
#' table(clusters)
#'
#' # Smaller 'k' usually yields finer clusters:
#' g <- makeSNNGraph(m, k=5)
#' clusters <- igraph::cluster_walktrap(g)$membership
#' table(clusters)
#'
#' @name makeSNNGraph
NULL

#' @export
#' @rdname makeSNNGraph
#' @importFrom BiocNeighbors KmknnParam findKNN
#' @importFrom BiocParallel SerialParam 
makeSNNGraph <- function(x, k=10, type=c("rank", "number", "jaccard"), BNPARAM=KmknnParam(), BPPARAM=SerialParam()) { 
    nn.out <- findKNN(x, k=k, BNPARAM=BNPARAM, BPPARAM=BPPARAM, get.distance=FALSE)
    neighborsToSNNGraph(nn.out$index, type=match.arg(type))
}

#' @export
#' @rdname makeSNNGraph
#' @importFrom BiocParallel SerialParam
#' @importFrom BiocNeighbors KmknnParam findKNN
makeKNNGraph <- function(x, k=10, directed=FALSE, BNPARAM=KmknnParam(), BPPARAM=SerialParam()) { 
    nn.out <- findKNN(x, k=k, BNPARAM=BNPARAM, BPPARAM=BPPARAM, get.distance=FALSE)
    neighborsToKNNGraph(nn.out$index, directed=directed)
}

#' @export
#' @rdname makeSNNGraph
#' @importFrom igraph make_graph simplify E "E<-"
neighborsToSNNGraph <- function(indices, type=c("rank", "number", "jaccard")) {
    type <- match.arg(type)
    if (type=="rank") {
        g.out <- build_snn_rank(indices)
    } else {
        g.out <- build_snn_number(indices)
    }

    edges <- g.out[[1]] 
    weights <- g.out[[2]]
    if (type=="jaccard") {
        weights <- weights / (2 * (ncol(indices) + 1) - weights)
    }

    g <- make_graph(edges, directed=FALSE)
    E(g)$weight <- weights
    simplify(g, edge.attr.comb="first") # symmetric, so doesn't really matter.
}


#' @export
#' @rdname makeSNNGraph
#' @importFrom igraph make_graph simplify
neighborsToKNNGraph <- function(indices, directed=FALSE) {
    start <- as.vector(row(indices))
    end <- as.vector(indices)
    interleaved <- as.vector(rbind(start, end))
    
    if (directed) { 
        g <- make_graph(interleaved, directed=TRUE)
    } else {
        g <- make_graph(interleaved, directed=FALSE)
        g <- simplify(g, edge.attr.comb = "first")
    }
    g
}

# Tests the pairwiseModularity function.
# library(bluster); library(testthat); source("test-modularity.R")

set.seed(20001)
test_that("pairwiseModularity computes the correct values", {
    exprs <- matrix(rnorm(1000), ncol=10)
    g <- makeSNNGraph(exprs)

    random <- sample(5, nrow(exprs), replace=TRUE)
    out <- pairwiseModularity(g, random) 
    expect_equal(sum(diag(out)), igraph::modularity(g, random, weight=igraph::E(g)$weight))
    expect_equal(sum(out), 0)

    # Some basic checks on the expected values.
    out <- pairwiseModularity(g, random, get.weights=TRUE)
    expect_equal(sum(out$observed), sum(out$expected))
    expect_equal(sum(out$observed), sum(igraph::E(g)$weight))
})

set.seed(20001)
test_that("pairwiseModularity computes correct values in a more realistic scenario", {
    exprs <- matrix(rnorm(5000), ncol=50)
    exprs[ 1:20, 1:10] <- rnorm(200, 2)
    exprs[21:50,11:20] <- rnorm(300, 2)
    exprs[51:90,21:30] <- rnorm(400, 2)
    g <- makeSNNGraph(exprs, k=10)

    actual <- igraph::cluster_walktrap(g)
    out <- pairwiseModularity(g, actual$membership) 
    expect_equal(sum(diag(out)), igraph::modularity(g, actual$membership, weight=igraph::E(g)$weight))
    expect_equal(sum(out), 0)

    # Some basic checks on the expected values.
    out <- pairwiseModularity(g, actual$membership, get.weights=TRUE)
    expect_equal(sum(out$observed), sum(out$expected))
    expect_equal(sum(out$observed), sum(igraph::E(g)$weight))
})

set.seed(20002)
test_that("pairwiseModularity handles unweighted graphs correctly", {
    exprs <- matrix(rnorm(1000), ncol=10)
    g <- makeKNNGraph(exprs)

    random <- sample(5, nrow(exprs), replace=TRUE)
    out <- pairwiseModularity(g, random) 
    expect_equal(sum(diag(out)), igraph::modularity(g, random))
    expect_equal(sum(out), 0)

    # Some basic checks on the expected values.
    out <- pairwiseModularity(g, random, get.weights=TRUE)
    expect_equal(sum(out$observed), sum(out$expected))
    expect_equal(sum(out$observed), length(igraph::E(g)))
})

set.seed(20003)
test_that("pairwiseModularity handles directed graphs correctly", {
    exprs <- matrix(rnorm(1000), ncol=10)
    g <- makeKNNGraph(exprs, directed=TRUE)

    random <- sample(5, nrow(exprs), replace=TRUE)
    out <- pairwiseModularity(g, random)
    ref <- pairwiseModularity(igraph::as.undirected(g, mode="each"), random) 
    expect_identical(out, ref)
})

set.seed(20003)
test_that("pairwiseModularity handles self-loops correctly", {
    exprs <- matrix(rnorm(1000), ncol=10)
    g <- makeSNNGraph(exprs)

    g <- igraph::add_edges(g, rep(nrow(exprs), each=2), weight=10)
    random <- sample(5, nrow(exprs), replace=TRUE)
    out <- pairwiseModularity(g, random) 
    expect_equal(sum(diag(out)), igraph::modularity(g, random, weight=igraph::E(g)$weight))

    # Works for unweighted graphs.
    exprs <- matrix(rnorm(10000), ncol=10)
    g <- makeKNNGraph(exprs)

    g <- igraph::add_edges(g, rep(nrow(exprs), each=2))
    random <- sample(5, nrow(exprs), replace=TRUE)
    out <- pairwiseModularity(g, random) 
    expect_equal(sum(diag(out)), igraph::modularity(g, random))
})

# Checks the construction of the SNN graph.
# library(bluster); require(testthat); source("test-make-snn.R")

library(igraph)

check <- function(vals, k=10, type="rank") {
    g <- makeSNNGraph(vals, k=k, type=type) 
    nobs <- nrow(vals)
    expect_identical(seq_len(nobs), as.vector(V(g)))
    nn.out <- BiocNeighbors::findKNN(vals, k=k)

    for (i in seq_len(nobs)) { 
        inn <- c(i, nn.out$index[i,])
        collected <- numeric(nobs)

        for (j in seq_len(nobs)) {
            jnn <- c(j, nn.out$index[j,])
            shared <- intersect(inn, jnn)
            if (length(shared)==0) {
                next
            }
            if (type=="rank") {
                s <- k + 1 - 0.5*(match(shared, inn) + match(shared, jnn))
                collected[j] <- max(c(s, 1e-6))
            } else if (type=="number") {
                collected[j] <- max(length(shared), 1e-6)
            } else {
                collected[j] <- max(length(shared), 1e-6) / length(union(inn, jnn))
            }
        }
        collected[i] <- 0
        expect_equal(collected, g[i])
    }
    return(NULL)
}

set.seed(20000)
nobs <- 200
ndim <- 50

test_that("makeSNNGraph gives same results as a reference", {
    dummy <- matrix(rnorm(ndim*nobs), nrow=nobs)
    check(dummy, k=10)
    
    dummy <- matrix(rnorm(ndim*nobs), nrow=nobs)
    check(dummy, k=20)
    
    dummy <- matrix(rnorm(ndim*nobs), nrow=nobs)
    check(dummy, k=5)

    # Checking 'number' mode.  
    dummy <- matrix(rnorm(ndim*nobs), nrow=nobs)
    check(dummy, k=10, type="number")
    
    dummy <- matrix(rnorm(ndim*nobs), nrow=nobs)
    check(dummy, k=20, type="number")
    
    dummy <- matrix(rnorm(ndim*nobs), nrow=nobs)
    check(dummy, k=5, type="number")

    # Checking 'jaccard' mode.  
    dummy <- matrix(rnorm(ndim*nobs), nrow=nobs)
    check(dummy, k=10, type="jaccard")
    
    dummy <- matrix(rnorm(ndim*nobs), nrow=nobs)
    check(dummy, k=20, type="jaccard")
    
    dummy <- matrix(rnorm(ndim*nobs), nrow=nobs)
    check(dummy, k=5, type="jaccard")
})

test_that("makeSNNGraph fails on silly inputs", {
    dummy <- matrix(rnorm(ndim*20), nrow=20)
    expect_warning(out <- makeSNNGraph(dummy, k=50), "capped")
    expect_warning(out2 <- makeSNNGraph(dummy, k=nrow(dummy)-1L), NA)
    expect_identical(out[], out2[])

    expect_error(makeSNNGraph(dummy[,0]), NA) # shouldn't fail, but shouldn't generate anything particularly useful.

    expect_warning(expect_error(makeSNNGraph(dummy[0,]), "must be positive"), "capped")
})

# Checking that makeKNNGraph also works.

KMAKE <- function(dummy, k, directed=FALSE) { 
    nobs <- nrow(dummy)
    collated <- matrix(0, nobs, nobs)
    td <- t(dummy)
    for (cell in seq_len(nobs)) {
        d2 <- colSums((td[,cell] - td)^2)
        chosen <- setdiff(order(d2), cell)[seq_len(k)]
        collated[cell,chosen] <- 1
        if (!directed) { 
            collated[chosen,cell] <- 1
        }
    }
    as(collated, "dgCMatrix")
}

test_that("makeKNNGraph works correctly", {
    dummy <- matrix(rnorm(ndim*nobs), nrow=nobs)
    g <- makeKNNGraph(dummy, k=10)
    expect_false(is.directed(g))
    expect_equal(g[], KMAKE(dummy, k=10))

    g <- makeKNNGraph(dummy, k=10, directed=TRUE)
    expect_true(is.directed(g))
    expect_equal(g[], KMAKE(dummy, k=10, directed=TRUE))

    g <- makeKNNGraph(dummy, k=20)
    expect_equal(g[], KMAKE(dummy, k=20))
})

#' Compute pairwise modularity
#' 
#' Calculate the modularity of each pair of clusters from a graph, based on a null model of random connections between nodes.
#' 
#' @param graph A \link{graph} object from \pkg{igraph}, usually where each node represents an observation.
#' @param clusters Factor specifying the cluster identity for each node.
#' @param get.weights Logical scalar indicating whether the observed and expected edge weights should be returned, rather than the modularity.
#' @param as.ratio Logical scalar indicating whether the log-ratio of observed to expected weights should be returned.
#' 
#' @return
#' By default, an upper triangular numeric matrix of order equal to the number of clusters is returned.
#' Each entry corresponds to a pair of clusters and is proportional to the difference between the observed and expected edge weights between those clusters.
#' 
#' If \code{as.ratio=TRUE}, an upper triangular numeric matrix is again returned.
#' Here, each entry is equal to the ratio between the observed and expected edge weights.
#' 
#' If \code{get.weights=TRUE}, a list is returned containing two upper triangular numeric matrices. 
#' The \code{observed} matrix contains the observed sum of edge weights between and within clusters,
#' while the \code{expected} matrix contains the expected sum of edge weights under the random model.
#' 
#' @details
#' This function computes a modularity score in the same manner as that from \code{\link{modularity}}.
#' The modularity is defined as the (scaled) difference between the observed and expected number of edges between nodes in the same cluster.
#' The expected number of edges is defined by a null model where edges are randomly distributed among nodes.
#' The same logic applies for weighted graphs, replacing the number of edges with the summed weight of edges.
#' 
#' Whereas \code{\link{modularity}} returns a modularity score for the entire graph, \code{pairwiseModularity} provides scores for the individual clusters.
#' The sum of the diagonal elements of the output matrix should be equal to the output of \code{\link{modularity}} 
#' (after supplying weights to the latter, if necessary).
#' A well-separated cluster should have mostly intra-cluster edges and a high modularity score on the corresponding diagonal entry,
#' while two closely related clusters that are weakly separated will have many inter-cluster edges and a high off-diagonal score.
#'
#' In practice, the modularity may not the most effective metric for evaluating cluster separatedness.
#' This is because the modularity is proportional to the number of observations, so larger clusters will naturally have a large score regardless of separation.
#' An alternative approach is to set \code{as.ratio=TRUE}, which returns the ratio of the observed to expected weights for each entry of the matrix.
#' This adjusts for differences in cluster size and improves resolution of differences between clusters.
#'
#' Directed graphs are treated as undirected inputs with \code{mode="each"} in \code{\link{as.undirected}}.
#' In the rare case that self-loops are present, these will also be handled correctly.
#' @author
#' Aaron Lun
#' 
#' @seealso
#' \code{\link{makeSNNGraph}}, for one method to construct \code{graph}.
#'
#' \code{\link{modularity}}, for the calculation of the entire graph modularity.
#'
#' \code{\link{pairwiseRand}}, which applies a similar breakdown to the Rand index.
#' 
#' @examples
#' m <- matrix(runif(10000), ncol=10)
#' clust.out <- clusterRows(m, BLUSPARAM=NNGraphParam(), full=TRUE)
#' clusters <- clust.out$clusters
#' g <- clust.out$objects$graph
#' 
#' # Examining the modularity values directly.
#' out <- pairwiseModularity(g, clusters)
#' out
#' 
#' # Compute the ratio instead, for visualization
#' # (log-transform to improve range of colors).
#' out <- pairwiseModularity(g, clusters, as.ratio=TRUE)
#' image(log2(out+1))
#'
#' # This can also be used to construct a graph of clusters,
#' # for use in further plotting, a.k.a. graph abstraction.
#' # (Fiddle with the scaling values for a nicer plot.)
#' g2 <- igraph::graph_from_adjacency_matrix(out, mode="upper",
#'     diag=FALSE, weighted=TRUE)
#' plot(g2, edge.width=igraph::E(g2)$weight*10,
#'     vertex.size=sqrt(table(clusters))*2)
#' 
#' # Alternatively, get the edge weights directly:
#' out <- pairwiseModularity(g, clusters, get.weights=TRUE)
#' out
#'
#' @export
#' @importFrom Matrix diag diag<-
#' @importFrom igraph is.directed
pairwiseModularity <- function(graph, clusters, get.weights=FALSE, as.ratio=FALSE) {
    by.clust <- split(seq_along(clusters), clusters)
    uclust <- names(by.clust)
    nclust <- length(uclust)

    mod.mat <- matrix(0, nclust, nclust)
    dimnames(mod.mat) <- list(uclust, uclust)
    clust.total <- numeric(nclust)
    names(clust.total) <- uclust

    fullmat <- graph[]

    # Calculating the observed weight within/between clusters.
    for (x in seq_along(by.clust)) {
        current <- by.clust[[x]] 
        for (y in seq_len(x)) { 
            other <- by.clust[[y]]
            grmat <- fullmat[current,other,drop=FALSE]
            grsum <- sum(grmat)
                
            if (x==y) {
                old.diag <- sum(diag(grmat))
                diag(grmat) <- 0

                self.sum <- sum(grmat)
                if (!is.directed(graph)) {
                    # Only count undirected edges within a cluster once. 
                    self.sum <- self.sum/2
                } else {
                    # Need to count directed edges between different nodes twice.
                    grsum <- grsum + self.sum
                }

                # Self-edges contribute twice to total node weight,
                # according to igraph::modularity.
                grsum <- grsum + old.diag

                mod.mat[x,y] <- self.sum + old.diag 
            } else {
                if (is.directed(graph)) {
                    grsum <- grsum + sum(graph[other,current])
                }
                mod.mat[y,x] <- grsum
            }

            # If x==y, this is equivalent to adding edge weights for each node twice.
            # THIS IS DELIBERATE; the total per-node weight is that for all edges 
            # involving each node, so edges between nodes in the same cluster are 
            # allowed to be double-counted here when aggregating node weights per cluster.
            clust.total[x] <- clust.total[x] + grsum
            if (x!=y) {
                clust.total[y] <- clust.total[y] + grsum
            }
        }
    }

    # Calcuating the expected weight if they were randomly distributed. 
    # Note some effort is involved in 'folding' it into an upper triangular.
    total.weight <- sum(mod.mat)
    clust.prop <- clust.total/sum(clust.total)
    expected.mat <- tcrossprod(clust.prop) * total.weight

    expected.mat[lower.tri(expected.mat)] <- 0
    old.diag <- diag(expected.mat)
    expected.mat <- expected.mat * 2
    diag(expected.mat) <- old.diag

    if (get.weights) {
        list(observed=mod.mat, expected=expected.mat)
    } else if (as.ratio) {
        mod.mat/expected.mat
    } else {
        1/total.weight * (mod.mat - expected.mat)
    }
}

#' Assess cluster stability by bootstrapping
#'
#' Generate bootstrap replicates and recluster on them to determine the stability of clusters with respect to sampling noise.
#'
#' @param x A numeric matrix-like object containing observations in the rows and variables in the columns.
#' If \code{transposed=TRUE}, observations are assumed to be in the columns instead.
#' @param FUN A function that takes \code{x} as its first argument and returns a vector or factor of cluster identities.
#' @param clusters A vector or factor of cluster identities equivalent to that obtained by calling \code{FUN(x, ...)}.
#' This is provided as an additional argument in the case that the clusters have already been computed,
#' in which case we can save a single round of computation.
#' @param iterations A positive integer scalar specifying the number of bootstrap iterations.
#' @param average String specifying the method to use to average across bootstrap iterations.
#' @param ... Further arguments to pass to \code{FUN} to control the clustering procedure.
#' @param compare A function that accepts the original clustering and the bootstrapped clustering,
#' and returns a numeric vector or matrix containing some measure of similarity between them - see Details.
#' @param transposed Logical scalar indicating that resampling should be done on the columns instead.
#' @param mode,adjusted Further arguments to pass to \code{\link{pairwiseRand}} when \code{compare=NULL}.
#'
#' @return 
#' If \code{compare=NULL} and \code{mode="ratio"}, a numeric matrix is returned with upper triangular entries set to the ratio of the adjusted observation pair counts (see \code{?\link{pairwiseRand}}) for each pair of clusters in \code{clusters}.
#' Each ratio is averaged across bootstrap iterations as specified by \code{average}.
#' 
#' If \code{compare=NULL} and \code{mode="index"}, a numeric scalar containing the average ARI between \code{clusters} and the bootstrap replicates across iterations is returned.
#'
#' If \code{compare} is provided, a numeric array of the same type as the output of \code{compare} is returned, containing the average statistic(s) across bootstrap replicates.
#'
#' @details
#' Bootstrapping is conventionally used to evaluate the precision of an estimator by applying it to an \emph{in silico}-generated replicate dataset.
#' We can (ab)use this framework to determine the stability of the clusters given the original dataset.
#' We sample observations with replacement from \code{x}, perform clustering with \code{FUN} and compare the new clusters to \code{clusters}.
#'
#' For comparing clusters, we compute the ratio matrix from \code{\link{pairwiseRand}} and average its values across bootstrap iterations.
#' High on-diagonal values indicate that the corresponding cluster remains coherent in the bootstrap replicates,
#' while high off-diagonal values indicate that the corresponding pair of clusters are still separated in the replicates.
#' If a single value is necessary, we can instead average the adjusted Rand indices across iterations with \code{mode="index"}.
#' 
#' We use the ratio matrix by default as it is more interpretable than a single value like the ARI or the Jaccard index (see the \pkg{fpc} package).
#' It focuses on the relevant differences between clusters, allowing us to determine which aspects of a clustering are stable.
#' For example, A and B may be well separated but A and C may not be, which is difficult to represent in a single stability measure for A.
#' If our main interest lies in the A/B separation, we do not want to be overly pessimistic about the stability of A, even though it might not be well-separated from all other clusters.
#'
#' @section Using another comparison function:
#' We can use a different method for comparing clusterings by setting \code{compare}.
#' This is expected to be a function that takes two arguments - 
#' the original clustering first, and the bootstrapped clustering second - 
#' and returns some kind of numeric scalar, vector or matrix containing 
#' statistics for the similarity or difference between the original and bootstrapped clustering.
#' These statistics are then averaged across all bootstrap iterations.
#' 
#' Any numeric output of \code{compare} is acceptable as long as the dimensions are only dependent on the \emph{levels} of the original clustering - including levels that have no observations, due to resampling! - and thus do not change across bootstrap iterations.
#'
#' @section Statistical note on bootstrap comparisons:
#' Technically speaking, some mental gymnastics are required to compare the original and bootstrap clusters in this manner.
#' After bootstrapping, the sampled observations represent distinct entities from the original dataset (otherwise it would be difficult to treat them as independent replicates) for which the original clusters do not immediately apply.
#' Instead, we assume that we perform label transfer using a nearest-neighbors approach - which, in this case, is the same as using the original label for each observation, as the nearest neighbor of each resampled observation to the original dataset is itself.
#'
#' Needless to say, bootstrapping will only generate replicates that differ by sampling noise.
#' Real replicates will differ due to composition differences, variability in expression across individuals, etc.
#' Thus, any stability inferences from bootstrapping are likely to be overly optimistic.
#' 
#' @author Aaron Lun
#' @examples
#' m <- matrix(runif(10000), ncol=10)
#' 
#' # BLUSPARAM just gets passed to the default FUN=clusterRows:
#' bootstrapStability(m, BLUSPARAM=KmeansParam(4), iterations=10)
#'
#' # Defining your own clustering function:
#' kFUN <- function(x) kmeans(x, 2)$cluster  
#' bootstrapStability(m, FUN=kFUN)
#'
#' # Using an alternative comparison, in this case the Rand index:
#' bootstrapStability(m, FUN=kFUN, compare=pairwiseRand)
#'
#' @seealso
#' \code{\link{clusterRows}}, for the default clustering function.
#'
#' \code{\link{pairwiseRand}}, for the calculation of the ARI.
#' 
#' @export
#' @importFrom stats median
bootstrapStability <- function(x, FUN=clusterRows, clusters=NULL, iterations=20, 
    average=c("median", "mean"), ..., compare=NULL, mode="ratio", adjusted=TRUE, transposed=FALSE)
{
    if (is.null(clusters)) {
        clusters <- FUN(x, ...)
    }
    clusters <- as.factor(clusters)

    if (iterations <= 0L) {
        stop("'iterations' must be a positive integer")
    }

    collated <- vector("list", iterations)
    if (is.null(compare)) {
        compare <- function(...) pairwiseRand(..., mode=mode, adjusted=adjusted)
    }

    for (i in seq_len(iterations)) {
        if (!transposed) {
            chosen <- sample(nrow(x), nrow(x), replace=TRUE)
            resampled <- x[chosen,,drop=FALSE]
        } else {
            chosen <- sample(ncol(x), ncol(x), replace=TRUE)
            resampled <- x[,chosen,drop=FALSE]
        }
        reclusters <- FUN(resampled, ...)
        collated[[i]] <- compare(clusters[chosen], reclusters)
    }

    if (length(unique(lapply(collated, dim))) > 1L) { 
        stop("'compare' output should have constant dimension")
    }

    # A robust way of computing the average that handles NAs.
    as.mat <- do.call(cbind, lapply(collated, as.numeric))

    average <- match.arg(average)
    if (average=="mean") {
        averaged <- rowMeans(as.mat, na.rm=TRUE)
    } else {
        averaged <- apply(as.mat, 1, median, na.rm=TRUE)
    }

    dim(averaged) <- dim(collated[[1]])
    dimnames(averaged) <- dimnames(collated[[1]])
    averaged 
}

# This tests the bootstrapStability functionality.
# library(testthat); library(bluster); source("test-bootstrap.R")

set.seed(50000)
nobs <- 700
ndims <- 100

set.seed(500001)
test_that("bootstrapStability works correctly with clear separation", {
    dummy <- matrix(rnbinom(nobs*ndims, mu=10, size=20), ncol=ndims)
    known.clusters <- sample(3, nobs, replace=TRUE)
    dummy[known.clusters==1L, 1:30] <- 0
    dummy[known.clusters==2L,31:60] <- 0
    dummy[known.clusters==3L,61:90] <- 0

    output <- bootstrapStability(dummy, FUN=function(x) { kmeans(log10(x+1), 3)$cluster })
    expect_true(all(output[upper.tri(output, diag=FALSE)] > 0.5))
    expect_true(all(diag(output) > 0.5))

    # Works with the mean.
    output <- bootstrapStability(dummy, FUN=function(x) { kmeans(log10(x+1), 3)$cluster }, average="mean")
    expect_true(all(output[upper.tri(output, diag=FALSE)] > 0.5))
    expect_true(all(diag(output) > 0.5))

    # Continues to work if vector is a character or factor.
    output <- bootstrapStability(dummy, FUN=function(x) { c("X", "Y", "Z")[kmeans(log10(x+1), 3)$cluster] })
    expect_true(all(output[upper.tri(output, diag=FALSE)] > 0.5))
    expect_true(all(diag(output) > 0.5))

    output <- bootstrapStability(dummy, FUN=function(x) { factor(kmeans(log10(x+1), 3)$cluster) })
    expect_true(all(output[upper.tri(output, diag=FALSE)] > 0.5))
    expect_true(all(diag(output) > 0.5))
})

set.seed(500002)
test_that("bootstrapStability works correctly with poor separation", {
    dummy <- matrix(rnbinom(nobs*ndims, mu=10, size=20), ncol=ndims)
    output <- bootstrapStability(dummy, FUN=function(x) { kmeans(log10(x+1), 3)$cluster })

    expect_true(all(output[upper.tri(output, diag=TRUE)] < 0.1))
    expect_true(all(output[upper.tri(output, diag=TRUE)] > -0.1))
    expect_true(all(diag(output) < 0.1))
    expect_true(all(diag(output) > -0.1))

    # Works with the mean.
    output <- bootstrapStability(dummy, FUN=function(x) { kmeans(log10(x+1), 3)$cluster }, average="mean")
    expect_true(all(output[upper.tri(output, diag=TRUE)] < 0.1))
    expect_true(all(output[upper.tri(output, diag=TRUE)] > -0.1))
    expect_true(all(diag(output) < 0.1))
    expect_true(all(diag(output) > -0.1))
})

set.seed(500004)
test_that("bootstrapStability works when some clusters are not in the bootstrap.", {
    dummy <- matrix(rnorm(10), nrow=10)

    # Guaranteed to get missing clusters from resampling.
    output <- bootstrapStability(dummy, FUN=function(x) { seq_len(ncol(x)) })

    expect_identical(rownames(output), as.character(seq_len(ncol(dummy))))
    expect_true(all(output[upper.tri(output, diag=FALSE)]==0))
})

set.seed(500003)
test_that("bootstrapStability works with alternative comparison functions", {
    dummy <- matrix(rnorm(nobs*20), nrow=nobs, ncol=20)
    output <- bootstrapStability(dummy, FUN=function(x) { kmeans(x, 3)$cluster }, compare=function(...) 1)
    expect_identical(output, 1)
})

set.seed(500003)
test_that("bootstrapStability works with transposition", {
    dummy <- matrix(rnorm(nobs*20), nrow=nobs, ncol=20)

    set.seed(20)
    ref <- bootstrapStability(dummy, FUN=function(x) { kmeans(x, 3)$cluster })

    set.seed(20)
    output <- bootstrapStability(t(dummy), FUN=function(x) { kmeans(t(x), 3)$cluster }, transposed=TRUE)

    expect_identical(ref, output)
})

set.seed(500004)
test_that("other miscellaneous tests for bootstrapStability", {
    dummy <- matrix(rnorm(nobs*20), nrow=nobs, ncol=20)
    expect_error(bootstrapStability(dummy, FUN=function(x) { seq_len(ncol(x)) }, iterations=0), "positive")
})

#' Divisive analysis clustering
#'
#' Use the \code{\link{diana}} function to perform divisive analysis clustering.
#'
#' @inheritParams HclustParam
#' @param stand Further arguments to pass to \code{\link{diana}}.
#'
#' @author Aaron Lun
#'
#' @details
#' To modify an existing DianaParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#'
#' If \code{cut.fun=NULL}, \code{cut.dynamic=FALSE} and \code{cut.params} does not have \code{h} or \code{k},
#' \code{\link{clusterRows}} will automatically set \code{h} to half the tree height when calling \code{\link{cutree}}. 
#'
#' @return 
#' The \code{DianaParam} constructor will return a \linkS4class{DianaParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
#' (a list containing \code{diana}, the function output; \code{dist}, the dissimilarity matrix; and \code{hclust}, a \link{hclust} object created from \code{diana}).
#'
#' @examples
#' clusterRows(iris[,1:4], DianaParam())
#' clusterRows(iris[,1:4], DianaParam(metric="manhattan"))
#'
#' @seealso
#' \code{\link{diana}}, which actually does all the heavy lifting.
#'
#' \linkS4class{HclustParam}, for the more commonly used implementation of hierarchical clustering.
#'
#' @name DianaParam-class
#' @docType class
#' @aliases 
#' show,DianaParam-method
#' .defaultScalarArguments,DianaParam-method
NULL

#' @export
#' @rdname DianaParam-class
setClass("DianaParam", contains="HierarchicalParam", slots=c(stand="ANY"))

#' @export
setMethod(".defaultScalarArguments", "DianaParam", function(x) c(callNextMethod(), stand="logical"))

#' @export
#' @rdname DianaParam-class
DianaParam <- function(metric=NULL, stand=NULL, cut.fun=NULL, cut.dynamic=FALSE, cut.params=list()) {
    new("DianaParam", metric=metric, stand=stand, cut.fun=cut.fun, cut.dynamic=cut.dynamic, cut.params=cut.params)
}

#' @export
#' @rdname DianaParam-class
#' @importFrom cluster diana 
#' @importFrom stats as.hclust
setMethod("clusterRows", c("ANY", "DianaParam"), function(x, BLUSPARAM, full=FALSE) {
    args <- c(
        list(
            quote(as.matrix(x)),
            keep.diss = TRUE,
            keep.data = FALSE
        ),
        .extractScalarArguments(BLUSPARAM)
    )

    out <- do.call(diana, args)
    dst <- out$diss
    hcl <- as.hclust(out)

    clusters <- .cut_hierarchical(hcl, dst, BLUSPARAM)
    if (full) {
        list(clusters=clusters, objects=list(diana=out, dist=dst, hclust=hcl))
    } else {
        clusters
    }
})

# Tests the DianaParam class.
# library(bluster); library(testthat); source('test-diana-param.R')

test_that("DianaParam constructor and utilities work correctly", {
    X <- DianaParam()
    expect_output(show(X), "DianaParam")

    expect_identical(X[["metric"]], NULL)
    X[["metric"]] <- "euclidean"
    expect_identical(X[["metric"]], "euclidean")

    X <- DianaParam(cut.params=list(metric="euclidean"))
    expect_identical(X[["cut.params"]], list(metric="euclidean"))
    X[["cut.params"]] <- list(whee=2)
    expect_identical(X[["cut.params"]], list(whee=2))

    # other show metrics
    expect_output(show(DianaParam(metric="euclidean")), "euclidean")
    expect_output(show(DianaParam(cut.dynamic=TRUE)), "cutreeDynamic")
    expect_output(show(DianaParam(cut.fun=identity)), "custom")
})

test_that("DianaParam validity works correctly", {
    expect_error(DianaParam(cut.dynamic=NA), "non-missing")
    expect_error(DianaParam(metric=1), "character")
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(1000), ncol=10)
    out <- clusterRows(m, DianaParam())
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))

    out2 <- clusterRows(m, DianaParam(cut.params=list(h=2)))
    expect_identical(length(out2), nrow(m))
    expect_false(identical(out, out2))

    out <- clusterRows(m, DianaParam(cut.params=list(k=5)))
    expect_identical(length(out), nrow(m))
    expect_identical(nlevels(out), 5L)

    ref <- clusterRows(m, DianaParam(cut.params=list(k=3)))
    out <- clusterRows(m, DianaParam(cut.fun=function(x) cutree(x, k=3)))
    expect_identical(out, ref)

    full <- clusterRows(m, DianaParam(), full=TRUE)
    expect_s3_class(full$objects$dist, "dist")
    expect_s3_class(full$objects$hclust, "hclust")

    # Default cut works as expected.
    maxh <- max(full$objects$hclust$height)
    check <- clusterRows(m, DianaParam(cut.params=list(h=maxh/2)))
    expect_identical(full$clusters, check)
})

test_that("clusterRows works with the dynamic tree cut", {
    m <- matrix(runif(1000), ncol=10)
    out <- clusterRows(m, DianaParam(cut.dynamic=TRUE))
    expect_true(is.factor(out))
    expect_identical(names(out), NULL)
    expect_identical(length(out), nrow(m))
})

test_that("clusterRows works with custom distance functions", {
    m <- matrix(runif(1000), ncol=10)

    out <- clusterRows(m, DianaParam(metric="manhattan"), full=TRUE)
    expect_equal(as.matrix(out$objects$dist), as.matrix(dist(m, method="manhattan")))

    vegout <- clusterRows(m, DianaParam(metric = "canberra", dist.fun = vegan::vegdist), full=TRUE)
    expected <- vegan::vegdist(m, "canberra")
    expect_equal(as.matrix(vegout$objects$dist), as.matrix(expected))
})

#' Clustering Large Applications
#'
#' Run the CLARA algorithm, an extension of the PAM method for large datasets.
#'
#' @param centers An integer scalar specifying the number of centers.
#' Alternatively, a function that takes the number of observations and returns the number of centers.
#' @param metric,stand,samples,sampsize Further arguments to pass to \code{\link{clara}}.
#' Set to the function defaults if not supplied.
#' @inheritParams clusterRows
#' @param BLUSPARAM A \linkS4class{ClaraParam} object.
#' @param full Logical scalar indicating whether the full PAM statistics should be returned.
#'
#' @author Aaron Lun
#'
#' @details
#' This class usually requires the user to specify the number of clusters beforehand.
#' However, we can also allow the number of clusters to vary as a function of the number of observations.
#' The latter is occasionally useful, e.g., to allow the clustering to automatically become more granular for large datasets.
#'
#' To modify an existing ClaraParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#'
#' Note that \code{clusterRows} will always use \code{rngR=TRUE}, for greater consistency with other algorithms of the \linkS4class{FixedNumberParam} class;
#' and \code{pamLike=TRUE}, for consistency with the PAM implementation from which it was derived.
#'
#' @return
#' The \code{ClaraParam} constructor will return a \linkS4class{ClaraParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
#' (a list containing \code{clara}, the direct output of \code{\link{clara}}).
#'
#' @examples
#' clusterRows(iris[,1:4], ClaraParam(centers=4))
#' clusterRows(iris[,1:4], ClaraParam(centers=4, sampsize=50))
#' clusterRows(iris[,1:4], ClaraParam(centers=sqrt))
#' @seealso
#' \code{\link{clara}}, which actually does all the heavy lifting.
#'
#' \linkS4class{PamParam}, for the original PAM algorithm.
#'
#' @name ClaraParam-class
#' @docType class
#' @aliases
#' show,ClaraParam-method
#' .defaultScalarArguments,ClaraParam-method
NULL

#' @export
#' @rdname ClaraParam-class
setClass("ClaraParam", contains="FixedNumberParam", slots=c(metric="ANY", stand="ANY", samples="ANY", sampsize="ANY"))

#' @export
#' @rdname ClaraParam-class
ClaraParam <- function(centers, metric=NULL, stand=NULL, samples=NULL, sampsize=NULL) {
    if (!is.function(centers)) {
        centers <- as.integer(centers)
    }
    new("ClaraParam", centers=centers, metric=metric, stand=stand, samples=samples, sampsize=sampsize)
}

#' @export
setMethod("show", "ClaraParam", function(object) {
    callNextMethod()
    .showScalarArguments(object)
})

#' @export
setMethod(".defaultScalarArguments", "ClaraParam", function(x) 
    c(metric="character", stand="logical", samples="numeric", sampsize="numeric"))

#' @export
#' @rdname ClaraParam-class
#' @importFrom cluster clara
setMethod("clusterRows", c("ANY", "ClaraParam"), function(x, BLUSPARAM, full=FALSE) {
    centerx <- centers(BLUSPARAM, n=nrow(x))

    args <- c(
        list(
            quote(as.matrix(x)),
            k=centerx,
            keep.data=FALSE,
            rngR=TRUE,
            pamLike=TRUE
        ),
        .extractScalarArguments(BLUSPARAM)
    )

    stats <- do.call(clara, args)
    clusters <- factor(stats$clustering)

    if (full) {
        list(clusters=clusters, objects=list(clara=stats))
    } else {
        clusters
    }
})

# Tests the ClaraParam class.
# library(bluster); library(testthat); source('test-clara-param.R')

test_that("ClaraParam constructor and utilities work correctly", {
    X <- ClaraParam(centers=10)
    expect_output(show(X), "ClaraParam")
    expect_output(show(X), "centers: 10")

    expect_identical(X[["centers"]], 10L)
    X[["centers"]] <- 2L
    expect_identical(X[["centers"]], 2L)

    X <- ClaraParam(centers=log)
    expect_true(is.function(X[["centers"]]))
    expect_identical(centers(X, 20), as.integer(round(log(20))))

    X <- ClaraParam(centers=10, sampsize=50)
    expect_identical(X[["sampsize"]], 50)
    X[["sampsize"]] <- 10L
    expect_identical(X[["sampsize"]], 10L)
})

test_that("ClaraParam validity works correctly", {
    expect_error(ClaraParam(-1), "positive")
    expect_error(ClaraParam(10, sampsize=LETTERS), "numeric scalar")
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(1000), ncol=10)
    out <- clusterRows(m, ClaraParam(5))
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))
    expect_identical(nlevels(out), 5L)

    set.seed(9999)
    out <- clusterRows(m, ClaraParam(sqrt))
    expect_identical(length(out), nrow(m))
    expect_equal(nlevels(out), round(sqrt(nrow(m))))

    set.seed(9999)
    full <- clusterRows(m, ClaraParam(sqrt), full=TRUE)
    expect_identical(out, full$cluster)
    expect_s3_class(full$objects$clara, "clara")

    # Responds to the options.
    set.seed(100000)
    suppressWarnings(ref <- cluster::clara(m, k=20, samples=10, sampsize=50, rngR=TRUE, pamLike=TRUE)$clustering)
    set.seed(100000)
    suppressWarnings(out <- clusterRows(m, ClaraParam(centers=20, samples=10, sampsize=50)))
    expect_identical(factor(ref), out)
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

build_snn_rank <- function(neighbors) {
    .Call('_bluster_build_snn_rank', PACKAGE = 'bluster', neighbors)
}

build_snn_number <- function(neighbors) {
    .Call('_bluster_build_snn_number', PACKAGE = 'bluster', neighbors)
}

sum_neighbor_weights <- function(nclusters, neighbors, clusters, weights) {
    .Call('_bluster_sum_neighbor_weights', PACKAGE = 'bluster', nclusters, neighbors, clusters, weights)
}

#' Mini-batch k-means clustering
#'
#' Run the mini-batch k-means \code{\link[mbkmeans]{mbkmeans}} function with the specified number of centers within \code{\link{clusterRows}}.
#' This sacrifices some accuracy for speed compared to the standard k-means algorithm.
#' Note that this requires installation of the \pkg{mbkmeans} package.
#'
#' @param centers An integer scalar specifying the number of centers.
#' Alternatively, a function that takes the number of observations and returns the number of centers.
#' @inheritParams clusterRows
#' @param batch_size,max_iters,num_init,init_fraction,initializer,calc_wcss,early_stop_iter,tol,BPPARAM
#' Further arguments to pass to \code{\link[mbkmeans]{mbkmeans}}.
#' @param BLUSPARAM A \linkS4class{MbkmeansParam} object.
#' @param full Logical scalar indicating whether the full mini-batch k-means statistics should be returned.
#'
#' @author Stephanie Hicks
#'
#' @details
#' This class usually requires the user to specify the number of clusters beforehand.
#' However, we can also allow the number of clusters to vary as a function of the number of observations.
#' The latter is occasionally useful, e.g., to allow the clustering to automatically become more granular for large datasets.
#'
#' To modify an existing MbkmeansParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#'
#' For \code{batch_size} and \code{init_fraction}, a value of \code{NULL} means that the default arguments in the \code{\link[mbkmeans]{mbkmeans}} function signature are used.
#' These defaults are data-dependent and so cannot be specified during construction of the MbkmeansParam object, but instead are defined within the \code{clusterRows} method.
#'
#' @return 
#' The \code{MbkmeansParam} constructor will return a \linkS4class{MbkmeansParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
#' (a list containing \code{mbkmeans}, the direct output of \code{\link[mbkmeans]{mbkmeans}}).
#'
#' @examples
#' clusterRows(iris[,1:4], MbkmeansParam(centers=3))
#' clusterRows(iris[,1:4], MbkmeansParam(centers=3, batch_size=10))
#' clusterRows(iris[,1:4], MbkmeansParam(centers=3, init_fraction=0.5))
#' @seealso
#' \code{\link[mbkmeans]{mbkmeans}} from the \pkg{mbkmeans} package, which actually does all the heavy lifting.
#'
#' \linkS4class{KmeansParam}, for dispatch to the standard k-means algorithm.
#'
#' @name MbkmeansParam-class
#' @docType class
#' @aliases 
#' show,MbkmeansParam-method
NULL

#' @export
#' @rdname MbkmeansParam-class
setClass("MbkmeansParam", contains="FixedNumberParam", 
    slots=c(batch_size="integer_OR_NULL", 
        max_iters="integer", 
        num_init="integer", 
        init_fraction="numeric_OR_NULL", 
        initializer="character",
        calc_wcss="logical", 
        early_stop_iter="integer", 
        tol="numeric", 
        BPPARAM="BiocParallelParam"))

setValidity2("MbkmeansParam", function(object) {
    msg <- character(0)

    msg <- c(msg, .check_positive_slots(object, c("batch_size", "max_iters", "init_fraction", "num_init", "early_stop_iter", "tol")))

    if (!is.null(val <- object@init_fraction)) {
        if (val > 1) {
            msg <- c(msg, "'init_fraction' cannot be greater than 1")
        }
    }

    msg <- c(msg, .check_nonna_slots(object, c("initializer", "calc_wcss")))

    if (length(msg)) {
        return(msg)
    }
    TRUE
})

#' @export
setMethod("show", "MbkmeansParam", function(object) {
    callNextMethod()
    cat(sprintf("batch_size: %s\n", if (is.null(object@batch_size)) "default" else object@batch_size)) 
    cat(sprintf("max_iters: %s\n", object@max_iters))
    cat(sprintf("num_init: %s\n", object@num_init))
    cat(sprintf("init_fraction: %s\n", if (is.null(object@init_fraction)) "default" else object@init_fraction)) 
    cat(sprintf("initializer: %s\n", object@initializer))
    cat(sprintf("calc_wcss: %s\n", object@calc_wcss))
    cat(sprintf("early_stop_iter: %s\n", object@early_stop_iter))
    cat(sprintf("tol: %s\n", object@tol))
    cat(sprintf("BPPARAM: %s\n", class(object@BPPARAM)[1]))
})

#' @export
#' @rdname MbkmeansParam-class
#' @importFrom BiocParallel SerialParam
MbkmeansParam <- function(centers, 
    batch_size = NULL,
    max_iters = 100,
    num_init = 1,
    init_fraction = NULL,
    initializer = "kmeans++",
    calc_wcss = FALSE,
    early_stop_iter = 10,
    tol = 1e-04,
    BPPARAM = SerialParam())
{
    if (!is.function(centers)) {
        centers <- as.integer(centers)
    }
    if (!is.null(batch_size)) {
        batch_size <- as.integer(batch_size)
    }

    new("MbkmeansParam", centers=centers, 
        batch_size=batch_size, 
        max_iters=as.integer(max_iters), 
        num_init=as.integer(num_init),
        init_fraction=init_fraction, 
        initializer=initializer, 
        calc_wcss=calc_wcss,
        early_stop_iter=as.integer(early_stop_iter), 
        tol=tol, 
        BPPARAM=BPPARAM)
}

#' @export
#' @rdname MbkmeansParam-class
setMethod("clusterRows", c("ANY", "MbkmeansParam"), function(x, BLUSPARAM, full=FALSE) {
    centerx <- centers(BLUSPARAM, n=nrow(x))

    # Setting the data-dependent defaults.
    batch_size <- BLUSPARAM[["batch_size"]]
    if (is.null(batch_size)) {
        batch_size <- min(500, nrow(x))
    }

    init_fraction <- BLUSPARAM[["init_fraction"]]
    if (is.null(init_fraction)) {
        init_fraction <- batch_size/nrow(x)
    }

    if (is.matrix(x) && is.object(x)) {
        # their C++ code can't handle S3 classes that inherit from base matrix objects.
        x <- unclass(x)
    }

    stats <- mbkmeans::mbkmeans(t(x), clusters=centerx,
        batch_size=batch_size,
        max_iters=BLUSPARAM[["max_iters"]], 
        num_init=BLUSPARAM[["num_init"]],
        init_fraction=init_fraction,
        initializer=BLUSPARAM[["initializer"]], 
        calc_wcss=BLUSPARAM[["calc_wcss"]],
        early_stop_iter=BLUSPARAM[["early_stop_iter"]],
        tol=BLUSPARAM[["tol"]], 
        BPPARAM=BLUSPARAM[["BPPARAM"]])

    vec_clusters <- factor(stats$Clusters)
    if (full) {
        list(clusters=vec_clusters, objects=list(mbkmeans=stats))
    } else {
        vec_clusters
    }
})

# Tests the MbkmeansParam class.
# library(bluster); library(testthat); source('test-mbkmeans-param.R')

test_that("MbkmeansParam constructor and utilities work correctly", {
    X <- MbkmeansParam(centers=10)
    expect_output(show(X), "MbkmeansParam")
    expect_output(show(X), "centers: 10")
    expect_output(show(X), "batch_size: default")
    expect_output(show(X), "BPPARAM: SerialParam")

    expect_identical(X[["centers"]], 10L)
    X[["centers"]] <- 2L
    expect_identical(X[["centers"]], 2L)

    X <- MbkmeansParam(centers=log)
    expect_true(is.function(X[["centers"]]))

    X <- MbkmeansParam(centers=10, batch_size=500)
    expect_identical(X[["batch_size"]], 500L)
    X[["batch_size"]] <- 100L 
    expect_identical(X[["batch_size"]], 100L)
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(10000), ncol=10)

    set.seed(100)
    out <- clusterRows(m, MbkmeansParam(5))
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))
    expect_identical(nlevels(out), 5L)

    # Consistent with the defaults in the original function. 
    set.seed(100)
    ref <- mbkmeans::mbkmeans(t(m), 5)
    expect_identical(out, factor(ref$Clusters))

    # Trying with fewer cells.
    m <- matrix(runif(1000), ncol=10)

    set.seed(100)
    out <- clusterRows(m, MbkmeansParam(5))
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))
    expect_identical(nlevels(out), 5L)

    set.seed(100)
    ref <- mbkmeans::mbkmeans(t(m), 5)
    expect_identical(out, factor(ref$Clusters))
})

test_that("clusterRows responds to the functions and full=TRUE", {
    m <- matrix(runif(10000), ncol=10)

    set.seed(9999)
    out <- clusterRows(m, MbkmeansParam(sqrt))
    expect_identical(length(out), nrow(m))
    expect_equal(nlevels(out), round(sqrt(nrow(m))))

    set.seed(9999)
    full <- clusterRows(m, MbkmeansParam(sqrt), full=TRUE)
    expect_identical(out, full$cluster)
    expect_true(is.list(full$objects))
    expect_identical(names(full$objects), "mbkmeans")
})

test_that("clusterRows responds to the options", {
    m <- matrix(runif(10000), ncol=10)
    set.seed(100000)
    suppressWarnings(ref <- mbkmeans::mbkmeans(t(m), 10, batch_size=120, max_iters=5, num_init=10, init_fraction=0.1)$Cluster)

    set.seed(100000)
    suppressWarnings(out <- clusterRows(m, MbkmeansParam(10, batch_size=120, max_iters=5, num_init=10, init_fraction=0.1)))
    expect_identical(factor(ref), out)
})

#' Compute neighborhood purity
#'
#' Use a hypersphere-based approach to compute the \dQuote{purity} of each cluster based on the number of contaminating observations from different clusters in its neighborhood.
#'
#' @param x A numeric matrix-like object containing observations in rows and variables in columns.
#' @param clusters Vector of length equal to \code{ncol(x)}, specifying the cluster assigned to each observation.
#' @param k Integer scalar specifying the number of nearest neighbors to use to determine the radius of the hyperspheres.
#' @param BNPARAM A \linkS4class{BiocNeighborParam} object specifying the nearest neighbor algorithm.
#' This should be an algorithm supported by \code{\link{findNeighbors}}.
#' @param BPPARAM A \linkS4class{BiocParallelParam} object indicating whether and how parallelization should be performed across genes.
#' @param weighted A logical scalar indicating whether to weight each observation in inverse proportion to the size of its cluster.
#' Alternatively, a numeric vector of length equal to \code{clusters} containing the weight to use for each observation.
#'
#' @return
#' A \linkS4class{DataFrame} with one row per observation in \code{x} and the columns:
#' \itemize{
#' \item \code{purity}, a numeric field containing the purity value for the current observation.
#' \item \code{maximum}, the cluster with the highest proportion of observations neighboring the current observation.
#' }
#' Row names are defined as the row names of \code{x}.
#' 
#' @details
#' The purity of a cluster is quantified by creating a hypersphere around each observation in the cluster
#' and computing the proportion of observations in that hypersphere from the same cluster.
#' If all observations in a cluster have proportions close to 1, this indicates that the cluster is highly pure,
#' i.e., there are few observations from other clusters in its region of the coordinate space.
#' The distribution of purities for each cluster can be used as a measure of separation from other clusters.
#'
#' In most cases, the majority of observations of a cluster will have high purities, corresponding to observations close to the cluster center.
#' A fraction of observations will have low values as these lie at the boundaries of two adjacent clusters.
#' A high degree of over-clustering will manifest as a majority of observations with purities close to zero.
#' The \code{maximum} field in the output can be used to determine the identity of the cluster 
#' with the greatest presence in a observation's neighborhood, usually an adjacent cluster for observations lying on the boundary.
#' 
#' The choice of \code{k} is used only to determine an appropriate value for the hypersphere radius.
#' We use hyperspheres as this is robust to changes in density throughout the coordinate space,
#' in contrast to computing purity based on the proportion of k-nearest neighbors in the same cluster.
#' For example, the latter will fail most obviously when the size of the cluster is less than \code{k}.
#'
#' @section Weighting by frequency:
#' By default, purity values are computed after weighting each observation by the reciprocal of the number of observations in the same cluster.
#' Otherwise, clusters with more observations will have higher purities as any contamination is offset by the bulk of observations,
#' which would compromise comparisons of purities between clusters.
#' One can interpret the weighted purities as the expected value after downsampling all clusters to the same size.
#'
#' Advanced users can achieve greater control by manually supplying a numeric vector of weights to \code{weighted}.
#' For example, we may wish to check the purity of batches after batch correction in single-cell RNA-seq.
#' In this application, \code{clusters} should be set to the \emph{batch blocking factor} (not the cluster identities!)
#' and \code{weighted} should be set to 1 over the frequency of each combination of cell type and batch.
#' This accounts for differences in cell type composition between batches when computing purities.
#' 
#' If \code{weighted=FALSE}, no weighting is performed.
#'
#' @author Aaron Lun
#' @examples
#' m <- matrix(runif(1000), ncol=10)
#' clusters <- clusterRows(m, BLUSPARAM=NNGraphParam())
#' out <- neighborPurity(m, clusters)
#' boxplot(split(out$purity, clusters))
#'
#' # Mocking up a stronger example:
#' centers <- matrix(rnorm(30), nrow=3)
#' clusters <- sample(1:3, 1000, replace=TRUE)
#' y <- centers[clusters,,drop=FALSE]
#' y <- y + rnorm(length(y))
#' 
#' out2 <- neighborPurity(y, clusters)
#' boxplot(split(out2$purity, clusters))
#'
#' @export
#' @importFrom BiocNeighbors KmknnParam buildIndex findKNN findNeighbors
#' @importFrom BiocParallel SerialParam bpstart bpstop bpisup
#' @importFrom stats median
#' @importFrom S4Vectors DataFrame
neighborPurity <- function(x, clusters, k=50, weighted=TRUE, BNPARAM=KmknnParam(), BPPARAM=SerialParam()) {
    x <- as.matrix(x)
    idx <- buildIndex(x, BNPARAM=BNPARAM)

    if (!bpisup(BPPARAM) && !is(BPPARAM, "MulticoreParam")) {
        bpstart(BPPARAM)
        on.exit(bpstop(BPPARAM))
    }

    dist <- median(findKNN(BNINDEX=idx, k=k, BNPARAM=BNPARAM, BPPARAM=BPPARAM, 
        last=1, warn.ties=FALSE, get.index=FALSE)$distance)
    nout <- findNeighbors(BNINDEX=idx, threshold=dist, BNPARAM=BNPARAM, get.distance=FALSE)$index

    # Constructing weights.
    if (isFALSE(weighted)) {
        w <- rep(1, nrow(x))
    } else if (isTRUE(weighted)) {
        w <- 1/table(clusters)
        w <- as.numeric(w[as.character(clusters)])
    } else {
        w <- as.numeric(weighted)
        if (length(w)!=length(clusters)) {
            stop("numeric 'weighted' must be of same length as 'clusters'")           
        }
    }

    uclust <- sort(unique(clusters)) # do NOT use as.factor(), we want to preserve type.
    m <- match(clusters, uclust)
    stopifnot(all(!is.na(m)))
    aggregated <- sum_neighbor_weights(length(uclust), nout, m - 1L, w)

    targets <- t(aggregated[[1]])
    totals <- aggregated[[2]]

    DataFrame( 
        purity=targets[cbind(seq_along(m), m)]/totals,
        maximum=uclust[max.col(targets, ties.method="first")],
        row.names=rownames(x) 
    )
}

# Tests the neighborPurity() function.
library(testthat); library(bluster); #source('test-purity.R')

set.seed(70000)
test_that('neighborPurity yields correct output for pure clusters', {
    clusters <- rep(seq_len(10), each=100)
    y <- matrix(clusters, ncol=5, nrow=length(clusters))
    y <- jitter(y)

    out <- neighborPurity(y, clusters)
    expect_identical(nrow(out), nrow(y))

    expect_true(all(out$purity==1))
    expect_true(all(out$maximum==clusters))
})

set.seed(70001)
test_that('neighborPurity yields correct output for compromised clusters', {
    y <- matrix(rep(0:1, each=100), ncol=5, nrow=200)
    y <- jitter(y)

    clusters <- rep(1:2, each=100)
    clusters[1] <- 2

    out <- neighborPurity(y, clusters)

    expect_true(out$purity[1] <= 0.05)
    expect_true(all(out$purity[-1] > 0.9))
    expect_identical(out$maximum[1], 1)
    expect_identical(out$maximum[-1], clusters[-1])
})

set.seed(700011)
test_that("neighborPurity handles the weighting correctly", {
    # Creating a bulk of points.
    y0 <- matrix(rnorm(10000), ncol=50)
    y1 <- y2 <- matrix(1000, ncol=50, nrow=10)

    y <- rbind(y1, y2, y0)
    clusters <- rep(1:3, c(nrow(y1), nrow(y2), nrow(y0)))
    out1 <- neighborPurity(y, clusters)$purity
    expect_true(all(abs(out1[1:20]-0.5) < 1e-8))
    expect_true(all(out1[-(1:20)]==1))

    # Unaffected by changes in the number of cells. Technically this should be
    # identical(), but the order of cells returned by findNeighbors() changes,
    # and this results in numeric precision changes due to order of addition of
    # double-precision values, resulting in very slightly different output.
    sub <- c(1:5, nrow(y1) + seq_len(nrow(y2) + nrow(y0)))
    out2 <- neighborPurity(y[sub,], clusters[sub])$purity
    expect_equal(out1[sub], out2)

    sub <- c(1, nrow(y1) + seq_len(nrow(y2) + nrow(y0)))
    out2 <- neighborPurity(y[sub,], clusters[sub])$purity
    expect_equal(out1[sub], out2)
})

set.seed(700012)
test_that("neighborPurity handles other weighting options", {
    # Creating a bulk of points.
    y0 <- matrix(rnorm(10000), ncol=50)
    y1 <- y2 <- matrix(1000, ncol=50, nrow=10)

    y <- rbind(y1, y2, y0)
    clusters <- rep(1:3, c(nrow(y1), nrow(y2), nrow(y0)))

    # Turning off weighting has no effect for balanced clusters.
    out1 <- neighborPurity(y, clusters, weighted=FALSE)$purity
    expect_true(all(abs(out1[1:20]-0.5) < 1e-8))
    expect_true(all(out1[-(1:20)]==1))

    # Turning off weighting has some effect for non-balanced clusters.
    sub <- c(1:5, nrow(y1) + seq_len(nrow(y2) + nrow(y0)))
    out2 <- neighborPurity(y[sub,], clusters[sub], weighted=FALSE)$purity
    expect_true(all(abs(out2[1:5]-1/3) < 1e-8))
    expect_true(all(abs(out2[6:15]-2/3) < 1e-8))
    expect_true(all(out2[-(1:20)]==1))

    # We can replace it with our own weighting to restore the balance.
    out3 <- neighborPurity(y[sub,], clusters[sub], weighted=rep(c(2, 1), c(5, nrow(y2)+nrow(y0))))$purity
    ref <- neighborPurity(y[sub,], clusters[sub])$purity
    expect_equal(out3, ref)
})

setValidity("FixedNumberParam", function(object) {
    msg <- character(0)

    centerx <- centers(object)
    if (is.integer(centerx)) {
        if (!.positive_number(centerx)) {
            msg <- c(msg, "integer 'centers' must be a positive scalar")
        }
    }

    if (length(msg)) {
        return(msg)
    }
    TRUE
})

#' @export
setMethod("centers", "FixedNumberParam", function(x, n=NULL) {
    xcenters <- x@centers
    if (is.integer(xcenters) || is.null(n)) {
        xcenters
    } else {
        as.integer(round(xcenters(n)))
    }
})

#' @export
setReplaceMethod("centers", "FixedNumberParam", function(x, value) {
    x@centers <- value
    x
})

#' @export
setMethod("show", "FixedNumberParam", function(object) {
    callNextMethod()
    cat(sprintf("centers: %s\n", if (is.function(object@centers)) "variable" else object@centers))
})

#' Affinity propogation 
#'
#' Use affinity propagation from the \pkg{apcluster} package to cluster observations.
#' Note that this requires the installation of the \pkg{apcluster} package.
#'
#' @inheritParams clusterRows
#' @param s A function that accepts a matrix of observations by dimensions and returns a similarity matrix.
#' If \code{NULL}, defaults to the output of \code{\link[apcluster]{negDistMat}} with \code{r=2}.
#' @param p,q Numeric scalars controlling the input preference, i.e., the resolution of the clustering.
#' These are passed to the \code{\link[apcluster]{apcluster}} function, where values of \code{NA} are the default.
#' @param maxits,convits,lam,nonoise
#' Further arguments to pass to the \code{\link[apcluster]{apcluster}} function.
#' @param BLUSPARAM A \linkS4class{AffinityParam} object.
#' @param full Logical scalar indicating whether the full affinity propagation statistics should be returned.
#' 
#' @details
#' To modify an existing AffinityParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#'
#' Setting \code{q} (and less typically, \code{p}) allows us to tune the resolution of the clustering.
#' In particular, when \code{p=NA}, it is computed based on the setting of \code{q}:
#' \itemize{
#' \item If the specified \code{q} lies in [0, 1], \code{p} is defined as the \code{q}-quantile of the finite similarities across all pairs of observations.
#' When \code{q=NA}, it defaults to 0.5.
#' \item If \code{q} is negative, \code{p} is defined as the \code{M + abs(M) * q} where \code{M} is the smallest finite similarity across all pairs.
#' This yields smaller \code{p} values while still responding to the scale of the similarities. 
#' }
#' The resulting value is used as the self-preference, i.e., the diagonal of the availability matrix.
#' Larger values yield more clusters as each data point is more inclined to form its own cluster.
#' 
#' @return
#' The \code{AffinityParam} constructor will return a \linkS4class{AffinityParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
#' (a list containing \code{similarity}, the similarity matrix; and \code{apcluster}, the direct output of \code{\link[apcluster]{apcluster}}).
#'
#' @author Aaron Lun
#' @examples
#' clusterRows(iris[,1:4], AffinityParam())
#' clusterRows(iris[,1:4], AffinityParam(q=0.9))
#' clusterRows(iris[,1:4], AffinityParam(s=apcluster::expSimMat()))
#'
#' @seealso
#' \code{\link[apcluster]{apcluster}} from the \pkg{apcluster} package, which does all of the heavy lifting.
#'
#' @name AffinityParam-class
#' @docType class
#' @aliases 
#' show,AffinityParam-method
NULL

#' @export
setClass("AffinityParam", contains="BlusterParam", slots=c(s="function_OR_NULL",
    p="numeric", q="numeric", maxits="integer", convits="integer", lam="numeric", nonoise="logical"))

#' @export
#' @rdname AffinityParam-class
AffinityParam <- function(s=NULL, p=NA, q=NA, maxits=1000, convits=100, lam=0.9, nonoise=FALSE) {
    new("AffinityParam", s=s, p=as.numeric(p), q=as.numeric(q), 
        maxits=as.integer(maxits), convits=as.integer(convits), 
        lam=as.numeric(lam), nonoise=as.logical(nonoise))
}

#' @export
setMethod("show", "AffinityParam", function(object) {
    callNextMethod()

    cat(sprintf("s: %s\n", if (is.null(object[["s"]])) "default" else "custom"))
    for (i in c("p", "q", "maxits", "convits", "lam", "nonoise")) {
        cat(sprintf("%s: %s\n", i, slot(object, i)))
    }
})

setValidity2("AffinityParam", function(object) {
    msg <- .check_positive_slots(object, c("maxits", "convits"))
    
    msg <- .check_nonna_slots(object, c("nonoise", "lam"))

    if (object@lam >= 1 || object@lam < 0.5) {
        msg <- c(msg, "'lam' should be in [0.5, 1)")
    }

    if (length(object[["p"]])!=1) {
        msg <- c(msg, "'p' should be of length 1")
    }

    q <- object[["q"]]
    if (length(q)!=1) {
        msg <- c(msg, "'q' should be of length 1")
    } else {
        if (!is.na(q) && q > 1) {
            msg <- c("'q' should lie in (-Inf, 1]")
        }
    }

    if (object@lam >= 1 || object@lam < 0.5) {
        msg <- c(msg, "'lam' should be in [0.5, 1)")
    }

    if (length(msg)) {
        return(msg)
    }
    TRUE
})

#' @export
#' @rdname AffinityParam-class
setMethod("clusterRows", c("ANY", "AffinityParam"), function(x, BLUSPARAM, full=FALSE) {
    s <- BLUSPARAM[["s"]]
    if (is.null(s)) {
        s <- apcluster::negDistMat(r=2)
    }
    mat <- s(as.matrix(x))

    p <- BLUSPARAM[["p"]]
    q <- BLUSPARAM[["q"]]
    if (is.na(p)) {
        if (!is.na(q) && q < 0) {
            keep <- mat > -Inf
            diag(keep) <- FALSE
            p <- min(mat[keep])
            p <- p + abs(p) * q
            q <- NA
        }
    }

    res <- apcluster::apcluster(s=mat, 
        p=p,
        q=q,
        maxits=BLUSPARAM[["maxits"]],
        convits=BLUSPARAM[["convits"]],
        lam=BLUSPARAM[["lam"]],
        nonoise=BLUSPARAM[["nonoise"]]
    )

    out <- apcluster::labels(res, type="enum")
    out <- factor(out)

    if (full) {
        list(clusters=out, objects=list(similarity=mat, apcluster=res))
    } else {
        out
    }
})

# Tests the AffinityParam class.
# library(bluster); library(testthat); source('test-affinity-param.R')

set.seed(1000)
test_that("AffinityParam constructor and utilities work correctly", {
    X <- AffinityParam()
    expect_output(show(X), "AffinityParam")
    expect_output(show(X), "s: default")
    expect_output(show(X), "p: NA")

    expect_identical(X[["s"]], NULL)
    X[["s"]] <- identity
    expect_identical(X[["s"]], identity)

    X <- AffinityParam(q=0.9)
    expect_identical(X[["q"]], 0.9)
    X[["q"]] <- 0.1
    expect_identical(X[["q"]], 0.1)
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(1000), ncol=10)

    out <- clusterRows(m, AffinityParam())
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))

    # Consistent with reference values.
    ref <- apcluster::apcluster(apcluster::negDistMat(m, r=2))
    expect_identical(out, factor(apcluster::labels(ref, type="enum")))

    # Trying with different parameters.
    set.seed(10)
    out <- clusterRows(m, AffinityParam(q=0.8))
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))

    set.seed(10)
    ref <- apcluster::apcluster(apcluster::negDistMat(m, r=2), q=0.8)
    expect_identical(out, factor(apcluster::labels(ref, type="enum")))
})

test_that("clusterRows responds to full=TRUE", {
    m <- matrix(runif(2960), ncol=10)
    out <- clusterRows(m, AffinityParam())

    full <- clusterRows(m, AffinityParam(), full=TRUE)
    expect_identical(out, full$cluster)
    expect_s4_class(full$objects$apcluster, "APResult")
})

set.seed(1010001)
test_that("clusterRows responds to the options", {
    m <- matrix(runif(1000), ncol=10)

    set.seed(1000)
    out <- clusterRows(m, AffinityParam(q=0.8, lam=0.7, s=apcluster::expSimMat()))

    set.seed(1000)
    ref <- apcluster::apcluster(apcluster::expSimMat(m), q=0.8, lam=0.7)
    expect_identical(out, factor(apcluster::labels(ref, type="enum")))

    # Responds to the negative q.
    set.seed(2000)
    out <- clusterRows(m, AffinityParam(q=-1, s=apcluster::expSimMat()))

    set.seed(2000)
    mat <- apcluster::expSimMat(m)
    ref <- apcluster::apcluster(mat, p=0)
    expect_identical(out, factor(apcluster::labels(ref, type="enum")))

    # Also works for negative distances.
    set.seed(300)
    out <- clusterRows(m, AffinityParam(q=-2, s=apcluster::negDistMat()))

    set.seed(300)
    mat <- apcluster::negDistMat(m)
    ref <- apcluster::apcluster(mat, p=min(mat) * 3)
    expect_identical(out, factor(apcluster::labels(ref, type="enum")))
})

#' Compare pairs of clusterings
#'
#' Compute the adjusted Rand index between all pairs of clusterings,
#' where larger values indicate a greater similarity between clusterings.
#'
#' @inheritParams linkClusters
#' @inheritParams pairwiseRand
#' 
#' @details
#' The aim of this function is to allow us to easily determine the relationships between clusterings.
#' For example, we might use this to determine which parameter settings have the greatest effect in a sweep by \code{\link{clusterSweep}}.
#' Alternatively, we could use this to obtain an \dQuote{ordering} of clusterings for visualization, e.g., with \pkg{clustree}.
#'
#' This function does not provide any insight into the relationships between individual clusters.
#' A large Rand index only means that two clusterings are similar but does not specify the corresponding set of clusters across clusterings.
#' For that task, we suggest using the \code{\link{linkClusters}} function instead.
#'
#' @return A symmetric square matrix of pairwise (adjusted) Rand indices between all pairs of clusters.
#'
#' @return Aaron Lun
#'
#' @examples
#' clusters <- list(
#'     nngraph = clusterRows(iris[,1:4], NNGraphParam()),
#'     hclust = clusterRows(iris[,1:4], HclustParam(cut.dynamic=TRUE)),
#'     kmeans = clusterRows(iris[,1:4], KmeansParam(20))
#' )
#'
#' aris <- compareClusterings(clusters)
#'
#' # Visualizing the relationships between clusterings.
#' # Here, k-means is forced to be least similar to the others.
#' ari.as.graph <- igraph::graph.adjacency(aris, mode="undirected", weighted=TRUE)
#' plot(ari.as.graph)
#'
#' # Obtain an ordering of clusterings, using the eigenvector 
#' # as a 1-dimensional summary of the matrix:
#' ev1 <- eigen(aris)$vectors[,1]
#' o <- order(ev1)
#' rownames(aris)[o]
#'
#' @seealso
#' \code{\link{linkClusters}}, which identifies relationships between individual clusters across clusterings.
#'
#' \code{\link{pairwiseRand}}, for calculation of the pairwise Rand index.
#' 
#' @export
compareClusterings <- function(clusters, adjusted=TRUE) {
    if (!length(unique(lengths(clusters)))) {
        stop("'clusters' must have elements of the same length")
    }

    xnames <- names(clusters)
    N <- length(xnames)
    output <- matrix(0, N, N, dimnames=list(xnames, xnames))

    for (x in seq_len(N)) {
        for (y in seq_len(x-1)) {
            output[x,y] <- pairwiseRand(clusters[[x]], clusters[[y]], mode="index", adjusted=adjusted)
        }
    }

    # Making it symmetric.
    output + t(output)
}

#' Create a graph between different clusterings
#'
#' Create a graph that links together clusters from different clusterings,
#' e.g., generated using different parameter settings or algorithms.
#' This is useful for identifying corresponding clusters between clusterings
#' and to create meta-clusters from multiple clusterings.
#'
#' @param clusters A list of factors or vectors where each entry corresponds to a clustering.
#' All vectors should be of the same length.
#' The list itself should usually be named with a suitable label for each clustering.
#' @param prefix Logical scalar indicating whether the cluster levels should be prefixed with its clustering.
#' If \code{clusters} is not named, numeric prefixes are used instead.
#' @param denominator String specifying how the strength of the correspondence between clusters should be computed.
#' @param x,y Factor or vector specifying a clustering of the same cells.
#'
#' @details
#' Links are only formed between clusters from different clusterings, e.g., between clusters \eqn{X} in clustering 1 and \eqn{Y} in clustering 2.
#' The edge weight of each link is set to the strength of the correspondence between the two clusters;
#' this is defined from the number of cells with those two labels in their respective clusterings.
#' A larger number of cells indicates that \eqn{X} and \eqn{Y} are corresponding clusters.
#'
#' Of course, the number of cells also depends on the total number of cells in each cluster.
#' To account for this, we normalize the strength by a function of the total number of cells in the two clusters.
#' The choice of function is determined by \code{denominator} and determines how the strength is adjusted for dissimilar cluster sizes.
#' \itemize{
#' \item For \code{"min"}, the number of shared cells is divided by the smaller of the totals between the two clusters.
#' \item For \code{"max"}, the number of shared cells is divided by the larger of the totals.
#' \item For \code{"union"}, the number of shared cells is divided by the size of the union of cells in the two clusters.
#' The result is equivalent to the Jaccard index.
#' }
#'
#' In situations where \eqn{X} splits into multiple smaller clusters \eqn{Y1}, \eqn{Y2}, etc. in another clustering,
#' \code{denominator="min"} will report strong links between \eqn{X} and its constituent subclusters while \code{"max"} and \code{"union"} will report weak links.
#' Conversely, \code{denominator="max"} and \code{"union"} can only form strong links when there is a 1:1 mapping between clusters in different clusterings.
#' This usually yields simpler correspondences between clusterings at the cost of orphaning some of the smaller subclusters.
#' \code{denominator="union"} is most stringent as it will penalize the presence of non-shared cells in both clusters, whereas \code{"max"} only does so for the larger cluster.
#'
#' The general idea is to use the graph returned by this function in visualization routines or for community-based clustering,
#' to identify \dQuote{clusters of clusters} that can inform about the relationships between clusterings.
#'
#' @return 
#' For \code{linkClusters}, a \link{graph} object where each node is a cluster level in one of the clusterings in \code{clusters}.
#' Edges are weighted by the strength of the correspondence between two clusters in different clusterings.
#'
#' For \code{linkClustersMatrix}, a matrix is returned where each row and column corresponds to a cluster in \code{x} and \code{y}, respectively.
#' Entries represent the strength of the correspondence between the associated clusters;
#' this is equivalent to a submatrix of the adjacency matrix from the graph returned by \code{linkClusters}. 
#'
#' @author Aaron Lun
#'
#' @examples
#' clusters <- list(
#'     nngraph = clusterRows(iris[,1:4], NNGraphParam()),
#'     hclust = clusterRows(iris[,1:4], HclustParam(cut.dynamic=TRUE)),
#'     kmeans = clusterRows(iris[,1:4], KmeansParam(5))
#' )
#'
#' g <- linkClusters(clusters)
#' plot(g)
#'
#' igraph::cluster_walktrap(g)
#'
#' # Results as a matrix, for two clusterings:
#' linkClustersMatrix(clusters[[1]], clusters[[2]], denominator="union")
#' @seealso
#' The \pkg{clustree} package, which provides another method for visualizing relationships between clusterings.
#'
#' \code{\link{compareClusterings}}, which computes similarities between the clusterings themselves.
#' @export
#' @importFrom igraph graph.adjacency
linkClusters <- function(clusters, prefix=TRUE, denominator=c("union", "min", "max")) {
    if (!length(unique(lengths(clusters)))) {
        stop("'clusters' must have elements of the same length")
    }

    totals <- lapply(clusters, table)
    N <- sum(lengths(totals))
    output <- matrix(0, N, N)

    newnames <- unlist(lapply(totals, names))
    if (prefix) {
        clust.names <- names(totals)
        if (is.null(clust.names)) {
            clust.names <- seq_along(totals)
        }
        newnames <- sprintf("%s.%s", rep(clust.names, lengths(totals)), newnames)
    }
    dimnames(output) <- list(newnames, newnames)

    denominator <- match.arg(denominator)
    lastx <- 0L

    for (x in seq_along(clusters)) {
        nx <- as.integer(totals[[x]])
        lasty <- 0L

        for (y in seq_len(x - 1L)) {
            ny <- as.integer(totals[[y]])
            ratio <- .compute_correspondence(clusters[[x]], clusters[[y]], denominator, nX=nx, nY=ny)
            output[lastx + seq_along(nx), lasty + seq_along(ny)] <- ratio
            lasty <- lasty + length(ny)
        }

        lastx <- lastx + length(nx)
    }

    graph.adjacency(output, mode="lower", weighted=TRUE)
}

#' @export
#' @rdname linkClusters
linkClustersMatrix <- function(x, y, denominator=c("union", "min", "max")) {
    out <- .compute_correspondence(x, y, denominator=match.arg(denominator))
    .untable(out)
}

.compute_correspondence <- function(X, Y, denominator, nX=as.integer(table(X)), nY=as.integer(table(Y))) {
    tab <- table(X, Y)

    if (denominator=="min") {
        denom <- outer(nX, nY, pmin)
    } else if (denominator=="union") {
        denom <- outer(nX, nY, `+`)
        denom <- denom - tab
    } else {
        denom <- outer(nX, nY, pmax)
    }

    tab/denom
}

# This tests the linkClusters function.
# library(testthat); library(bluster); source("test-link-clusters.R")

set.seed(9999999)
x1 <- sample(LETTERS[1:10], 100, replace=TRUE)
x2 <- sample(1:15, 100, replace=TRUE)
x3 <- sample(letters[1:5], 100, replace=TRUE)
together <- list(first=x1, second=x2, third=x3)

set.seed(99999999)
test_that("linkClusters works correctly", {
    linked <- linkClusters(together, denominator="min")
    ulinked <- linkClusters(together, denominator="union")
    xlinked <- linkClusters(together, denominator="max")

    # Prefixing works correctly.
    out <- names(igraph::V(linked))
    expect_true(all(grepl("^(first|second|third)", out)))

    edges <- igraph::ends(linked, igraph::E(linked))
    expect_identical(nrow(edges), length(unique(c(paste(x1, x2), paste(x1, x3), paste(x2, x3)))))

    expect_identical(edges, igraph::ends(ulinked, igraph::E(ulinked)))
    expect_identical(edges, igraph::ends(xlinked, igraph::E(xlinked)))

    # Edge weight calculations are performed correctly.
    ref.weights <- uref.weights <- xref.weights <- numeric(nrow(edges)) 
    for (e in seq_len(nrow(edges))) {
        left <- edges[e,1]
        left.cluster <- sub("\\..*", "", left)
        left.level <- sub(".*\\.", "", left)

        right <- edges[e,2]
        right.cluster <- sub("\\..*", "", right)
        right.level <- sub(".*\\.", "", right)

        has.left <- together[[left.cluster]]==left.level 
        has.right <- together[[right.cluster]]==right.level 
        shared <- sum(has.left & has.right)
        n.left <- sum(has.left)
        n.right <- sum(has.right)

        ref.weights[e] <- shared / min(n.left, n.right)
        uref.weights[e] <- shared / sum(has.left | has.right)
        xref.weights[e] <- shared / max(n.left, n.right)
    }

    expect_identical(ref.weights, igraph::E(linked)$weight)
    expect_identical(uref.weights, igraph::E(ulinked)$weight)
    expect_identical(xref.weights, igraph::E(xlinked)$weight)
})

set.seed(999999991)
test_that("linkClusters behaves correctly with unnumbered elements", {
    ref <- linkClusters(together)
    stripped <- ref[]
    nofix <- sub(".*\\.", "", rownames(stripped))
    colnames(stripped) <- nofix
    rownames(stripped) <- nofix

    linked <- linkClusters(together, prefix=FALSE)
    expect_identical(stripped, linked[])

    ref <- linkClusters(unname(together))
    together2 <- together
    names(together2) <- seq_along(together)
    linked <- linkClusters(together2)
    expect_identical(linked[], ref[])
})

test_that("linkClusters handles error states smoothly", {
    expect_error(linkClusters(list(letters, 1:10)), "same length")

    empty <- linkClusters(list(integer(0), character(0)))
    expect_identical(length(igraph::E(empty)), 0L)
    expect_identical(length(igraph::V(empty)), 0L)
})

#' Map nested clusterings 
#'
#' Map an alternative clustering to a reference clustering,
#' where the latter is expected to be nested within the former.
#'
#' @inheritParams pairwiseRand
#'
#' @return
#' A list containing:
#' \itemize{
#' \item \code{proportions}, a matrix where each row corresponds to one of the \code{alt} clusters and each column corresponds to one of the \code{ref} clusters.
#' Each matrix entry represents the proportion of cells in \code{alt} that are assigned to each cluster in \code{ref}.
#' (That is, the proportions across all \code{ref} clusters should sum to unity for each \code{alt} cluster.)
#' \item \code{alt.mapping}, a \linkS4class{DataFrame} with one row per cluster in \code{alt}.
#' This contains the columns \code{max}, a numeric vector specifying the maximum value of \code{statistic} for that \code{alt} cluster;
#' and \code{which}, a character vector specifying the \code{ref} cluster in which the maximum value occurs.
#' \item \code{ref.score}, a numeric vector of length equal to the number of \code{ref} clusters.
#' This represents the degree of nesting of \code{alt} clusters within each \code{ref} cluster, see Details.
#' }
#' 
#' @details
#' This function identifies mappings between two clusterings on the same set of cells where \code{alt} is potentially nested within \code{ref} (e.g., as it is computed at higher resolution).
#' To do so, we take each \code{alt} cluster and compute the the proportion of its cells that are derived from each \code{ref} cluster.
#' The corresponding \code{ref} cluster is identified as that with the highest proportion, as reported by the \code{which} field in the \code{mapping} DataFrame. 
#'
#' The quality of the mapping is determined by \code{max} in the output \code{mapping} DataFrame.
#' A low value indicates that \code{alt} does not have a clear counterpart in \code{ref}, representing loss of heterogeneity.
#' Note that this is not a symmetrical inference; multiple \code{alt} clusters can map to the same \code{ref} cluster without manifesting as a low \code{max}.
#' This implicitly assumes that an increase in resolution in \code{alt} is not problematic.
#'
#' The \code{ref.score} value for each cluster \code{ref} is formally defined as the probability of randomly picking a cell that belongs to \code{ref},
#' conditional on the event that the chosen cell belongs to the same \code{alt} cluster as a randomly chosen cell from \code{ref}.
#' This probability is equal to unity when \code{ref} is an exact superset of all \code{alt} clusters that contain its cells, corresponding to perfect 1:many nesting.
#' In contrast, if the \code{alt} clusters contain a mix of cells from different \code{ref}, this probability will be low and can be used as a diagnostic for imperfect nesting.
#'
#' @examples
#' m <- matrix(runif(10000), ncol=10)
#' clust1 <- kmeans(m,10)$cluster
#' clust2 <- kmeans(m,20)$cluster
#' nestedClusters(clust1, clust2)
#'
#' # The ref.score is 1 in cases of perfect nesting.
#' nestedClusters(clust1, clust1)$ref.score
#'
#' nest.clust <- paste0(clust1, sample(letters, length(clust1), replace=TRUE))
#' nestedClusters(clust1, nest.clust)$ref.score
#'
#' # In contrast, it is much lower when nesting is bad.
#' nestedClusters(clust1, sample(clust1))$ref.score
#' 
#' @seealso
#' \code{\link{linkClusters}}, to do this in a symmetric manner (i.e., without nesting).
#'
#' \code{\link{pairwiseRand}}, for another way of comparing two sets of clusterings.
#'
#' @export
#' @importFrom S4Vectors DataFrame
nestedClusters <- function(ref, alt) {
    tab <- table(alt, ref)
    stats <- tab/rowSums(tab)
    m <- max.col(stats, ties.method="first")
    list(
         proportions=.untable(stats),
         alt.mapping=DataFrame(row.names=rownames(stats),
                               max=stats[cbind(seq_along(m), m)],
                               which=colnames(stats)[m]),

         # To explain: 'stats' represents the probability of picking a cell
         # that belongs to 'ref', conditional on having chosen a particular
         # 'alt' cluster.  The probability of choosing that 'alt' cluster is
         # equal to the distribution of cells across 'alt' clusters for a given
         # 'ref', which is why we multiply by 'tab'. We then divide by the
         # total number of cells in each 'ref' (i.e., 'colSums(tab)') to obtain
         # an actual probability.
         ref.score=colSums(stats * tab) / colSums(tab)
    )
}

# This contains some cursory tests for nestedClusters().
# library(testthat); library(bluster); source("test-nested-clusters.R")

set.seed(1000)
m <- matrix(runif(10000), ncol=10)
clust1 <- kmeans(m,10)$cluster
clust2 <- kmeans(m,20)$cluster

test_that("nestedClusters works as expected", {
    out <- nestedClusters(clust1, clust2)
    expect_identical(rownames(out$alt.mapping), as.character(sort(unique(clust2))))
    expect_identical(out$alt.mapping$max, unname(apply(out$proportions, 1, max)))
    expect_identical(out$alt.mapping$which, as.character(apply(out$proportions, 1, which.max)))

    # Works with some more interesting names.
    out <- nestedClusters(letters[clust1], LETTERS[clust2])
    expect_identical(rownames(out$alt.mapping), LETTERS[1:20])
    expect_identical(out$alt.mapping$which, letters[apply(out$proportions, 1, which.max)])

    # Works with factors.
    out <- nestedClusters(factor(clust1, 10:1), factor(clust2, 20:1))
    expect_identical(rownames(out$alt.mapping), as.character(20:1))
    expect_identical(out$alt.mapping$which, as.character(10:1)[apply(out$proportions, 1, which.max)])
})

test_that("nestedClusters ref.score behaves correctly", {
    # The ref.score is 1 in cases of perfect nesting.
    out <- nestedClusters(clust1, clust1)$ref.score
    expect_true(all(out==1))
                                                                                
    nest.clust <- paste0(clust1, sample(letters, length(clust1), replace=TRUE))
    out <- nestedClusters(clust1, nest.clust)$ref.score
    expect_true(all(out==1))
                                                                            
    # In contrast, it is much lower when nesting is bad.
    out <- nestedClusters(clust1, sample(clust1))$ref.score
    expect_true(all(out < 0.2))
})

.non_na_scalar <- function(val) {
    length(val)==1L && !is.na(val)
}

.non_na_number <- function(val) {
    .non_na_scalar(val) && is.numeric(val)
}

.positive_number <- function(val) {
    .non_na_number(val) && val > 0
}

.check_positive_slots <- function(object, names) {
    msg <- character(0)
    for (i in names) {
        val <- slot(object, i)
        if (!is.null(val) && !.positive_number(val)) {
            msg <- c(msg, sprintf("'%s' should be positive", i))
        }
    }
    msg
}

.check_nonna_slots <- function(object, names) {
    msg <- character(0)
    for (i in names) {
        val <- slot(object, i)
        if (!is.null(val) && !.non_na_scalar(val)) {
            msg <- c(msg, sprintf("'%s' should not be NA", i))
        }
    }
    msg
}

.untable <- function(tab) {
    tab <- unclass(tab)
    dimnames(tab) <- unname(dimnames(tab))
    tab
}

#' Two step clustering with vector quantization
#'
#' For large datasets, we can perform vector quantization (e.g., with k-means clustering) to create centroids.
#' These centroids are then subjected to a slower clustering technique such as graph-based community detection.
#' The label for each cell is set to the label of the centroid to which it was assigned.
#'
#' @param first A \linkS4class{BlusterParam} object specifying a fast vector quantization technique.
#' @param second A \linkS4class{BlusterParam} object specifying the second clustering technique on the centroids.
#' @inheritParams clusterRows
#' @param BLUSPARAM A \linkS4class{KmeansParam} object.
#' @param full Logical scalar indicating whether the clustering statistics from both steps should be returned.
#' 
#' @details
#' Here, the idea is to use a fast clustering algorithm to perform vector quantization and reduce the size of the dataset,
#' followed by a slower algorithm that aggregates the centroids for easier interpretation.
#' The exact choice of the number of clusters is less relevant to the first clustering step
#' as long as not too many centroids are generated but the clusters are still sufficiently granular.
#' The second step can take more care (and computational time) summarizing the centroids into meaningful \dQuote{meta-clusters}.
#' 
#' The default choice is to use k-means for the first step, with number of clusters set to the root of the number of observations;
#' and graph-based clustering for the second step, which automatically detects a suitable number of clusters.
#' K-means also eliminates density differences in the data that can introduce variable resolution from graph-based methods.
#'
#' To modify an existing TwoStepParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#'
#' @return 
#' The \code{TwoStepParam} constructor will return a \linkS4class{TwoStepParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with a \code{clusters} factor and an \code{objects} list containing:
#' \itemize{
#' \item \code{first}, a list of objects from the first clustering step.
#' This is equal to the \code{objects} list in the output of \code{\link{clusterRows}} with the \code{first} BlusterParam.
#' \item \code{centroids}, a numeric matrix of centroids generated from the first clustering step.
#' \item \code{second}, a list of objects from the second clustering step on the centroids.
#' This is equal to the \code{objects} list in the output of \code{\link{clusterRows}} with the \code{second} BlusterParam.
#' }
#'
#' @author Aaron Lun
#' @examples
#' m <- matrix(runif(100000), ncol=10)
#' stuff <- clusterRows(m, TwoStepParam())
#' table(stuff)
#'
#' @name TwoStepParam-class
#' @aliases
#' show,TwoStepParam-method
NULL

#' @export
#' @rdname TwoStepParam-class
setClass("TwoStepParam", contains="BlusterParam", slots=c(first="BlusterParam", second="BlusterParam"))

#' @export
#' @rdname TwoStepParam-class
TwoStepParam <- function(first=KmeansParam(centers=sqrt), second=NNGraphParam()) {
    new("TwoStepParam", first=first, second=second)
}

#' @export
#' @importFrom utils capture.output
setMethod("show", "TwoStepParam", function(object) {
    callNextMethod()
    cat("first:\n")
    fout <- capture.output(show(object@first))
    cat(paste0("  ", paste(fout, collapse="\n  "), "\n"))
    cat("second:\n")
    sout <- capture.output(show(object@second))
    cat(paste0("  ", paste(sout, collapse="\n  "), "\n"))
})

#' @export
#' @rdname TwoStepParam-class
setMethod("clusterRows", c("ANY", "TwoStepParam"), function(x, BLUSPARAM, full=FALSE) {
    first <- clusterRows(x, BLUSPARAM@first, full=TRUE)

    fclust <- first$clusters
    centroids <- rowsum(x, fclust)
    rn <- rownames(centroids)
    centroids <- centroids/as.integer(table(fclust)[rn])

    second <- clusterRows(centroids, BLUSPARAM@second, full=TRUE)
    clusters <- second$clusters[match(levels(fclust), rn)[fclust]]
    clusters <- factor(clusters)

    if (!full) {
        clusters
    } else {
        list(clusters=clusters, objects=list(centroids=centroids, first=first$objects, second=second$objects))
    }
})

# Tests the TwoStepParam class.
# library(bluster); library(testthat); source('test-twostep-param.R')

test_that("TwoStepParam constructor and utilities work correctly", {
    X <- TwoStepParam()
    expect_output(show(X), "TwoStepParam")

    expect_s4_class(X[["first"]], "KmeansParam")
    X[["first"]] <- HclustParam()
    expect_s4_class(X[["first"]], "HclustParam")
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(10000), ncol=10)
    out <- clusterRows(m, TwoStepParam())
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))

    out2 <- clusterRows(m, TwoStepParam(first=KmeansParam(100)))
    expect_identical(length(out2), nrow(m))
    expect_false(identical(out, out2))

    out <- clusterRows(m, TwoStepParam(second=HclustParam()))
    expect_identical(length(out), nrow(m))
    expect_false(identical(out, out2))

    full <- clusterRows(m, TwoStepParam(), full=TRUE)
    expect_identical(length(full$cluster), nrow(m))
    expect_type(full$objects$first, "list")
    expect_type(full$objects$centroids, "double")
    expect_type(full$objects$second, "list")
})

#' Clustering with self-organizing maps
#'
#' Use the self-organizing map implementation in the \pkg{kohonen} package to cluster observations into the specified number of nodes.
#' Note that this requires the installation of the \pkg{kohonen} package.
#'
#' @param centers An integer scalar specifying the number of centers.
#' Alternatively, a function that takes the number of observations and returns the number of centers.
#' @inheritParams clusterRows
#' @param dim.ratio A positive numeric scalar in specifying how \code{centers} should be distributed between the \code{x} and \code{y} dimensions.
#' Defaults to equal distribution, i.e., both dimensions will be of length equal to the square root of \code{centers}.
#' Values above 1 will distribute more nodes to \code{x} while values below 1 will distribute mode nodes to \code{y}.
#' @param topo,neighbourhood.fct,toroidal
#' Further arguments to pass to the \code{\link[kohonen]{somgrid}} function in the \pkg{kohonen} package.
#' @param rlen,alpha,radius,dist.fct
#' Further arguments to pass to the \code{\link[kohonen]{som}} function in the \pkg{kohonen} package.
#' @param BLUSPARAM A \linkS4class{SomParam} object.
#' @param full Logical scalar indicating whether the full SOM statistics should be returned.
#' 
#' @details
#' This class usually requires the user to specify the number of clusters beforehand.
#' However, we can also allow the number of clusters to vary as a function of the number of observations.
#' The latter is occasionally useful, e.g., to allow the clustering to automatically become more granular for large datasets.
#'
#' Note that the final number of clusters may not be exactly equal to \code{centers}, depending on how \code{dim.ratio} is specified.
#' For example, if \code{centers} is a perfect square and \code{dim.ratio=1}, we will get exactly the requested number of points.
#'
#' To modify an existing SomParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#'
#' For \code{radius}, a value of \code{NULL} means that the default argument in the \code{\link[kohonen]{som}} function signature is used.
#' This is are data-dependent and so cannot be specified during construction of the SomParam object.
#'
#' For \code{dist.fct}, users can specify any string that can be used in the \code{dist.fcts} arguments in \code{\link[kohonen]{som}}.
#' In practice, the only real alternative is \code{"manhattan"}.
#'
#' @return
#' The \code{SomParam} constructor will return a \linkS4class{SomParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
#' (a list containing \code{som}, the direct output of \code{\link[kohonen]{som}}).
#'
#' @author Aaron Lun
#' @examples
#' clusterRows(iris[,1:4], SomParam(centers=16))
#' clusterRows(iris[,1:4], SomParam(centers=12, dim.ratio=3/4))
#'
#' @seealso
#' \code{\link[kohonen]{som}} from the \pkg{kohonen} package, which does all of the heavy lifting.
#'
#' \linkS4class{FixedNumberParam}, the parent of the SomParam class.
#' @name SomParam-class
#' @docType class
#' @aliases 
#' show,SomParam-method
NULL

#' @export
#' @rdname SomParam-class
setClass("SomParam", contains="FixedNumberParam", slots=c(
    dim.ratio="numeric", 
    topo="character",
    neighbourhood.fct="character",
    toroidal="logical",
    rlen="integer", 
    alpha="numeric",
    radius="numeric_OR_NULL",
    dist.fct="character"
))

setValidity2("SomParam", function(object) {
    msg <- character(0)

    msg <- c(msg, .check_positive_slots(object, c("dim.ratio", "rlen"))) 

    val <- object@alpha
    if (length(val)!=2 || !all(is.finite(val)) || !all(val > 0)) {
        msg <- c(msg, "'alpha' must be a positive numeric vector of length 2")
    }

    val <- object@radius
    if (!is.null(val)) {
        if (length(val)!=2 || !all(is.finite(val)) || !all(val > 0)) {
            msg <- c(msg, "'radius' must be a positive numeric vector of length 2")
        }
    }

    msg <- c(msg, .check_nonna_slots(object, c("dist.fct", "topo", "neighbourhood.fct", "toroidal")))

    if (length(msg)) {
        return(msg)
    }
    TRUE
})

#' @export
setMethod("show", "SomParam", function(object) {
    callNextMethod()

    for (i in c("dim.ratio", "topo", "neighbourhood.fct", "toroidal", "rlen")) {
        cat(sprintf("%s: %s\n", i, slot(object, i)))
    }

    cat(sprintf("alpha: %s\n", paste(object@alpha, collapse=" ")))
    cat(sprintf("radius: %s\n", if (is.null(object@radius)) "default" else paste(object@radius, collapse=" ")))
    cat(sprintf("dist.fct: %s\n", object@dist.fct))
})

#' @export
#' @rdname SomParam-class
SomParam <- function(centers, 
    dim.ratio = 1,
    topo = "rectangular",
    neighbourhood.fct = "bubble",
    toroidal = FALSE,
    rlen = 100,
    alpha = c(0.05, 0.01),
    radius = NULL,
    dist.fct = "sumofsquares")
{
    if (!is.function(centers)) {
        centers <- as.integer(centers)
    }
    new("SomParam", 
        centers=centers, 
        dim.ratio=dim.ratio,
        topo=topo,
        neighbourhood.fct=neighbourhood.fct,
        toroidal=toroidal,
        rlen=as.integer(rlen),
        alpha=alpha,
        radius=radius,
        dist.fct=dist.fct
    )
}

#' @export
#' @rdname SomParam-class
setMethod("clusterRows", c("ANY", "SomParam"), function(x, BLUSPARAM, full=FALSE) {
    centerx <- centers(BLUSPARAM, n=nrow(x))
    dim.ratio <- BLUSPARAM[["dim.ratio"]]
    xdim <- max(1, round(sqrt(centerx * dim.ratio)))
    ydim <- max(1, round(sqrt(centerx / dim.ratio)))

    args <- list(
        X = as.matrix(x),
        grid = kohonen::somgrid(
            xdim = xdim,
            ydim = ydim,
            topo = BLUSPARAM@topo,
            neighbourhood.fct = BLUSPARAM@neighbourhood.fct,
            toroidal = BLUSPARAM@toroidal
        ),
        rlen = BLUSPARAM@rlen,
        alpha = BLUSPARAM@alpha,
        dist.fcts = list(BLUSPARAM@dist.fct)
    )

    if (!is.null(BLUSPARAM[["radius"]])) {
        args$radius <- BLUSPARAM[["radius"]]
    }

    stats <- do.call(kohonen::som, args)
    clusters <- factor(stats$unit.classif)

    if (full) {
        list(clusters=clusters, objects=list(som=stats))
    } else {
        clusters 
    }
})

# Tests the SomParam class.
# library(bluster); library(testthat); source('test-som-param.R')

set.seed(1000)
test_that("SomParam constructor and utilities work correctly", {
    X <- SomParam(centers=10)
    expect_output(show(X), "SomParam")
    expect_output(show(X), "centers: 10")
    expect_output(show(X), "alpha: 0.05 0.01")
    expect_output(show(X), "radius: default")
    expect_output(show(X), "rlen: 10")
    expect_output(show(X), "dist.fct: sumofsquares")

    expect_identical(X[["centers"]], 10L)
    X[["centers"]] <- 2L
    expect_identical(X[["centers"]], 2L)

    X <- SomParam(centers=log)
    expect_true(is.function(X[["centers"]]))

    X[["radius"]] <- c(0, 0.1)
    expect_output(show(X), "radius: 0 0.1")

    X <- SomParam(centers=10, rlen=50)
    expect_identical(X[["rlen"]], 50L)
    X[["rlen"]] <- 100L 
    expect_identical(X[["rlen"]], 100L)
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(10000), ncol=10)

    set.seed(100)
    out <- clusterRows(m, SomParam(25))
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))
    expect_identical(nlevels(out), 25L)

    # Beahves the same with a seed.
    set.seed(100)
    out2 <- clusterRows(m, SomParam(25))
    expect_identical(out, out2)

    set.seed(200)
    out3 <- clusterRows(m, SomParam(25))
    expect_false(identical(out, out3))

    # Consistent with reference values.
    set.seed(100)
    ref <- kohonen::som(m, kohonen::somgrid(5,5))
    expect_identical(out, factor(ref$unit.classif))

    # Trying with fewer cells.
    m <- matrix(runif(1000), ncol=10)
    colnames(m) <- seq_len(ncol(m))

    set.seed(100)
    out <- clusterRows(m, SomParam(12, dim.ratio=3/4))
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))
    expect_identical(nlevels(out), 12L)

    set.seed(100)
    ref <- kohonen::som(m, kohonen::somgrid(3,4))
    expect_identical(out, factor(ref$unit.classif))
})

test_that("clusterRows responds to the functions and full=TRUE", {
    m <- matrix(runif(12960), ncol=10)

    set.seed(9999)
    out <- clusterRows(m, SomParam(sqrt))
    expect_identical(length(out), nrow(m))
    expect_equal(nlevels(out), round(sqrt(nrow(m))))

    set.seed(9999)
    full <- clusterRows(m, SomParam(sqrt), full=TRUE)
    expect_identical(out, full$cluster)
    expect_true(is.list(full$objects))
    expect_identical(names(full$objects), "som")
})

test_that("clusterRows responds to the options", {
    m <- matrix(runif(10000), ncol=10)

    set.seed(100000)
    suppressWarnings(ref<- clusterRows(m, SomParam(16)))

    set.seed(100000)
    suppressWarnings(con <- clusterRows(m, SomParam(16)))
    expect_true(identical(ref, con)) # as a control

    set.seed(100000)
    suppressWarnings(out <- clusterRows(m, SomParam(16, rlen=20, alpha=c(0.12, 0.06)))) 
    expect_false(identical(ref, out)) 

    set.seed(100000)
    ref <- kohonen::som(m, kohonen::somgrid(4,4), rlen=20, alpha=c(0.12, 0.06))
    expect_identical(out, factor(ref$unit.classif))
})

#' The BlusterParam class
#'
#' The BlusterParam class is a virtual base class controlling S4 dispatch in \code{\link{clusterRows}} and friends.
#' Concrete subclasses specify the choice of clustering algorithm,
#' while the slots of an instance of such a subclass represent the parameters for that algorithm.
#'
#' @section Available methods:
#' In the following code snippets, \code{x} is a \linkS4class{BlusterParam} object or one of its subclasses.
#' \itemize{
#' \item \code{x[[i]]} will return the value of the parameter \code{i}.
#' Refer to the documentation for each concrete subclass for more details on the available parameters.
#' \item \code{x[[i]] <- value} will set the value of the parameter \code{i} to \code{value}.
#' \item \code{show(x)} will print some information about the class instance.
#' }
#' 
#' @author Aaron Lun
#' @seealso
#' \linkS4class{HclustParam}, \linkS4class{KmeansParam} and \linkS4class{NNGraphParam} 
#' for some examples of concrete subclasses.
#'
#' @aliases
#' [[,BlusterParam-method
#' [[<-,BlusterParam-method
#' show,BlusterParam-method
#' @docType class
#' @export
setClass("BlusterParam", contains="VIRTUAL")

setClassUnion("function_OR_NULL", c("function", "NULL"))

setClassUnion("numeric_OR_NULL", c("numeric", "NULL"))

setClassUnion("integer_OR_NULL", c("integer", "NULL"))

setClassUnion("integer_OR_function", c("function", "integer"))

setClassUnion("character_OR_function", c("function", "character"))

#' The FixedNumberParam class
#'
#' The FixedNumberParam is a virtual subclass of the \linkS4class{BlusterParam} class.
#' It causes \code{\link{clusterRows}} to dispatch to clustering algorithms that rely on a pre-specified number of clusters, e.g., \linkS4class{KmeansParam}.
#' 
#' @section Available methods:
#' \code{centers(x, n=NULL)} will return the specified number of centers in a FixedNumberParam \code{x}.
#' This can be an positive integer, or a function that accepts the number of observations and returns a positive number.
#' If a function and \code{n} is supplied, the function is called on \code{n} and the result is rounded to obtain an integer.
#'
#' \code{centers(x) <- value} will replace the specified number of centers in \code{x} with an integer scalar or function \code{value}.
#' The function should accept a single argument and return a positive integer.
#' 
#' @author Aaron Lun
#'
#' @seealso
#' \linkS4class{KmeansParam}, for the archetypal example of a concrete subclass.
#'
#' @aliases
#' centers
#' centers,FixedNumberParam-method
#' centers<-
#' centers<-,FixedNumberParam-method
#' show,FixedNumberParam-method
#' 
#' @docType class
#' @export
setClass("FixedNumberParam", contains=c("BlusterParam", "VIRTUAL"), slots=c(centers="integer_OR_function"))

#' The HierarchicalParam class
#'
#' The HierarchicalParam is a virtual subclass of the \linkS4class{BlusterParam} class.
#' It causes \code{\link{clusterRows}} to dispatch to clustering algorithms that produce a dissimilarity matrix and a dendrogram.
#' 
#' @section Available slots:
#' The virtual class provides \code{metric}, the choice of distance metric.
#' This is conventionally passed to \code{\link{dist}} and defaults to a Euclidean distance in most subclasses.
#'
#' It also provides a number of slots to manage the final tree cut:
#' \itemize{
#' \item \code{cut.fun}, a function that takes a \link{hclust} object as its first argument and returns a vector of cluster assignments.
#' If \code{NULL}, the choice of function is determined from \code{cut.dynamic}.
#' \item \code{cut.dynamic}, a logical scalar indicating whether a dynamic tree cut should be performed by \code{\link[dynamicTreeCut]{cutreeDynamic}}.
#' Otherwise \code{\link{cutree}} is used.
#' Ignored if \code{cut.fun} is not \code{NULL}.
#' \item \code{cut.params}, further arguments to pass to the tree cut function specified by the previous arguments.
#' }
#' 
#' @section Return value:
#' The contract is that, when \code{full=TRUE}, the \code{objects} field of the \code{\link{clusterRows}} return value will always contain at least the following elements:
#' \itemize{
#' \item \code{dist}, a \link{dist} object containing a dissimilarity matrix, usually a distance matrix.
#' \item \code{hclust}, a \link{hclust} object containing a dendrogram.
#' }
#'
#' @seealso
#' \linkS4class{HclustParam}, for the archetypal example of a concrete subclass.
#'
#' @docType class
#' @aliases
#' .defaultScalarArguments,HierarchicalParam-method
#' show,HierarchicalParam-method
#' @export
setClass("HierarchicalParam", contains=c("BlusterParam", "VIRTUAL"), 
    slots=c(
        metric="ANY",
        cut.fun="function_OR_NULL", 
        cut.dynamic="logical", 
        cut.params="list"
    )
)

#' Clustering parameter sweeps
#'
#' Perform a sweep across combinations of parameters to obtain different clusterings from the same algorithm.
#'
#' @inheritParams clusterRows
#' @param ... Named vectors or lists specifying the parameters to sweep over.
#' @param BPPARAM A \linkS4class{BiocParallelParam} specifying how the sweep should be parallelized.
#' @param args A named list of additional arguments to use with \code{...}.
#' This is provided in case there is a name conflict with the existing arguments in this function signature.
#' 
#' @return
#' A \linkS4class{List} containing:
#' \itemize{
#' \item \code{clusters}, a \linkS4class{DataFrame} with number of rows equal to that of \code{x},
#' where each column corresponds to (and is named after) a specific combination of clustering parameters.
#' \item \code{parameters}, another DataFrame with number of rows equal to the number of columns in the previous \code{clusters} DataFrame.
#' Each row contains the specific parameter combination for each column of \code{clusters}.
#' \item If \code{full=TRUE}, \code{objects} is an additional list of length equal to the number of rows in \code{clusters}.
#' This contains the \code{objects} produced by each run.
#' }
#'
#' @details
#' This function allows users to conveniently test out a range of clustering parameters in a single call.
#' The name of each argument in \code{...} should be a legitimate argument to \code{x[[i]]},
#' and will be used to modify any existing values in \code{BLUSPARAM} to obtain a new set of parameters.
#' (For all other parameters, the existing values in \code{BLUSPARAM} are used.)
#' If multiple arguments are provided, all combinations are tested.
#'
#' We attempt to create a unique name for each column based on its parameter combination.
#' This has the format of \code{<NAME1>.<VALUE1>_<NAME2>.<VALUE2>_...} based on the parameter names and values.
#' Note that any non-atomic values are simply represented by the name of their class;
#' no attempt is made to convert these into a compact string.
#' 
#' If an entry of \code{...} is a \emph{named} list of vectors, we expand those to generate all possible combinations of values.
#' For example, if we passed:
#' \preformatted{    blah.args = list(a = 1:5, b = LETTERS[1:3])}
#' This would be equivalent to manually specifying:
#' \preformatted{    blah.args = list(list(a = 1, b = "A"), list(a = 1, b = "B"), ...)}
#' The auto-expansion mechanism allows us to conveniently test parameter combinations when those parameters are stored inside \code{x} as a list.
#' The algorithm is recursive so any internal named lists containing vectors are similarly expanded.
#' Expansion can be disabled by wrapping vectors in \code{\link{I}}, in which case they are passed verbatim.
#' No expansion is performed for non-vector arguments.
#'
#' @author Aaron Lun
#'
#' @examples
#' out <- clusterSweep(iris[,1:4], KmeansParam(10), 
#'     centers=4:10, algorithm=c("Lloyd", "Hartigan-Wong"))
#' out$clusters[,1:5]
#' out$parameters
#'
#' out <- clusterSweep(iris[,1:4], NNGraphParam(), k=c(5L, 10L, 15L, 20L),
#'     cluster.fun=c("louvain", "walktrap"))
#' out$clusters[,1:5]
#' out$parameters
#' 
#' # Combinations are automatically expanded inside named lists:
#' out <- clusterSweep(iris[,1:4], NNGraphParam(), k=c(5L, 10L, 15L, 20L), 
#'     cluster.args=list(steps=3:4))
#' out$clusters[,1:5]
#' out$parameters
#' 
#' @seealso
#' \code{\link{clusterRows}}, which manages the dispatch to specific methods based on \code{BLUSPARAM}.
#'
#' \linkS4class{BlusterParam}, which determines which algorithm is actually used.
#'
#' @export
#' @importFrom BiocParallel bplapply SerialParam
#' @importFrom S4Vectors I DataFrame List
clusterSweep <- function(x, BLUSPARAM, ..., full=FALSE, BPPARAM=SerialParam(), args=list()) {
    args <- c(list(...), args)
    if (is.null(names(args)) || anyDuplicated(names(args))) {
        stop("all arguments in '...' and 'args' must have unique names")
    }

    for (i in names(args)) {
        current <- args[[i]]
        if (is.list(current) && !is.null(names(current))) {
            args[[i]] <- .expand_list(current)
        }
    }

    N <- lengths(args)
    all.choices <- do.call(expand.grid, lapply(N, seq_len))

    out <- bplapply(seq_len(nrow(all.choices)), FUN=.run_combination, BPPARAM=BPPARAM, 
        x=x, combinations=all.choices, parameters=args, full=full, BLUSPARAM=BLUSPARAM)

    param.list <- lapply(seq_along(all.choices), FUN=function(i) args[[i]][all.choices[,i]])
    names(param.list) <- names(args)
    param.df <- do.call(DataFrame, lapply(param.list, I))
    rownames(param.df) <- names(out) <- .create_combination_names(param.list)

    output <- List()
    if (full) {
        output$clusters <- DataFrame(lapply(out, function(x) x$clusters), check.names=FALSE)
        output$parameters <- param.df
        output$objects <- lapply(out, function(x) x$objects)
    } else {
        output$clusters <- DataFrame(out, check.names=FALSE)
        output$parameters <- param.df
    }

    output
}

.run_combination <- function(i, x, BLUSPARAM, combinations, parameters, full) {
    current <- combinations[i,]
    for (j in seq_along(parameters)) {
        curparams <- parameters[[j]]
        BLUSPARAM[[names(parameters)[j]]] <- curparams[[current[[j]]]]
    }
    clusterRows(x, BLUSPARAM=BLUSPARAM, full=full)
}

.expand_list <- function(values) {
    collected <- list(list())
    for (field in names(values)) {
        current <- values[[field]]
        if (is.list(current) && !is.null(names(current))) {
            current <- .expand_list(current)
        } else if (!is.vector(current)) {
            if (is(current, "AsIs")) {
                classes <- class(current)
                new_class <- classes[classes != "AsIs"]
                attr(new_class, "package") <- attr(classes, "package")
                class(current) <- new_class
            }
            current <- list(current)
        }

        # Expanding by these values.
        for (i in seq_along(collected)) {
            added <- vector("list", length(current))
            for (j in seq_along(current)) {
                stub <- collected[[i]]
                stub[[field]] <- current[[j]]
                added[[j]] <- stub
            }
            collected[[i]] <- added
        }
        collected <- unlist(collected, recursive=FALSE)
    }

    collected 
}

.create_combination_names <- function(parameters) {
    for (i in names(parameters)) {
        values <- parameters[[i]]
        if (!is.atomic(values)) {
            values <- rep(class(values)[1], length(values))
        }
        parameters[[i]] <- sprintf("%s.%s", i, values)
    }
    do.call(paste, c(parameters, list(sep="_")))
}

# This tests the sweeping functionality.
# library(testthat); library(bluster); source("test-sweep.R")

test_that("clusterSweep works correctly in simple cases", {
    x <- matrix(rnorm(1000), ncol=5)
    out <- clusterSweep(x, NNGraphParam(), k=c(5L, 10L, 20L), cluster.fun=c("louvain", "walktrap"))
    for (choice in seq_len(nrow(out$parameters))) {
        curk <- out$parameters$k[[choice]]
        curf <- out$parameters$cluster.fun[[choice]]
        param <- NNGraphParam(k=curk, cluster.fun=curf)
        expect_identical(out$clusters[,choice], clusterRows(x, param))
        expect_identical(colnames(out$clusters)[choice], sprintf("k.%s_cluster.fun.%s", curk, curf))
    }

    # Respects existing parameters.
    x <- matrix(rnorm(1000), ncol=5)
    out <- clusterSweep(x, NNGraphParam(k=12), cluster.fun=c("louvain", "walktrap"))
    for (choice in 1:2) {
        curf <- out$parameters$cluster.fun[[choice]]
        param <- NNGraphParam(k=12, cluster.fun=curf)
        expect_identical(out$clusters[,choice], clusterRows(x, param))
        expect_identical(colnames(out$clusters)[choice], sprintf("cluster.fun.%s", curf))
    }
})

test_that("clusterSweep reports full objects correctly", {
    x <- matrix(rnorm(1000), ncol=5)

    out <- clusterSweep(x, NNGraphParam(), cluster.fun=c("louvain", "walktrap"), full=TRUE)
    ref <- clusterSweep(x, NNGraphParam(), cluster.fun=c("louvain", "walktrap"))
    expect_identical(out[1:2], ref)

    counter <- clusterRows(x, NNGraphParam(cluster.fun="walktrap"), full=TRUE)
    expect_identical(out$objects$cluster.fun.walktrap$communities, counter$objects$communities)
})

test_that("clusterSweep handles difficult names correctly", {
    x <- matrix(rnorm(1000), ncol=5)
    out <- clusterSweep(x, NNGraphParam(cluster.fun="walktrap"), cluster.args=list(list(steps=3), list(steps=4)))
    expect_identical(colnames(out$clusters), rep("cluster.args.list", 2))
})

test_that("clusterSweep loops across list elements correctly", {
    set.seed(100)
    x <- matrix(rnorm(1000), ncol=5)
    out <- clusterSweep(x, NNGraphParam(cluster.fun="walktrap"), cluster.args=list(list(steps=3), list(steps=4)))

    # Checking that the list is actually extracted properly.
    ref <- clusterRows(x, NNGraphParam(cluster.fun="walktrap"))
    expect_identical(out$clusters[,2], ref)
    expect_false(identical(out$clusters[,1], ref)) # actually has an effect.

    alt <- clusterSweep(x, NNGraphParam(cluster.fun="walktrap"), cluster.args=list(steps=3:4))
    expect_equal(out, alt)

    expanded <- clusterSweep(x[1:20,], NNGraphParam(cluster.fun="spinglass", k=2), cluster.args=list(spins=c(20, 25, 30), start.temp=c(0.9, 1, 1.1)))
    expect_identical(nrow(expanded$parameters), 9L)
    expect_identical(expanded$parameters[1,1][[1]], list(spins=20, start.temp=0.9))
    expect_identical(expanded$parameters[9,1][[1]], list(spins=30, start.temp=1.1))
})

test_that("clusterSweep skips looping over AsIs'd list elements", {
    out <- bluster:::.expand_list(list(k=1:50))
    expect_identical(length(out), 50L)

    out <- bluster:::.expand_list(list(k=I(1:50)))
    expect_identical(out, list(list(k=1:50)))

    out <- bluster:::.expand_list(list(X=HclustParam()))
    expect_identical(out, list(list(X=HclustParam())))
})

#' K-means clustering
#'
#' Run the base \code{\link{kmeans}} function with the specified number of centers within \code{\link{clusterRows}}.
#'
#' @param centers An integer scalar specifying the number of centers.
#' Alternatively, a function that takes the number of observations and returns the number of centers.
#' @param iter.max,nstart,algorithm Further arguments to pass to \code{\link{kmeans}}.
#' Set to the \code{\link{kmeans}} defaults if not supplied.
#' @inheritParams clusterRows
#' @param BLUSPARAM A \linkS4class{KmeansParam} object.
#' @param full Logical scalar indicating whether the full k-means statistics should be returned.
#'
#' @author Aaron Lun
#'
#' @details
#' This class usually requires the user to specify the number of clusters beforehand.
#' However, we can also allow the number of clusters to vary as a function of the number of observations.
#' The latter is occasionally useful, e.g., to allow the clustering to automatically become more granular for large datasets.
#'
#' To modify an existing KmeansParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#'
#' @return 
#' The \code{KmeansParam} constructor will return a \linkS4class{KmeansParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
#' (a list containing \code{kmeans}, the direct output of \code{\link{kmeans}}).
#'
#' @examples
#' clusterRows(iris[,1:4], KmeansParam(centers=4))
#' clusterRows(iris[,1:4], KmeansParam(centers=4, algorithm="Lloyd"))
#' clusterRows(iris[,1:4], KmeansParam(centers=sqrt))
#' @seealso
#' \code{\link{kmeans}}, which actually does all the heavy lifting.
#'
#' \linkS4class{MbkmeansParam}, for a faster but more approximate version of the k-means algorithm.
#' @name KmeansParam-class
#' @docType class
#' @aliases 
#' show,KmeansParam-method
#' updateObject,KmeansParam-method
NULL

#' @export
#' @rdname KmeansParam-class
setClass("KmeansParam", contains="FixedNumberParam", slots=c(iter.max="integer", nstart="integer", algorithm="character"))

#' @export
#' @rdname KmeansParam-class
KmeansParam <- function(centers, iter.max = NULL, nstart = NULL, algorithm = NULL) {
    if (!is.function(centers)) {
        centers <- as.integer(centers)
    }

    # Filling in missing values with the defaults.
    current <- list(iter.max=iter.max, nstart=nstart, algorithm=algorithm)
    notpresent <- vapply(current, is.null, FALSE)
    if (any(notpresent)) {
        defaults <- .get_kmeans_defaults()
        current[notpresent] <- defaults[notpresent]
    }

    new("KmeansParam", centers=centers, iter.max=as.integer(current$iter.max), 
        nstart=as.integer(current$nstart), algorithm=current$algorithm)
}

.get_kmeans_defaults <- function() {
    args <- formals(kmeans)
    out <- args[c("iter.max", "nstart", "algorithm")]
    out$algorithm <- eval(out$algorithm)[1]
    out
}

#' @export
setMethod("show", "KmeansParam", function(object) {
    callNextMethod()
    cat(sprintf("iter.max: %s\n", object@iter.max))
    cat(sprintf("nstart: %s\n", object@nstart))
    cat(sprintf("algorithm: %s\n", object@algorithm))
})

#' @importFrom S4Vectors setValidity2
setValidity2("KmeansParam", function(object) {
    msg <- character(0)

    if (length(object@iter.max) > 1) {
        msg <- c(msg, "'iter.max' must be an integer scalar")
    }
    if (length(object@nstart) > 1) {
        msg <- c(msg, "'nstart' must be an integer scalar")
    }
    if (length(object@algorithm) > 1) {
        msg <- c(msg, "'algorithm' must be a string")
    }

    if (length(msg)) return(msg)
    TRUE
})

#' @export
#' @rdname KmeansParam-class
#' @importFrom stats kmeans
setMethod("clusterRows", c("ANY", "KmeansParam"), function(x, BLUSPARAM, full=FALSE) {
    centerx <- centers(BLUSPARAM, n=nrow(x))

    stats <- kmeans(as.matrix(x), centers=centerx, iter.max=BLUSPARAM@iter.max,
        nstart=BLUSPARAM@nstart, algorithm=BLUSPARAM@algorithm)
    clusters <- factor(stats$cluster)

    if (full) {
        list(clusters=clusters, objects=list(kmeans=stats))
    } else {
        clusters
    }
})

# Tests the KmeansParam class.
# library(bluster); library(testthat); source('test-kmeans-param.R')

test_that("KmeansParam constructor and utilities work correctly", {
    X <- KmeansParam(centers=10)
    expect_output(show(X), "KmeansParam")
    expect_output(show(X), "centers: 10")

    expect_identical(X[["centers"]], 10L)
    X[["centers"]] <- 2L
    expect_identical(X[["centers"]], 2L)

    X <- KmeansParam(centers=log)
    expect_true(is.function(X[["centers"]]))
    expect_identical(centers(X, 20), as.integer(round(log(20))))

    X <- KmeansParam(centers=10, algorithm="Lloyd")
    expect_identical(X[["algorithm"]], "Lloyd")
    X[["algorithm"]] <- "Hartigan"
    expect_identical(X[["algorithm"]], "Hartigan")
})

test_that("KmeansParam validity works correctly", {
    expect_error(KmeansParam(-1), "positive")
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(1000), ncol=10)
    out <- clusterRows(m, KmeansParam(5))
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))
    expect_identical(nlevels(out), 5L)

    set.seed(9999)
    out <- clusterRows(m, KmeansParam(sqrt))
    expect_identical(length(out), nrow(m))
    expect_equal(nlevels(out), round(sqrt(nrow(m))))

    set.seed(9999)
    full <- clusterRows(m, KmeansParam(sqrt), full=TRUE)
    expect_identical(out, full$cluster)
    expect_s3_class(full$objects$kmeans, "kmeans")

    # Responds to the options.
    set.seed(100000)
    suppressWarnings(ref <- kmeans(m, centers=20, nstart=5, iter.max=3, algorithm="Lloyd")$cluster)
    set.seed(100000)
    suppressWarnings(out <- clusterRows(m, KmeansParam(centers=20, nstart=5, iter.max=3, algorithm="Lloyd")))
    expect_identical(factor(ref), out)
})

# Tests the approxSilhouette() function.
# library(testthat); library(bluster); source('test-silhouette.R')

set.seed(80000)
test_that('approxSilhouette yields sensible output for pure clusters', {
    clusters <- rep(seq_len(10), each=100)
    y <- matrix(clusters, ncol=50, nrow=length(clusters))

    out <- approxSilhouette(y, clusters)
    expect_identical(nrow(out), nrow(y))
    expect_true(all(out$width == 1))
    expect_true(all(clusters != out$other))

    # Throwing in some jitter.
    y <- jitter(y)
    out <- approxSilhouette(y, clusters)
    expect_true(all(out$width >= 0.5))
    expect_true(all(clusters != out$other))
})

test_that('approxSilhouette yields correct output for perfectly randomized clusters', {
    clusters <- rep(1:5, each=10)
    y0 <- matrix(rnorm(100), ncol=10)
    y <- rbind(y0, y0, y0, y0, y0)    

    out <- approxSilhouette(y, clusters)
    expect_identical(nrow(out), nrow(y))
    expect_true(all(out$width == 0))
    expect_true(all(clusters != out$other))
})

test_that("approxSilhouette preserves type", {
    y <- matrix(rnorm(1000), ncol=10)
    kclust <- kmeans(y, 5)$cluster
    ref <- approxSilhouette(y, kclust)

    f <- LETTERS[kclust]
    out2 <- approxSilhouette(y, f)
    expect_identical(out2$cluster, f)
    expect_identical(out2$other, LETTERS[ref$other])
    expect_identical(out2$width, ref$width)

    f <- factor(f)
    out2 <- approxSilhouette(y, f)
    expect_identical(out2$cluster, f)
    expect_identical(out2$other, factor(LETTERS[1:5])[ref$other])
    expect_identical(out2$width, ref$width)
})

set.seed(80001)
test_that('approxSilhouette computes the right approximation', {
    y <- matrix(rnorm(1000), ncol=1)
    cout <- clusterRows(y, BLUSPARAM=KmeansParam(4))

    tY <- t(y)
    collated <- numeric(nrow(y))
    for (i in seq_len(nrow(y))) {
        d <- colSums((tY - tY[,i])^2)
        by.clust <- split(d, cout)
        ave.d <- sqrt(vapply(by.clust, mean, 0))

        m <- match(as.character(cout[i]), names(ave.d))
        rest <- ave.d[-m]
        other <- min(rest)
        collated[i] <- (other - ave.d[m])/max(other, ave.d[m])
    } 

    X <- approxSilhouette(y, cout)
    expect_equal(X$width, collated)
    expect_true(all(X$other!=cout))
})

#' Compute the RMSD per cluster
#'
#' Compute the root mean-squared deviation (RMSD) for each cluster.
#'
#' @param x Numeric matrix containing observations in rows and variables in columns.
#' @param clusters Vector containing the assigned cluster for each observation.
#' @param sum Logical scalar indicating whether to compute the sum of squares.
#'
#' @return 
#' Numeric vector of RMSD values per cluster.
#' If \code{sum=TRUE}, a numeric vector of the sum of squares per cluster is returned instead.
#'
#' @details
#' The RMSD for each cluster is a measure of its dispersion;
#' clusters with large internal heterogeneity will have high RMSDs and are good candidates for further subclustering.
#'
#' @examples
#' x <- matrix(rnorm(10000), ncol=10)
#' kout <- kmeans(x, 5)
#' clusterRMSD(x, kout$cluster)
#'
#' @author Aaron Lun
#' @export
#' @importFrom stats var
clusterRMSD <- function(x, clusters, sum=FALSE) {
    by.cluster <- split(seq_along(clusters), clusters)
    for (i in names(by.cluster)) {
        chosen <- x[by.cluster[[i]],,drop=FALSE]
        by.cluster[[i]] <- sum(apply(chosen, 2, var))
        if (sum) {
            by.cluster[[i]] <- by.cluster[[i]] * (nrow(chosen) - 1)            
        } else {
            by.cluster[[i]] <- sqrt(by.cluster[[i]])
        }
    }

    if (length(by.cluster)) {
        unlist(by.cluster)    
    } else {
        numeric(0)
    }
}

# This tests the clusterRMSD function.
# library(testthat); library(bluster); source("test-cluster-rmsd.R")

set.seed(1000)
x <- matrix(rnorm(10000), ncol=10)
kout <- kmeans(x, 5)

test_that("clusterRMSD works as expected", {
    expect_equal(kout$withinss, unname(clusterRMSD(x, kout$cluster, sum=TRUE)))
    expect_identical(names(clusterRMSD(x, kout$cluster)), as.character(1:5))

    ref <- kout$withinss / as.integer(table(kout$cluster) - 1)
    expect_equal(sqrt(ref), unname(clusterRMSD(x, kout$cluster)))

    out <- clusterRMSD(x, letters[kout$cluster])
    expect_equal(sqrt(ref), unname(out))
    expect_identical(names(out), letters[1:5])

    expect_identical(clusterRMSD(x[0,], numeric(0)), numeric(0))
})

#' Approximate silhouette width
#'
#' Given a clustering, quickly compute an approximate silhouette width for each observation.
#'
#' @inheritParams neighborPurity 
#' 
#' @details
#' The silhouette width is a general-purpose method for evaluating the separation between clusters 
#' but requires calculating the average distance between pairs of observations within or between clusters.
#' This function instead approximates the average distance with the root-mean-squared-distance, which can be computed very efficiently for large datasets.
#' The approximated averages are then used to compute the silhouette width using the usual definition.
#'
#' % In case you're wondering, here it is. Let O be the current observation.
#' % - The squared distance to the centroid is (O - Y)^2 where Y = mean(X) for the vector of observations X.
#' % - The mean of squares within X is mean[(X - Y)^2].
#' % - Their sum is:
#' %   (O - Y)^2 + mean[(X - Y)^2] = O^2 - 2OY + Y^2 + mean(X^2) - 2 mean(X) Y + Y^2
#' %                               = O^2 - 2OY + mean(X^2)
#' %                               = O^2 - 2OE(X) + mean(X^2)
#' %                               = mean[(O - X)^2]
#' %   After which we can just take the square root.
#' 
#' @return
#' A \linkS4class{DataFrame} with one row per observation in \code{x} and the columns:
#' \itemize{
#' \item \code{cluster}, the assigned cluster for each observation in \code{x}.
#' \item \code{other}, the closest cluster other than the one to which the current observation is assigned.
#' \item \code{width}, a numeric field containing the approximate silhouette width of the current cell.
#' }
#' Row names are defined as the row names of \code{x}.
#' 
#' @author Aaron Lun
#' 
#' @examples
#' m <- matrix(rnorm(10000), ncol=10)
#' clusters <- clusterRows(m, BLUSPARAM=KmeansParam(5))
#' out <- approxSilhouette(m, clusters)
#' boxplot(split(out$width, clusters))
#'
#' # Mocking up a stronger example:
#' centers <- matrix(rnorm(30), nrow=3)
#' clusters <- sample(1:3, 1000, replace=TRUE)
#'
#' y <- centers[clusters,]
#' y <- y + rnorm(length(y), sd=0.1)
#' 
#' out2 <- approxSilhouette(y, clusters)
#' boxplot(split(out2$width, clusters))
#'
#' @seealso
#' \code{silhouette} from the \pkg{cluster} package, for the exact calculation.
#'
#' \code{\link{neighborPurity}}, for another method of evaluating cluster separation.
#'
#' @export
#' @importFrom S4Vectors DataFrame
approxSilhouette <- function(x, clusters) {
    x <- as.matrix(x)
    uclust <- sort(unique(clusters))
    averaged <- list(length(uclust))
    clust.var <- numeric(length(uclust))

    for (i in seq_along(uclust)) {
        current <- uclust[i]==clusters
        xcurrent <- x[current,,drop=FALSE]
        centroid <- colMeans(xcurrent)
        averaged[[i]] <- centroid
        clust.var[i] <- sum(colMeans(sweep(xcurrent, 2, centroid)^2))
    }

    self.dist <- other.dist <- rep(Inf, nrow(x))
    other.clust <- integer(nrow(x))
    tx <- t(x)

    for (i in seq_along(uclust)) {
        D <- sqrt(colSums((tx - averaged[[i]])^2) + clust.var[i])

        is.self <- uclust[i]==clusters
        self.dist[is.self] <- D[is.self]

        is.other <- !is.self
        other.D <- D[is.other]
        better <- other.D < other.dist[is.other]
        other.dist[is.other][better] <- other.D[better]
        other.clust[is.other][better] <- i
    }

    DataFrame(
        cluster=clusters,
        other=uclust[other.clust],
        width=(other.dist - self.dist)/pmax(other.dist, self.dist),
        row.names=rownames(x)
    )
}

#' @export
setMethod(".defaultScalarArguments", "HierarchicalParam", function(x) c(metric="character"))

#' @importFrom S4Vectors setValidity2
setValidity2("HierarchicalParam", function(object) {
    if (!.non_na_scalar(slot(object, "cut.dynamic"))) {
        return("'cut.dynamic' must be a non-missing logical scalar")
    }
    TRUE
})

#' @importFrom S4Vectors coolcat
setMethod("show", "HierarchicalParam", function(object) {
    callNextMethod()
    .showScalarArguments(object)

    fun <- object@cut.fun
    if (is.null(fun)) {
        if (object@cut.dynamic) {
            cat("cut.fun: cutreeDynamic\n")
        } else {
            cat("cut.fun: cutree\n")
        }
    } else {
        cat("cut.fun: custom\n")
    }
    coolcat("cut.params(%i): %s", names(object@cut.params))
})

.cut_hierarchical <- function(hcl, dst, BLUSPARAM) {
    fun <- BLUSPARAM@cut.fun
    args <- BLUSPARAM@cut.params

    if (is.null(fun)) {
        if (!BLUSPARAM@cut.dynamic) {
            fun <- cutree
            if (is.null(args$k)) {
                if (is.null(args$h)) {
                    args$h <- max(hcl$height)/2
                }
            }

        } else {
            fun <- function(...) unname(dynamicTreeCut::cutreeDynamic(..., verbose=0))
            args$dist <- as.matrix(dst)
        }
    }

    clusters <- do.call(fun, c(list(hcl), args))
    factor(clusters)
}

#' Partitioning around medoids
#'
#' Partition observations into k-medoids as a more robust version of k-means.
#'
#' @param centers An integer scalar specifying the number of centers.
#' Alternatively, a function that takes the number of observations and returns the number of centers.
#' @param metric,medoids,nstart,stand,do.swap,variant Further arguments to pass to \code{\link{pam}}.
#' Set to the function defaults if not supplied.
#' @inheritParams clusterRows
#' @param BLUSPARAM A \linkS4class{PamParam} object.
#' @param full Logical scalar indicating whether the full PAM statistics should be returned.
#'
#' @author Aaron Lun
#'
#' @details
#' This class usually requires the user to specify the number of clusters beforehand.
#' However, we can also allow the number of clusters to vary as a function of the number of observations.
#' The latter is occasionally useful, e.g., to allow the clustering to automatically become more granular for large datasets.
#'
#' To modify an existing PamParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#'
#' @return
#' The \code{PamParam} constructor will return a \linkS4class{PamParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
#' (a list containing \code{pam}, the direct output of \code{\link{pam}}).
#'
#' @examples
#' clusterRows(iris[,1:4], PamParam(centers=4))
#' clusterRows(iris[,1:4], PamParam(centers=4, variant="faster", do.swap=FALSE))
#' clusterRows(iris[,1:4], PamParam(centers=sqrt))
#' @seealso
#' \code{\link{pam}}, which actually does all the heavy lifting.
#'
#' \linkS4class{KmeansParam}, for the more commonly used k-means algorithm.
#'
#' \linkS4class{ClaraParam}, for a scalable extension to the PAM approach.
#' @name PamParam-class
#' @docType class
#' @aliases
#' show,PamParam-method
#' .defaultScalarArguments,PamParam-method
NULL

#' @export
#' @rdname PamParam-class
setClass("PamParam", contains="FixedNumberParam", slots=c(metric="ANY", medoids="ANY", nstart="ANY", stand="ANY", do.swap="ANY", variant="ANY"))

#' @export
#' @rdname PamParam-class
PamParam <- function(centers, metric=NULL, medoids=NULL, nstart=NULL, stand=NULL, do.swap=NULL, variant=NULL) {
    if (!is.function(centers)) {
        centers <- as.integer(centers)
    }
    new("PamParam", centers=centers, metric=metric, medoids=medoids, nstart=nstart, stand=stand, do.swap=do.swap, variant=variant)
}

#' @export
setMethod("show", "PamParam", function(object) {
    callNextMethod()
    .showScalarArguments(object)
})

#' @export
setMethod(".defaultScalarArguments", "PamParam", function(x) 
    c(metric="character", medoids="character", nstart="numeric", stand="logical", do.swap="logical", variant="character"))

#' @export
#' @rdname PamParam-class
#' @importFrom cluster pam
setMethod("clusterRows", c("ANY", "PamParam"), function(x, BLUSPARAM, full=FALSE) {
    centerx <- centers(BLUSPARAM, n=nrow(x))

    args <- c(
        list(
            quote(as.matrix(x)),
            k=centerx,
            keep.data=FALSE
        ),
        .extractScalarArguments(BLUSPARAM)
    )

    stats <- do.call(pam, args)
    clusters <- factor(stats$clustering)

    if (full) {
        list(clusters=clusters, objects=list(pam=stats))
    } else {
        clusters
    }
})

# Tests the PamParam class.
# library(bluster); library(testthat); source('test-pam-param.R')

test_that("PamParam constructor and utilities work correctly", {
    X <- PamParam(centers=10)
    expect_output(show(X), "PamParam")
    expect_output(show(X), "centers: 10")

    expect_identical(X[["centers"]], 10L)
    X[["centers"]] <- 2L
    expect_identical(X[["centers"]], 2L)

    X <- PamParam(centers=log)
    expect_true(is.function(X[["centers"]]))
    expect_identical(centers(X, 20), as.integer(round(log(20))))

    X <- PamParam(centers=10, variant="faster")
    expect_identical(X[["variant"]], "faster")
    X[["variant"]] <- "original"
    expect_identical(X[["variant"]], "original")
})

test_that("PamParam validity works correctly", {
    expect_error(PamParam(-1), "positive")
    expect_error(PamParam(10, variant=LETTERS), "character scalar")
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(1000), ncol=10)
    out <- clusterRows(m, PamParam(5))
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))
    expect_identical(nlevels(out), 5L)

    set.seed(9999)
    out <- clusterRows(m, PamParam(sqrt))
    expect_identical(length(out), nrow(m))
    expect_equal(nlevels(out), round(sqrt(nrow(m))))

    set.seed(9999)
    full <- clusterRows(m, PamParam(sqrt), full=TRUE)
    expect_identical(out, full$cluster)
    expect_s3_class(full$objects$pam, "pam")

    # Responds to the options.
    set.seed(100000)
    suppressWarnings(ref <- cluster::pam(m, k=20, variant="faster", do.swap=FALSE)$clustering)
    set.seed(100000)
    suppressWarnings(out <- clusterRows(m, PamParam(centers=20, variant="faster", do.swap=FALSE)))
    expect_identical(factor(ref), out)
})

## ---- echo=FALSE--------------------------------------------------------------
knitr::opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE)
library(BiocStyle)

## -----------------------------------------------------------------------------
library(scRNAseq)
sce <- GrunPancreasData()

# Quality control to remove bad cells.
library(scuttle)
qcstats <- perCellQCMetrics(sce)
qcfilter <- quickPerCellQC(qcstats, sub.fields="altexps_ERCC_percent")
sce <- sce[,!qcfilter$discard]

# Normalization by library size.
sce <- logNormCounts(sce)

# Feature selection.
library(scran)
dec <- modelGeneVar(sce)
hvgs <- getTopHVGs(dec, n=1000)

# Dimensionality reduction.
set.seed(1000)
library(scater)
sce <- runPCA(sce, ncomponents=20, subset_row=hvgs)

# Clustering.
library(bluster)
mat <- reducedDim(sce)
clust.info <- clusterRows(mat, NNGraphParam(), full=TRUE)
clusters <- clust.info$clusters
table(clusters)

## -----------------------------------------------------------------------------
sil <- approxSilhouette(mat, clusters)
sil
boxplot(split(sil$width, clusters))

## -----------------------------------------------------------------------------
best.choice <- ifelse(sil$width > 0, clusters, sil$other)
table(Assigned=clusters, Closest=best.choice)

## -----------------------------------------------------------------------------
pure <- neighborPurity(mat, clusters)
pure
boxplot(split(pure$purity, clusters))

## -----------------------------------------------------------------------------
table(Assigned=clusters, Max=pure$maximum)

## -----------------------------------------------------------------------------
rmsd <- clusterRMSD(mat, clusters)
barplot(rmsd)

## -----------------------------------------------------------------------------
clusterRMSD(mat, clusters, sum=TRUE)

## -----------------------------------------------------------------------------
g <- clust.info$objects$graph
ratio <- pairwiseModularity(g, clusters, as.ratio=TRUE)

library(pheatmap)
pheatmap(log10(ratio+1), cluster_cols=FALSE, cluster_rows=FALSE,
    col=rev(heat.colors(100)))

## -----------------------------------------------------------------------------
cluster.gr <- igraph::graph_from_adjacency_matrix(log2(ratio+1), 
    mode="upper", weighted=TRUE, diag=FALSE)

# Increasing the weight to increase the visibility of the lines.
set.seed(1100101)
plot(cluster.gr, edge.width=igraph::E(cluster.gr)$weight*5,
    layout=igraph::layout_with_lgl)

## -----------------------------------------------------------------------------
merged <- mergeCommunities(g, clusters, number=10)
table(merged)

## -----------------------------------------------------------------------------
hclusters <- clusterRows(mat, HclustParam(cut.dynamic=TRUE))
pairwiseRand(clusters, hclusters, mode="index")

## -----------------------------------------------------------------------------
ratio <- pairwiseRand(clusters, hclusters, mode="ratio")

library(pheatmap)
pheatmap(ratio, cluster_cols=FALSE, cluster_rows=FALSE,
    col=viridis::viridis(100), breaks=seq(-1, 1, length=101))

## -----------------------------------------------------------------------------
set.seed(1001010)
ari <-bootstrapStability(mat, clusters=clusters, 
    mode="index", BLUSPARAM=NNGraphParam())
ari

## -----------------------------------------------------------------------------
set.seed(1001010)
ratio <-bootstrapStability(mat, clusters=clusters,
    mode="ratio", BLUSPARAM=NNGraphParam())

library(pheatmap)
pheatmap(ratio, cluster_cols=FALSE, cluster_rows=FALSE,
    col=viridis::viridis(100), breaks=seq(-1, 1, length=101))

## -----------------------------------------------------------------------------
combinations <- clusterSweep(mat, BLUSPARAM=SNNGraphParam(),
    k=c(5L, 10L, 15L, 20L), cluster.fun=c("walktrap", "louvain", "infomap"))

## -----------------------------------------------------------------------------
colnames(combinations$clusters)
combinations$parameters

## -----------------------------------------------------------------------------
set.seed(10)
nclusters <- 3:25
kcombos <- clusterSweep(mat, BLUSPARAM=KmeansParam(centers=5), centers=nclusters)

sil <- vapply(as.list(kcombos$clusters), function(x) mean(approxSilhouette(mat, x)$width), 0)
plot(nclusters, sil, xlab="Number of clusters", ylab="Average silhouette width")

pur <- vapply(as.list(kcombos$clusters), function(x) mean(neighborPurity(mat, x)$purity), 0)
plot(nclusters, pur, xlab="Number of clusters", ylab="Average purity")

wcss <- vapply(as.list(kcombos$clusters), function(x) sum(clusterRMSD(mat, x, sum=TRUE)), 0)
plot(nclusters, wcss, xlab="Number of clusters", ylab="Within-cluster sum of squares")

## -----------------------------------------------------------------------------
linked <- linkClusters(
    list(
        walktrap=combinations$clusters$k.10_cluster.fun.walktrap,
        louvain=combinations$clusters$k.10_cluster.fun.louvain,
        infomap=combinations$clusters$k.10_cluster.fun.infomap
    )
)
linked

## -----------------------------------------------------------------------------
meta <- igraph::cluster_walktrap(linked)
plot(linked, mark.groups=meta)

## -----------------------------------------------------------------------------
aris <- compareClusterings(combinations$clusters)
g <- igraph::graph.adjacency(aris, mode="undirected", weighted=TRUE)
meta2 <- igraph::cluster_walktrap(g)
plot(g, mark.groups=meta2)

## -----------------------------------------------------------------------------
sessionInfo()

#' Graph-based clustering
#'
#' Run community detection algorithms on a nearest-neighbor (NN) graph within \code{\link{clusterRows}}.
#'
#' @param shared Logical scalar indicating whether a shared NN graph should be constructed.
#' @param ... Further arguments to pass to \code{\link{SNNGraphParam}} (if \code{shared=TRUE}) or \code{\link{KNNGraphParam}}.
#' @inheritParams makeSNNGraph
#' @param cluster.fun Function specifying the method to use to detect communities in the NN graph.
#' The first argument of this function should be the NN graph and the return value should be a \link{communities} object.
#'
#' Alternatively, this may be a string containing the suffix of any \pkg{igraph} community detection algorithm.
#' For example, \code{cluster.fun="louvain"} will instruct \code{\link{clusterRows}} to use \code{\link{cluster_louvain}}.
#' Defaults to \code{\link{cluster_walktrap}}.
#' @param cluster.args Further arguments to pass to the chosen \code{cluster.fun}.
#' @inheritParams clusterRows
#' @param BLUSPARAM A \linkS4class{NNGraphParam} object.
#' @param full Logical scalar indicating whether the graph-based clustering objects should be returned.
#'
#' @author Aaron Lun
#'
#' @details
#' The SNNGraphParam and KNNGraphParam classes are both derived from the NNGraphParam virtual class.
#' This former will perform clustering with a shared nearest-neighbor (SNN) graph 
#' while the latter will use a simpler k-nearest neighbor (KNN) graph - see \code{?\link{makeSNNGraph}} for details.
#'
#' To modify an existing NNGraphParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#' The exception is that of \code{shared}, which is not a valid \code{i} as it is implicit in the identity of the class.
#' 
#' @return 
#' The constructors will return a \linkS4class{NNGraphParam} object with the specified parameters.
#' If \code{shared=TRUE}, this is a SNNGraphParam object; otherwise it is a KNNGraphParam object.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects};
#' the latter is a list with \code{graph} (the graph) and \code{communities} (the output of \code{cluster.fun}).
#'
#' @examples
#' clusterRows(iris[,1:4], NNGraphParam())
#' clusterRows(iris[,1:4], NNGraphParam(k=5))
#' clusterRows(iris[,1:4], NNGraphParam(cluster.fun="louvain"))
#'
#' @seealso
#' \code{\link{makeSNNGraph}} and related functions, to build the graph.
#'
#' \code{\link{cluster_walktrap}} and related functions, to perform community detection.
#'
#' @name NNGraphParam-class
#' @docType class
#' @aliases
#' show,NNGraphParam-method
NULL

#' @export
#' @rdname NNGraphParam-class
setClass("NNGraphParam", contains=c("BlusterParam", "VIRTUAL"),
    slots=c(k="integer", cluster.fun="character_OR_function", cluster.args="list",
        BNPARAM="BiocNeighborParam", BPPARAM="BiocParallelParam"))

#' @export
#' @rdname NNGraphParam-class
setClass("SNNGraphParam", contains="NNGraphParam", slot=c(type="character"))

#' @export
#' @rdname NNGraphParam-class
setClass("KNNGraphParam", contains="NNGraphParam", slot=c(directed="logical"))

#' @export
#' @rdname NNGraphParam-class
#' @importFrom BiocParallel SerialParam
#' @importFrom BiocNeighbors KmknnParam
NNGraphParam <- function(shared=TRUE, k=10, ..., BNPARAM=KmknnParam(), BPPARAM=SerialParam(), cluster.fun="walktrap", cluster.args=list()) {
    if (shared) {
        SNNGraphParam(k=k, ..., BNPARAM=BNPARAM, BPPARAM=BPPARAM, cluster.fun=cluster.fun, cluster.args=cluster.args)
    } else {
        KNNGraphParam(k=k, ..., BNPARAM=BNPARAM, BPPARAM=BPPARAM, cluster.fun=cluster.fun, cluster.args=cluster.args)
    }
}

#' @export
#' @rdname NNGraphParam-class
#' @importFrom BiocParallel SerialParam
#' @importFrom BiocNeighbors KmknnParam
SNNGraphParam <- function(k=10, type="rank", BNPARAM=KmknnParam(), BPPARAM=SerialParam(), cluster.fun="walktrap", cluster.args=list()) {
    new("SNNGraphParam", k=as.integer(k), type=type, BNPARAM=BNPARAM, BPPARAM=BPPARAM, cluster.fun=cluster.fun, cluster.args=cluster.args)
}

#' @export
#' @rdname NNGraphParam-class
#' @importFrom BiocParallel SerialParam
#' @importFrom BiocNeighbors KmknnParam
KNNGraphParam <- function(k=10, directed=FALSE, BNPARAM=KmknnParam(), BPPARAM=SerialParam(), cluster.fun="walktrap", cluster.args=list()) {
    new("KNNGraphParam", k=as.integer(k), directed=directed, BNPARAM=BNPARAM, BPPARAM=BPPARAM, cluster.fun=cluster.fun, cluster.args=cluster.args)
}

#' @importFrom S4Vectors setValidity2
setValidity2("NNGraphParam", function(object) {
    msg <- character(0)

    if (!.positive_number(object@k)) {
        msg <- c(msg, "'k' must be a positive integer scalar")
    }

    cluster.fun <- object@cluster.fun
    if (!is.function(cluster.fun) && !.non_na_scalar(cluster.fun)) {
        msg <- c(msg, "'cluster.fun' must be a non-missing string")
    }

    if (length(msg)) return(msg)
    TRUE
})

#' @importFrom S4Vectors setValidity2
setValidity2("SNNGraphParam", function(object) {
    msg <- character(0)

    if (!.non_na_scalar(object@type)) {
        msg <- c(msg, "'type' should be a non-NA string")
    }

    if (length(msg)) return(msg)
    TRUE
})

#' @importFrom S4Vectors setValidity2
setValidity2("KNNGraphParam", function(object) {
    msg <- character(0)

    if (!.non_na_scalar(object@directed)) {
        msg <- c(msg, "'directed' should be a non-NA logical scalar")
    }

    if (length(msg)) return(msg)
    TRUE
})

#' @export
#' @importFrom S4Vectors coolcat
setMethod("show", "NNGraphParam", function(object) {
    callNextMethod()
    cat(sprintf("k: %s\n", object@k))
    sub_graph_show(object)
    cat(sprintf("BNPARAM: %s\n", class(object@BNPARAM)[1]))
    cat(sprintf("BPPARAM: %s\n", class(object@BPPARAM)[1]))
    cat(sprintf("cluster.fun: %s\n", if (is.function(object@cluster.fun)) "custom" else object@cluster.fun))
    coolcat("cluster.args(%i): %s", names(object@cluster.args))
})

# Can't quite be bothered to capture the output and re-print it, so I'm just
# going to define more generics to handle the right ordering of show outputs.

setGeneric("sub_graph_show", function(object) standardGeneric("sub_graph_show"))

setMethod("sub_graph_show", "SNNGraphParam", function(object) {
    cat(sprintf("type: %s\n", object@type))
})

setMethod("sub_graph_show", "KNNGraphParam", function(object) {
    cat(sprintf("directed: %s\n", object@directed))
})

#' @export
#' @rdname NNGraphParam-class
setMethod("clusterRows", c("ANY", "SNNGraphParam"), function(x, BLUSPARAM, full=FALSE) {
    g <- makeSNNGraph(x, k=BLUSPARAM[["k"]], type=BLUSPARAM[["type"]],
        BNPARAM=BLUSPARAM[["BNPARAM"]], BPPARAM=BLUSPARAM[["BPPARAM"]]) 
    .cluster_igraph(g, BLUSPARAM, full=full)
})

#' @export
#' @rdname NNGraphParam-class
setMethod("clusterRows", c("ANY", "KNNGraphParam"), function(x, BLUSPARAM, full=FALSE) {
    g <- makeKNNGraph(x, k=BLUSPARAM[["k"]], directed=BLUSPARAM[["directed"]],
        BNPARAM=BLUSPARAM[["BNPARAM"]], BPPARAM=BLUSPARAM[["BPPARAM"]]) 
    .cluster_igraph(g, BLUSPARAM, full=full)
})

#' @importFrom utils getFromNamespace
#' @importFrom igraph membership cluster_walktrap
.cluster_igraph <- function(g, BLUSPARAM, full=TRUE) {
    cFUN <- BLUSPARAM@cluster.fun
    if (is.character(cFUN)) {
        cFUN <- getFromNamespace(paste0("cluster_", cFUN), ns="igraph")
    }

    comm <- do.call(cFUN, c(list(g), BLUSPARAM@cluster.args))
    clusters <- factor(membership(comm))

    if (full) {
        list(clusters=clusters, objects=list(graph=g, communities=comm))
    } else {
        clusters
    }
}

# Tests the NNGraphParam class.
# library(bluster); library(testthat); source('test-nngraph-param.R')

test_that("NNGraphParam constructor and utilities work correctly", {
    X <- NNGraphParam()
    expect_output(show(X), "SNNGraphParam")

    X <- NNGraphParam(shared=FALSE)
    expect_output(show(X), "KNNGraphParam")

    expect_identical(X[["cluster.fun"]], "walktrap")
    X <- NNGraphParam(cluster.fun=igraph::cluster_louvain)
    expect_true(is.function(X[["cluster.fun"]]))

    X <- NNGraphParam(k=20)
    expect_identical(X[["k"]], 20L)
    X[["k"]] <- 30L
    expect_identical(X[["k"]], 30L)
})

test_that("NNGraphParam validity works correctly", {
    expect_error(NNGraphParam(k=NA), "integer scalar")
    expect_error(NNGraphParam(type=letters), "string")
    expect_error(NNGraphParam(shared=FALSE, directed=NA), "logical scalar")
    expect_error(NNGraphParam(cluster.fun=LETTERS), "non-missing string")
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(1000), ncol=10)
    out <- clusterRows(m, NNGraphParam())
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))

    out2 <- clusterRows(m, NNGraphParam(shared=FALSE))
    expect_true(is.factor(out2))
    expect_identical(length(out2), nrow(m))
    expect_false(identical(out, out2))

    out3 <- clusterRows(m, NNGraphParam(cluster.fun="louvain"))
    expect_true(is.factor(out3))
    expect_identical(length(out3), nrow(m))
    expect_false(identical(out, out3))

    out4 <- clusterRows(m, NNGraphParam(cluster.fun=igraph::cluster_louvain))
    expect_identical(out3, out4)

    full <- clusterRows(m, NNGraphParam(), full=TRUE)
    expect_s3_class(full$objects$graph, "igraph")
    expect_s3_class(full$objects$communities, "communities")
})

#' Dirichlet multinomial mixture clustering
#' 
#' Apply the Dirichlet multinomial mixture (DMM) algorithm from the \pkg{DirichletMultinomial} package.
#' This is commonly used in microbial ecology and in analyses of metagenomic and 16S rRNA count data.
#'
#' @param x A numeric matrix-like object where rows represent observations and columns represent variables.
#' Values are expected to be counts.
#' @inheritParams clusterRows
#' @param k An integer vector indicating the number of clusters to create with the DMM algorithm.
#' A vector containing two or more values will instruct \code{\link{clusterRows}} to perform clustering on each number,
#' and choose the optimal number of clusters based on \code{type}.
#' @param type A string specifying the method to use to find the optimal number of clusters. 
#' Must be equal to \code{"laplace"}, \code{"AIC"} or \code{"BIC"}.
#' Only used when \code{k} contains multiple values.
#' @param seed Integer scalar specifying the seed to use.
#' If \code{NULL}, a random value is used on each invocation of \code{\link{clusterRows}}.
#' @param BPPARAM A \linkS4class{BiocParallelParam} object indicating how multiple clusterings should be parallelized.
#' Only relevant if \code{k} contains multiple values.
#' 
#' @author Basil Courbayre
#'
#' @references
#' Holmes I, Harris K and Quince C (2012).
#' Dirichlet multinomial mixtures: generative models for microbial metagenomics.
#' \emph{PLoS ONE}, 7(2), 1-15
#' 
#' @details
#' To modify an existing DmmParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#' 
#' @return 
#' The \code{DmmParam} constructor will return a \linkS4class{DmmParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects};
#' the latter is a list containing:
#' \itemize{
#' \item \code{dmm}, a list containing the output of \code{\link[DirichletMultinomial]{dmn}} for each value of \code{k}.
#' \item \code{best}, an integer scalar specifying the best choice of \code{k} according to the method of \code{type}.
#' \item \code{prob}, a matrix array of probabilities where each row is an observation and each column is a cluster.
#' The number of columns is set to the best number of clusters in \code{best}.
#' \item \code{seed}, an integer scalar specifying the seed used for clustering.
#' }
#'
#' @examples
#' # Mocking up a small example.
#' nfeatures <- 50
#' out1 <- matrix(rpois(20 * nfeatures, lambda = rgamma(nfeatures, 5)), ncol=nfeatures, byrow=TRUE)
#' out2 <- matrix(rpois(20 * nfeatures, lambda = rgamma(nfeatures, 5)), ncol=nfeatures, byrow=TRUE)
#' out <- rbind(out1, out2)
#' clusterRows(out, DmmParam())
#'
#' @name DmmParam-class
#' @docType class
#' @aliases 
#' show,DmmParam-method
NULL

#' @export
#' @rdname DmmParam-class
setClass("DmmParam", contains="BlusterParam", slots=c(
    k = "integer_OR_NULL", 
    type = "character", 
    seed = "integer_OR_NULL",
    BPPARAM = "BiocParallelParam"
))

.allowed.dmm.type <- c("laplace", "AIC", "BIC")

#' @export
#' @rdname DmmParam-class
#' @importFrom BiocParallel SerialParam
DmmParam <- function(k = 1:3, type = "laplace", seed = NULL, BPPARAM = SerialParam()) {
    new("DmmParam", k = as.integer(k), type=match.arg(type, .allowed.dmm.type), seed=seed, BPPARAM = BPPARAM)
}

#' @export
#' @importFrom S4Vectors coolcat
setMethod("show", "DmmParam", function(object) {
    callNextMethod()
    coolcat("k(%i): %s", as.character(object@k))
    cat(sprintf("type: %s\n", object@type))
    cat(sprintf("seed: %s\n", object@seed))
    cat(sprintf("BPPARAM: %s\n", class(object@seed)[1]))
})

setValidity2("DmmParam", function(object) {
    msg <- character(0)

    if (!is.integer(object@k) || length(object@k) < 1 || any(object@k <= 0)) {
        msg <- c(msg, "'k' must be a strictly positive integer vector")
    }

    if (!(object@type %in% .allowed.dmm.type)) {
        msg <- c(msg, "'type' must be one of \"AIC\", \"BIC\" or \"laplace\"")
    }

    if (length(msg) > 0) {
        msg
    } else {
        TRUE
    }
})

#' @export
#' @rdname DmmParam-class
#' @importFrom stats runif
#' @importFrom BiocParallel bplapply bpisup bpstart bpstop
setMethod("clusterRows", c("ANY", "DmmParam"), function(x, BLUSPARAM, full=FALSE) {
    k <- BLUSPARAM[["k"]]
    type <- BLUSPARAM[["type"]]

    # Assigning the seed here to ensure reproducibility
    # when multiple threads are used.
    seed <- BLUSPARAM[["seed"]]
    if (is.null(seed)) {
        seed <- sample(.Machine$integer.max, 1)
    }

    BPPARAM <- BLUSPARAM[["BPPARAM"]]
    if (!(bpisup(BPPARAM) || is(BPPARAM, "MulticoreParam"))) {
        bpstart(BPPARAM)
        on.exit(bpstop(BPPARAM), add = TRUE)
    }

    if (is.null(dimnames(x))) { # otherwise dmn() complains.
        dimnames(x) <- list(NULL, NULL)
    }
    dmm <- bplapply(k, DirichletMultinomial::dmn, count = x, seed = seed, BPPARAM = BPPARAM)

    # Finding optimal number of clusters if necessary
    best <- 1L
    if (length(k) > 1) {
        fit_FUN <- switch(type,
            laplace = DirichletMultinomial::laplace,
            AIC = DirichletMultinomial::AIC,
            BIC = DirichletMultinomial::BIC)
        fit <- vapply(dmm, fit_FUN, numeric(1))
        best <- which.min(fit)
    }

    prob <- DirichletMultinomial::mixture(dmm[[best]])
    colnames(prob) <- seq_len(ncol(prob))
    clusters <- colnames(prob)[max.col(prob, ties.method = "first")]
    clusters <- factor(clusters)
    names(clusters) <- rownames(x)

    if (full) {
        list(clusters=clusters, objects=list(dmm=dmm, best=k[[best]], prob=prob, seed=seed))
    } else {
        clusters
    }
})

# Tests the DmmParam class.
# library(bluster); library(testthat); source('test-dmm-param.R')

set.seed(1000)
test_that("DmmParam constructor and utilities work correctly", {
    X <- DmmParam()
    expect_output(show(X), "DmmParam")
    expect_output(show(X), "k(3): 1 2 3", fixed=TRUE)
    expect_output(show(X), "type: laplace")

    expect_identical(X[["k"]], 1:3)
    X[["k"]] <- as.integer(2:4)
    expect_identical(X[["k"]], 2:4)

    X <- DmmParam(k=3)
    expect_equal(X[["k"]], 3)
    X[["k"]] <- as.integer(2)
    expect_equal(X[["k"]], 2)

    X <- DmmParam(type="BIC")
    expect_identical(X[["type"]], "BIC")
    X[["type"]] <- "AIC"
    expect_identical(X[["type"]], "AIC")
})


test_that("DmmParam validity works correctly", {
    expect_error(DmmParam(type="unknown"), "should be one of")
    expect_error(DmmParam(k=-1), "strictly positive")
    expect_error(DmmParam(k=0:3), "strictly positive")
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(1000), ncol = 100, nrow = 10)
    m <- ceiling(m * 10)
    rownames(m) <- paste0("A",1:10)
    colnames(m) <- 1:100

    out <- clusterRows(m, DmmParam(), full = TRUE)
    expect_true(is.factor(out$clusters))
    expect_identical(length(out$clusters), nrow(m))

    # Trying with different parameters.
    set.seed(10)
    out <- clusterRows(m, DmmParam(k=4))
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))
})

test_that("clusterRows responds to full=TRUE", {
    m <- matrix(runif(1000), ncol = 100, nrow = 10)
    m <- ceiling(m * 10)
    rownames(m) <- paste0("A",1:10)
    colnames(m) <- 1:100
    out <- clusterRows(m, DmmParam())

    full <- clusterRows(m, DmmParam(), full=TRUE)
    expect_identical(out, full$clusters)
    expect_s4_class(full$objects$dmm[[1]], "DMN")
})

#' Cluster rows of a matrix
#'
#' Cluster rows of a matrix-like object with a variety of algorithms.
#'
#' @param x A numeric matrix-like object where rows represent observations and columns represent variables.
#' @param BLUSPARAM A \linkS4class{BlusterParam} object specifying the algorithm to use.
#' @param full Logical scalar indicating whether the full clustering statistics should be returned for each method.
#'
#' @return
#' By default, a factor of length equal to \code{nrow(x)} containing cluster assignments for each row of \code{x}.
#'
#' If \code{full=TRUE}, a list is returned containing \code{clusters}, a factor as described above;
#' and \code{objects}, an arbitrary object containing algorithm-specific statistics or intermediate objects.
#'
#' @details
#' This generic allows users to write agile code that can use a variety of clustering algorithms.
#' By simply changing \code{BLUSPARAM}, we can tune the clustering procedure in analysis workflows and package functions.
#'
#' @seealso
#' \linkS4class{HclustParam}, \linkS4class{KmeansParam} and \linkS4class{NNGraphParam} 
#' for some examples of values for \code{BLUSPARAM}.
#' 
#' @author Aaron Lun
#'
#' @examples
#' m <- matrix(runif(10000), ncol=10)
#'
#' clusterRows(m, KmeansParam(10L))
#' clusterRows(m, HclustParam())
#' clusterRows(m, NNGraphParam())
#' @export
setGeneric("clusterRows", function(x, BLUSPARAM, full=FALSE) standardGeneric("clusterRows"))

#' @export
setGeneric("centers", function(x, n=NULL) standardGeneric("centers"))

#' @export
setGeneric("centers<-", function(x, value) standardGeneric("centers<-"))

#' Define the default arguments
#'
#' Provide a consistent mechanism to handle specification of default arguments to the underlying clustering functions.
#'
#' @param x,object A \linkS4class{BlusterParam} object.
#'
#' @return
#' For \code{.defaultScalarArguments}, a named character vector is returned.
#' Each entry corresponds to an argument to the clustering function - the name is the argument name, and the value is the argument type.
#'
#' For \code{.extractScalarArguments}, a named list of non-default scalar arguments is returned.
#' Any arguments set to their default values are omitted from the list.
#'
#' For \code{.showScalarArguments}, the values of the arguments are printed to screen.
#' Default values are marked with \code{[default]}.
#'
#' @details
#' The idea is to simplify the derivation of new \linkS4class{BlusterParam} objects,
#' by allowing developers to indicate that the underlying function default should be used for particular arguments.
#' This avoids duplication of the default arguments in the object constructor;
#' instead, default arguments can be indicated as such by setting them to \code{NULL},
#' in which case they will not be explicitly passed to the underlying clustering function.
#'
#' @author Aaron Lun
#' @examples
#' .defaultScalarArguments(PamParam(10))
#' .extractScalarArguments(PamParam(10))
#' .extractScalarArguments(PamParam(10, variant="faster"))
#'
#' @export
#' @aliases
#' .defaultScalarArguments,BlusterParam-method
#'
#' @rdname defaultArguments
setGeneric(".defaultScalarArguments", function(x) standardGeneric(".defaultScalarArguments"))

#' Hierarchical clustering
#'
#' Run the base \code{\link{hclust}} function on a distance matrix within \code{\link{clusterRows}}.
#'
#' @param metric String specifying the distance metric to use in \code{\link{dist}}.
#' @param method String specifying the agglomeration method to use in \code{\link{hclust}}.
#' @param cut.fun Function specifying the method to use to cut the dendrogram.
#' The first argument of this function should be the output of \code{\link{hclust}},
#' and the return value should be an atomic vector specifying the cluster assignment for each observation.
#' Defaults to \code{\link{cutree}} if \code{cut.dynamic=FALSE} and \code{\link[dynamicTreeCut]{cutreeDynamic}} otherwise.
#' @param cut.dynamic Logical scalar indicating whether a dynamic tree cut should be performed using the \pkg{dynamicTreeCut} package.
#' @param cut.height,cut.number Deprecated, use \code{h} and \code{k} in \code{cut.params} instead.
#' @param cut.params Further arguments to pass to \code{cut.fun}.
#' @param ... Deprecated, more arguments to add to \code{cut.params}.
#' @inheritParams clusterRows
#' @param BLUSPARAM A \linkS4class{HclustParam} object.
#' @param full Logical scalar indicating whether the hierarchical clustering statistics should be returned.
#'
#' @author Aaron Lun
#'
#' @details
#' To modify an existing HclustParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#'
#' If \code{cut.fun=NULL}, \code{cut.dynamic=FALSE} and \code{cut.params} does not have \code{h} or \code{k},
#' \code{\link{clusterRows}} will automatically set \code{h} to half the tree height when calling \code{\link{cutree}}. 
#'
#' @return 
#' The \code{HclustParam} constructor will return a \linkS4class{HclustParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects} 
#' (a list containing \code{dist}, the distance matrix; and \code{hclust}, the output of \code{\link{hclust}}).
#'
#' @examples
#' clusterRows(iris[,1:4], HclustParam())
#' clusterRows(iris[,1:4], HclustParam(method="ward.D2"))
#'
#' @seealso
#' \code{\link{dist}}, \code{\link{hclust}} and \code{\link{cutree}}, which actually do all the heavy lifting.
#'
#' \code{\link[dynamicTreeCut]{cutreeDynamic}}, for an alternative tree cutting method to use in \code{cut.fun}.
#' @name HclustParam-class
#' @docType class
#' @aliases 
#' .defaultScalarArguments,HclustParam-method
#' updateObject,HclustParam-method
#' show,HclustParam-method
#' [[,HclustParam-method
NULL

#' @export
#' @rdname HclustParam-class
setClass("HclustParam", contains="HierarchicalParam", slots=c(method="ANY"))

#' @export
setMethod(".defaultScalarArguments", "HclustParam", function(x) c(callNextMethod(), method="character")) 

#' @export
#' @rdname HclustParam-class
HclustParam <- function(metric=NULL, method=NULL, cut.fun=NULL, cut.dynamic=FALSE, cut.height=NULL, cut.number=NULL, cut.params=list(), ...) {
    if (!is.null(cut.number)) {
        .Deprecated(old="cut.number=", new="cut.params=list(k=...)")
        cut.params$k <- cut.number
    } else if (!is.null(cut.height)) {
        .Deprecated(old="cut.height=", new="cut.params=list(h=...)")
        cut.params$h <- cut.height
    }

    extra.args <- list(...)
    if (length(extra.args)) {
        .Deprecated(old="...", new="cut.params=")
        cut.params <- c(cut.params, extra.args)
    }

    new("HclustParam", metric=metric, method=method, cut.fun=cut.fun, cut.dynamic=cut.dynamic, cut.params=cut.params)
}

#' @export
setMethod("[[", "HclustParam", function(x, i) {
    x <- updateObject(x)
    callNextMethod()
})

#' @export
setMethod("show", "HclustParam", function(object) {
    object <- updateObject(object)
    callNextMethod()
})

#' @export
#' @rdname HclustParam-class
#' @importFrom stats dist hclust cutree
setMethod("clusterRows", c("ANY", "HclustParam"), function(x, BLUSPARAM, full=FALSE) {
    dargs <- list(quote(as.matrix(x)))
    if (!is.null(BLUSPARAM@metric)) {
        dargs$metric <- BLUSPARAM@metric
    }
    dst <- do.call(dist, dargs)

    hargs <- list(quote(dst))
    if (!is.null(BLUSPARAM@method)) {
        hargs$method <- BLUSPARAM@method
    }
    hcl <- do.call(hclust, hargs)

    # Moving arguments to their new homes.
    BLUSPARAM <- updateObject(BLUSPARAM)

    clusters <- .cut_hierarchical(hcl, dst, BLUSPARAM)
    if (full) {
        list(clusters=clusters, objects=list(dist=dst, hclust=hcl))
    } else {
        clusters
    }
})

#' @export
#' @importFrom S4Vectors updateObject
setMethod("updateObject", "HclustParam", function(object, ..., verbose=FALSE) {
    has.k <- .hasSlot(object, "cut.number")
    has.h <- .hasSlot(object, "cut.height")
    
    if (has.k || has.h) {
        if (has.k && !is.null(object@cut.number)) {
            object@cut.params$k <- object@cut.number
            if (verbose) {
                message("[updateObject] moving 'cut.number' to 'cut.params$k'")
            }
        } else if (!is.null(object@cut.height)) {
            object@cut.params$h <- object@cut.height
            if (verbose) {
                message("[updateObject] moving 'cut.height' to 'cut.params$h'")
            }
        }
    }

    object
})

#' @export
setMethod("[[", "BlusterParam", function(x, i) {
    slot(x, i)
})

#' @export
setReplaceMethod("[[", "BlusterParam", function(x, i, j, ..., value) {
    slot(x, i) <- value
    x
})

#' @export
setMethod("show", "BlusterParam", function(object) {
    cat(paste0("class: ", class(object)[1], "\n"))
})

setValidity2("BlusterParam", function(object) {
    what <- .defaultScalarArguments(object)
    for (x in names(what)) {
        val <- slot(object, x)
        if (!is.null(val)) {
            if (length(val)!=1 || !is(val, what[[x]])) {
                return(sprintf("'%s' should be NULL or a %s scalar", x, what[[x]]))
            }
        }
    }
})

#' @export
setMethod(".defaultScalarArguments", "BlusterParam", function(x) character(0))

#' @export
#' @rdname defaultArguments
.showScalarArguments <- function(object) {
    what <- .defaultScalarArguments(object)
    for (x in names(what)) {
        val <- slot(object, x)
        if (is.null(val)) {
            val <- "[default]"
        }
        cat(sprintf("%s: %s\n", x, val))
    }
}

#' @export
#' @rdname defaultArguments
.extractScalarArguments <- function(x) {
    args <- list()
    what <- .defaultScalarArguments(x)
    for (i in names(what)) {
        val <- slot(x, i)
        if (!is.null(val)) {
            args[[i]] <- val
        }
    }
    args

}

#' Density-based clustering with DBSCAN
#'
#' Perform density-based clustering with a fast re-implementation of the DBSCAN algorithm.
#'
#' @inheritParams clusterRows
#' @param eps Numeric scalar specifying the distance to use to define neighborhoods.
#' If \code{NULL}, this is determined from \code{min.pts} and \code{core.prop}.
#' @param min.pts Integer scalar specifying the minimum number of neighboring observations required for an observation to be a core point.
#' @param core.prop Numeric scalar specifying the proportion of observations to treat as core points.
#' This is only used when \code{eps=NULL}, see Details.
#' @param chunk.size Integer scalar specifying the number of points to process per chunk.
#' @param BNPARAM A \linkS4class{BiocNeighborParam} object specifying the algorithm to use for the neighbor searches.
#' This should be able to support both nearest-neighbor and range queries.
#' @param BPPARAM A \linkS4class{BiocParallelParam} object specifying the parallelization to use for the neighbor searches.
#' @param full Logical scalar indicating whether additional statistics should be returned.
#'
#' @details
#' DBSCAN operates by identifying core points, i.e., observations with at least \code{min.pts} neighbors within a distance of \code{eps}.
#' It then identifies which core points are neighbors of each other, forming components of connected core points.
#' All non-core points are then connected to the closest core point within \code{eps}.
#' All groups of points that are connected in this manner are considered to be part of the same cluster.
#' Any unconnected non-core points are treated as noise and reported as \code{NA}.
#'
#' As a suitable value of \code{eps} may not be known beforehand, we can automatically determine it from the data.
#' For all observations, we compute the distance to the \eqn{k}th neighbor where \eqn{k} is defined as \code{round(min.pts * core.prop)}.
#' We then define \code{eps} as the \code{core.prop} quantile of the distances across all observations.
#' The default of \code{core.prop=0.5} means that around half of the observations will be treated as core points.
#'
#' Larger values of \code{eps} will generally result in fewer observations classified as noise, as they are more likely to connect to a core point.
#' It may also promote agglomeration of existing clusters into larger entities if they are connected by regions of (relatively) low density.
#' Conversely, larger values of \code{min.pts} will generally increase the number of noise points and may fragment larger clusters into subclusters.
#' 
#' To modify an existing DbscanParam object \code{x},
#' users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
#' 
#' @return
#' The \code{DbscanParam} constructor will return a \linkS4class{DbscanParam} object with the specified parameters.
#'
#' The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
#' Note that this may contain \code{NA} values corresponding to noise points.
#' If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects} 
#' (a list containing the \code{eps} and \code{min.pts} used in the analysis).
#' 
#' @examples
#' clusterRows(iris[,1:4], DbscanParam())
#' clusterRows(iris[,1:4], DbscanParam(core.prop=0.8))
#'
#' @references
#' Ester M et al. (1996).
#' A density-based algorithm for discovering clusters in large spatial databases with noise.
#' \emph{Proceedings of the Second International Conference on Knowledge Discovery and Data Mining}, 226-231.
#' 
#' @author Aaron Lun
#' @name DbscanParam-class
#' @docType class
#' @aliases
#' show,DbscanParam-method
NULL

#' @export
setClass("DbscanParam", contains="BlusterParam", slots=c(eps="numeric_OR_NULL", min.pts="integer", core.prop="numeric", 
    chunk.size="integer", BNPARAM="BiocNeighborParam", BPPARAM="BiocParallelParam"))

#' @export
#' @rdname DbscanParam-class
DbscanParam <- function(eps=NULL, min.pts=5, core.prop=0.5, chunk.size=1000, BNPARAM=KmknnParam(), BPPARAM=SerialParam()) {
    new("DbscanParam", eps=eps, min.pts=as.integer(min.pts), 
        core.prop=core.prop, chunk.size=as.integer(chunk.size),
        BNPARAM=BNPARAM, BPPARAM=BPPARAM)
}

#' @export
setMethod("show", "DbscanParam", function(object) {
    callNextMethod()
    cat(sprintf("eps: %s\n", if (is.null(object[["eps"]])) "default" else object[["eps"]]))
    for (i in c("min.pts", "core.prop", "chunk.size")) {
        cat(sprintf("%s: %s\n", i, object[[i]]))
    }
    for (i in c("BNPARAM", "BPPARAM")) {
        cat(sprintf("%s: %s\n", i, class(object[[i]])[1]))
    }
})

setValidity2("DbscanParam", function(object) {
    msg <- character(0)

    if (!is.null(object[["eps"]])) {
        msg <- c(msg, .check_positive_slots(object, "eps"))
    }

    msg <- c(msg, .check_positive_slots(object, c("min.pts", "core.prop", "chunk.size")))

    if (object[["core.prop"]] > 1) {
        msg <- c(msg, "'core.prop' should not be greater than 1")
    }

    if (length(msg)) {
        return(msg)
    }
    TRUE
})

#' @export
#' @rdname DbscanParam-class
setMethod("clusterRows", c("ANY", "DbscanParam"), function(x, BLUSPARAM, full=FALSE) {
    out <- .DBSCAN(x, 
        eps=BLUSPARAM[["eps"]],
        min.pts=BLUSPARAM[["min.pts"]],
        core.prop=BLUSPARAM[["core.prop"]],
        chunk.size=BLUSPARAM[["chunk.size"]],
        BNPARAM=BLUSPARAM[["BNPARAM"]],
        BPPARAM=BLUSPARAM[["BPPARAM"]]
    )

    clusters <- out$clusters
    clusters[clusters==0] <- NA_integer_
    clusters <- factor(clusters)

    if (full) {
        out$clusters <- NULL
        list(clusters=clusters, objects=out)
    } else {
        clusters
    }
})

#' @importFrom BiocNeighbors buildIndex findNeighbors findKNN KmknnParam queryKNN
#' @importFrom BiocParallel SerialParam bpstart bpstop bpisup
#' @importFrom stats quantile
#' @importFrom igraph make_graph components
#' @importFrom utils head
.DBSCAN <- function(x, eps=NULL, min.pts=5, core.prop=0.5, chunk.size=1000, BNPARAM=KmknnParam(), BPPARAM=SerialParam()) {
    if (!bpisup(BPPARAM) && !is(BPPARAM, "MulticoreParam")) {
        bpstart(BPPARAM)
        on.exit(bpstop(BPPARAM))
    }

    # Finding all core points, using a quick NN algorithm.
    all.dist <- findKNN(X=x, k=min.pts, last=1, get.index=FALSE, BNPARAM=BNPARAM, BPPARAM=BPPARAM)$distance
    if (is.null(eps)) {
        eps <- quantile(all.dist, core.prop)
    } 
    is.core <- all.dist <= eps

    # Looping through and finding core-core neighbors in chunks to keep memory usage low.
    n.core <- sum(is.core)
    core.idx <- buildIndex(x[is.core,,drop=FALSE], BNPARAM=BNPARAM)
    remaining <- seq_len(n.core)
    to <- vector("list", n.core)

    while (length(remaining)) {
        chosen <- head(remaining, chunk.size)
        neighbors <- findNeighbors(BNINDEX=core.idx, subset=chosen, threshold=eps, get.distance=FALSE, BPPARAM=BPPARAM)$index
        to[chosen] <- neighbors
        remaining <- setdiff(remaining, unlist(neighbors)) 
    }

    # Identifying core-core components.
    from <- rep(seq_along(to), lengths(to))
    to <- unlist(to)
    g <- make_graph(rbind(from, to), n=n.core, directed=FALSE)
    core.clusters <- components(g)$membership

    # Finding closest core point (within range) for all non-core points.
    clusters <- integer(nrow(x))
    clusters[is.core] <- core.clusters

    if (n.core) {
        closest.core <- queryKNN(BNINDEX=core.idx, query=x[!is.core,,drop=FALSE], BPPARAM=BPPARAM, k=1)
        in.range <- closest.core$distance <= eps
        clusters[which(!is.core)[in.range]] <- core.clusters[closest.core$index[in.range]]
    }

    list(clusters=clusters, eps=eps, min.pts=min.pts)
}

# Tests the DbscanParam class.
# library(bluster); library(testthat); source('test-dbscan-param.R')

set.seed(1000)
test_that("DbscanParam constructor and utilities work correctly", {
    X <- DbscanParam()
    expect_output(show(X), "DbscanParam")
    expect_output(show(X), "eps: default")
    expect_output(show(X), "min.pts: 5")

    expect_identical(X[["eps"]], NULL)
    X[["eps"]] <- 5 
    expect_identical(X[["eps"]], 5)

    X <- DbscanParam(min.pts=10)
    expect_identical(X[["min.pts"]], 10L)
    X[["min.pts"]] <- 2L
    expect_identical(X[["min.pts"]], 2L)
})

set.seed(10)
raw <- matrix(rnorm(100), ncol=10)
m <- matrix(raw[sample(nrow(raw), 1000, replace=TRUE),], ncol=10)
m <- m + rnorm(length(m), sd=0.1)

# Creating an alternative implementation with less subsetting all over the place.
REF <- function(m, eps, min.pts) {
    # Identifying the core points.
    neighbors <- BiocNeighbors::findNeighbors(m, threshold=eps, get.distance=FALSE)$index
    is.core <- lengths(neighbors) > min.pts # >, not >=, as self is included in range.
    core.id <- which(is.core)
    core.neighbors <- neighbors[is.core]

    # Constructing the core components.
    from <- rep(core.id, lengths(core.neighbors))
    to <- unlist(core.neighbors)
    keep <- to %in% from
    g <- igraph::make_graph(rbind(as.character(from[keep]), as.character(to[keep])), directed=FALSE)

    comp <- igraph::components(g)$membership
    clusters <- integer(nrow(m))
    clusters[as.integer(names(comp))] <- unname(comp)

    # Matching all points to their closest core point.
    closest.core <- BiocNeighbors::queryKNN(X=m[is.core,,drop=FALSE], query=m, k=1)
    reassign <- closest.core$distance <= eps
    clusters[reassign] <- comp[as.character(core.id)][closest.core$index[reassign]]

    factor(ifelse(clusters == 0, NA_integer_, clusters))
}

test_that("clusterRows works correctly", {
    out <- clusterRows(m, DbscanParam())
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))
    expect_identical(nlevels(out), 10L)

    # Trying with different parameters.
    out2 <- clusterRows(m, DbscanParam(core.prop=0.8))
    expect_true(is.factor(out2))
    expect_identical(length(out2), nrow(m))
    expect_true(sum(is.na(out2)) < sum(is.na(out)))

    # Comparing to another implementation.
    out3 <- clusterRows(m, DbscanParam(eps=0.3, min.pts=5))
    expect_identical(nlevels(out3), 10L)
    ref <- REF(m, eps=0.3, min.pts=5)
    expect_identical(out3, ref)

    out4 <- clusterRows(m, DbscanParam(eps=0.5, min.pts=10))
    expect_identical(nlevels(out4), 10L)
    ref <- REF(m, eps=0.5, min.pts=10)
    expect_identical(out4, ref)

    out5 <- clusterRows(m, DbscanParam(eps=0.25, min.pts=10))
    expect_true(nlevels(out5) > 5)
    ref <- REF(m, eps=0.25, min.pts=10)
    expect_identical(out5, ref)

    # Edge case behaves well.
    none <- clusterRows(m, DbscanParam(eps=0.001))
    expect_true(all(is.na(none)))
})

test_that("clusterRows responds to full=TRUE", {
    out <- clusterRows(m, DbscanParam())
    full <- clusterRows(m, DbscanParam(), full=TRUE)
    expect_identical(out, full$cluster)
    expect_type(full$objects$min.pts, "integer")
    expect_type(full$objects$eps, "double")
})

# Tests the AgnesParam class.
# library(bluster); library(testthat); source('test-agnes-param.R')

test_that("AgnesParam constructor and utilities work correctly", {
    X <- AgnesParam()
    expect_output(show(X), "AgnesParam")

    expect_identical(X[["method"]], NULL)
    X[["method"]] <- "average"
    expect_identical(X[["method"]], "average")

    X <- AgnesParam(cut.params=list(method="average"))
    expect_identical(X[["cut.params"]], list(method="average"))
    X[["cut.params"]] <- list(whee=2)
    expect_identical(X[["cut.params"]], list(whee=2))

    # other show methods
    expect_output(show(AgnesParam(method="average")), "average")
    expect_output(show(AgnesParam(cut.dynamic=TRUE)), "cutreeDynamic")
    expect_output(show(AgnesParam(cut.fun=identity)), "custom")
})

test_that("AgnesParam validity works correctly", {
    expect_error(AgnesParam(cut.dynamic=NA), "non-missing")
    expect_error(AgnesParam(method=1), "character")
})

test_that("clusterRows works correctly", {
    m <- matrix(runif(1000), ncol=10)
    out <- clusterRows(m, AgnesParam())
    expect_true(is.factor(out))
    expect_identical(length(out), nrow(m))

    out2 <- clusterRows(m, AgnesParam(cut.params=list(h=2)))
    expect_identical(length(out2), nrow(m))
    expect_false(identical(out, out2))

    out <- clusterRows(m, AgnesParam(cut.params=list(k=5)))
    expect_identical(length(out), nrow(m))
    expect_identical(nlevels(out), 5L)

    ref <- clusterRows(m, AgnesParam(cut.params=list(k=3)))
    out <- clusterRows(m, AgnesParam(cut.fun=function(x) cutree(x, k=3)))
    expect_identical(out, ref)

    full <- clusterRows(m, AgnesParam(), full=TRUE)
    expect_s3_class(full$objects$dist, "dist")
    expect_s3_class(full$objects$hclust, "hclust")

    # Default cut works as expected.
    maxh <- max(full$objects$hclust$height)
    check <- clusterRows(m, AgnesParam(cut.params=list(h=maxh/2)))
    expect_identical(full$clusters, check)
})

test_that("clusterRows works with the dynamic tree cut", {
    m <- matrix(runif(1000), ncol=10)
    out <- clusterRows(m, AgnesParam(cut.dynamic=TRUE))
    expect_true(is.factor(out))
    expect_identical(names(out), NULL)
    expect_identical(length(out), nrow(m))
})

---
title: Flexible clustering for Bioconductor 
author: 
- name: Aaron Lun
  email: infinite.monkeys.with.keyboards@gmail.com
output: BiocStyle::html_document
package: bluster
vignette: >
  %\VignetteIndexEntry{1. Clustering algorithms}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}    
---

```{r, echo=FALSE}
knitr::opts_chunk$set(error=FALSE, message=FALSE, warnings=FALSE)
library(BiocStyle)
```

# Introduction

The `r Biocpkg("bluster")` package provides a flexible and extensible framework for clustering in Bioconductor packages/workflows.
At its core is the `clusterRows()` generic that controls dispatch to different clustering algorithms. 
We will demonstrate on some single-cell RNA sequencing data from the `r Biocpkg("scRNAseq")` package;
our aim is to cluster cells into cell populations based on their PC coordinates.

```{r}
library(scRNAseq)
sce <- ZeiselBrainData()

# Trusting the authors' quality control, and going straight to normalization.
library(scuttle)
sce <- logNormCounts(sce)

# Feature selection based on highly variable genes.
library(scran)
dec <- modelGeneVar(sce)
hvgs <- getTopHVGs(dec, n=1000)

# Dimensionality reduction for work (PCA) and pleasure (t-SNE).
set.seed(1000)
library(scater)
sce <- runPCA(sce, ncomponents=20, subset_row=hvgs)
sce <- runUMAP(sce, dimred="PCA")

mat <- reducedDim(sce, "PCA")
dim(mat)
```

# Based on distance matrices

## Hierarchical clustering

Our first algorithm is good old hierarchical clustering, as implemented using `hclust()` from the _stats_ package.
This automatically sets the cut height to half the dendrogram height.

```{r}
library(bluster)
hclust.out <- clusterRows(mat, HclustParam())
plotUMAP(sce, colour_by=I(hclust.out))
```

Advanced users can achieve greater control of the procedure by passing more parameters to the `HclustParam()` constructor.
Here, we use Ward's criterion for the agglomeration with a dynamic tree cut from the `r CRANpkg("dynamicTreeCut")` package.

```{r}
hp2 <- HclustParam(method="ward.D2", cut.dynamic=TRUE)
hp2
hclust.out <- clusterRows(mat, hp2)
plotUMAP(sce, colour_by=I(hclust.out))
```

## Affinity propagation

Another option is to use affinity propagation, as implemented using the `r CRANpkg("apcluster")` package.
Here, messages are passed between observations to decide on a set of exemplars, each of which form the center of a cluster.

This is not particularly fast as it involves the calculation of a square similarity matrix between all pairs of observations.
So, we'll speed it up by taking analyzing a subset of the data:

```{r}
set.seed(1000)
sub <- sce[,sample(ncol(sce), 200)]
ap.out <- clusterRows(reducedDim(sub), AffinityParam())
plotUMAP(sub, colour_by=I(ap.out))
```

The `q` parameter is probably the most important and determines the resolution of the clustering.
This can be set to any value below 1, with smaller (possibly negative) values corresponding to coarser clusters.

```{r}
set.seed(1000)
ap.out <- clusterRows(reducedDim(sub), AffinityParam(q=-2))
plotUMAP(sub, colour_by=I(ap.out))
```

# With a fixed number of clusters

## $k$-means clustering

A classic algorithm is $k$-means clustering, as implemented using the `kmeans()` function.
This requires us to pass in the number of clusters, either as a number: 

```{r}
set.seed(100)
kmeans.out <- clusterRows(mat, KmeansParam(10))
plotUMAP(sce, colour_by=I(kmeans.out))
```

Or as a function of the number of observations, which is useful for vector quantization purposes:

```{r}
kp <- KmeansParam(sqrt)
kp
set.seed(100)
kmeans.out <- clusterRows(mat, kp)
plotUMAP(sce, colour_by=I(kmeans.out))
```

A variant of this approach is mini-batch $k$-means, as implemented in the `r Biocpkg("mbkmeans")` package.
This uses mini-batching to approximate the full $k$-means algorithm for greater speed.

```{r}
set.seed(100)
mbkmeans.out <- clusterRows(mat, MbkmeansParam(20))
plotUMAP(sce, colour_by=I(mbkmeans.out))
```

## Self-organizing maps

We can also use self-organizing maps (SOMs) from the `r CRANpkg("kohonen")` package.
This allocates observations to nodes of a simple neural network; each node is then treated as a cluster.

```{r}
set.seed(1000)
som.out <- clusterRows(mat, SomParam(20))
plotUMAP(sce, colour_by=I(som.out))
```

The key feature of SOMs is that they apply additional topological constraints on the relative positions of the nodes.
This allows us to naturally determine the relationships between clusters based on the proximity of the nodes.

```{r, fig.wide=TRUE}
set.seed(1000)
som.out <- clusterRows(mat, SomParam(100), full=TRUE)

par(mfrow=c(1,2))
plot(som.out$objects$som, "counts")
grid <- som.out$objects$som$grid$pts
text(grid[,1], grid[,2], seq_len(nrow(grid)))
```

# Graph-based clustering

We can build shared or direct nearest neighbor graphs and perform community detection with `r CRANpkg("igraph")`.
Here, the number of neighbors `k` controls the resolution of the clusters.

```{r}
set.seed(101) # just in case there are ties.
graph.out <- clusterRows(mat, NNGraphParam(k=10))
plotUMAP(sce, colour_by=I(graph.out))
```

It is again straightforward to tune the procedure by passing more arguments such as the community detection algorithm to use.

```{r}
set.seed(101) # just in case there are ties.
np <- NNGraphParam(k=20, cluster.fun="louvain")
np
graph.out <- clusterRows(mat, np)
plotUMAP(sce, colour_by=I(graph.out))
```

# Density-based clustering

We also implement a version of the DBSCAN algorithm for density-based clustering.
This focuses on identifying masses of continuous density.

```{r}
dbscan.out <- clusterRows(mat, DbscanParam())
plotUMAP(sce, colour_by=I(dbscan.out))
```

Unlike the other methods, it will consider certain points to be noise and discard them from the output.

```{r}
summary(is.na(dbscan.out))
```

The resolution of the clustering can be modified by tinkering with the `core.prop`.
Smaller values will generally result in smaller clusters as well as more noise points.

```{r}
dbscan.out <- clusterRows(mat, DbscanParam(core.prop=0.1))
summary(is.na(dbscan.out))
```

# Two-phase clustering

We also provide a wrapper for a hybrid "two-step" approach for handling larger datasets.
Here, a fast agglomeration is performed with $k$-means to compact the data,
followed by a slower graph-based clustering step to obtain interpretable meta-clusters.
(This dataset is not, by and large, big enough for this approach to work particularly well.)

```{r}
set.seed(100) # for the k-means
two.out <- clusterRows(mat, TwoStepParam())
plotUMAP(sce, colour_by=I(two.out))
```

Each step is itself parametrized by `BlusterParam` objects, so it is possible to tune them individually:

```{r}
twop <- TwoStepParam(second=NNGraphParam(k=5))
twop
set.seed(100) # for the k-means
two.out <- clusterRows(mat, TwoStepParam())
plotUMAP(sce, colour_by=I(two.out))
```

# Obtaining full clustering statistics

Sometimes the vector of cluster assignments is not enough.
We can obtain more information about the clustering procedure by setting `full=TRUE` in `clusterRows()`.
For example, we could obtain the actual graph generated by `NNGraphParam()`:

```{r}
nn.out <- clusterRows(mat, NNGraphParam(), full=TRUE)
nn.out$objects$graph
```

The assignment vector is still reported in the `clusters` entry:

```{r}
table(nn.out$clusters)
```

# Further comments

`clusterRows()` enables users or developers to easily switch between clustering algorithms by changing a single argument.
Indeed, by passing the `BlusterParam` object across functions, we can ensure that the same algorithm is used through a workflow.
It is also helpful for package functions as it provides diverse functionality without compromising a clean function signature.
However, the true power of this framework lies in its extensibility.
Anyone can write a `clusterRows()` method for a new clustering algorithm with an associated `BlusterParam` subclass,
and that procedure is immediately compatible with any workflow or function that was already using `clusterRows()`.

# Session information {-}

```{r}
sessionInfo()
```

---
title: Assorted clustering diagnostics
author: 
- name: Aaron Lun
  email: infinite.monkeys.with.keyboards@gmail.com
output: BiocStyle::html_document
package: bluster
vignette: >
  %\VignetteIndexEntry{2. Clustering diagnostics}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}    
---

```{r, echo=FALSE}
knitr::opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE)
library(BiocStyle)
```

# Introduction 

The `r Biocpkg("bluster")` package provides a few diagnostics for quantitatively examining the cluster output.
These diagnostics 
We will demonstrate on another dataset from the `r Biocpkg("scRNAseq")` package,
clustered with graph-based methods via the `clusterRows()` generic as described in the previous vignette.

```{r}
library(scRNAseq)
sce <- GrunPancreasData()

# Quality control to remove bad cells.
library(scuttle)
qcstats <- perCellQCMetrics(sce)
qcfilter <- quickPerCellQC(qcstats, sub.fields="altexps_ERCC_percent")
sce <- sce[,!qcfilter$discard]

# Normalization by library size.
sce <- logNormCounts(sce)

# Feature selection.
library(scran)
dec <- modelGeneVar(sce)
hvgs <- getTopHVGs(dec, n=1000)

# Dimensionality reduction.
set.seed(1000)
library(scater)
sce <- runPCA(sce, ncomponents=20, subset_row=hvgs)

# Clustering.
library(bluster)
mat <- reducedDim(sce)
clust.info <- clusterRows(mat, NNGraphParam(), full=TRUE)
clusters <- clust.info$clusters
table(clusters)
```

# Computing the silhouette width

The silhouette width is a standard metric to quantify the separation between clusters generated by any procedure.
A cell with a large positive width is closer to other cells from the same cluster compared to cells from different clusters.
On the other hand, low or negative widths indicate that cells from different clusters are not well separated.

The exact silhouette calculation is rather computationally intensive so `r Biocpkg("bluster")` implements an approximation instead.
This is provided in the `approxSilhouette()` function, which returns the width for each cell and its closest (non-self) cluster.
Clusters consisting of cells with lower widths may warrant some more care during interpretation.

```{r}
sil <- approxSilhouette(mat, clusters)
sil
boxplot(split(sil$width, clusters))
```

The function also returns the identity of the closest "other" cluster for each cell.
This can be helpful to identify which clusters are easily confused to each other, based on how many of one cluster's cells are closer to the other cluster.

```{r}
best.choice <- ifelse(sil$width > 0, clusters, sil$other)
table(Assigned=clusters, Closest=best.choice)
```

# Computing the neighborhood purity

Another diagnostic uses the percentage of neighbors for each cell that belong to the same cluster.
Well-separated clusters should exhibit high percentages (i.e., "purities") as cells from different clusters do not mix.
Low purities are symptomatic of overclustering where cluster boundaries become more ambiguous.

The `neighborPurity()` function computes the purity of the neighborhood for each cell.
Clusters with systematically low purities may warrant some more care during interpretation.
By default, we perform some weighting so that large clusters do not have large purities simply because there are few cells assigned to other clusters in the dataset.

```{r}
pure <- neighborPurity(mat, clusters)
pure
boxplot(split(pure$purity, clusters))
```

The function also returns the identity of the other cluster with the highest percentage. 
This can again be useful to identify the relationships between clusters based on which pairs have the greatest intermingling in their neighborhoods.

```{r}
table(Assigned=clusters, Max=pure$maximum)
```

# Computing per-cluster RMSD

The root-mean-squared deviation (RMSD) for each cluster represents the dispersion of cells within each cluster.
A large RMSD value indicates that a cluster has high internal heterogeneity,
making it a good candidate for further subclustering.

```{r}
rmsd <- clusterRMSD(mat, clusters)
barplot(rmsd)
```

Alternatively, we can compute the within-cluster sum of squares (WCSS), a metric commonly seen in $k$-means clustering.
One could pick a "sensible" choice for $k$ by computing the WCSS for a range of values and picking the value the WCSS begins to plateau -
see Section \@ref(clustering-parameter-sweeps) for more details.

```{r}
clusterRMSD(mat, clusters, sum=TRUE)
```

# Computing graph modularity

For graph-based methods, we can compute the cluster modularity within clusters and between pairs of clusters.
Specifically, we examine the ratio of observed to expected edge weights for each pair of clusters (closely related to the modularity score used in many `cluster_*`  functions from `r CRANpkg("igraph")`).
We would usually expect to see high observed weights between cells in the same cluster with minimal weights between clusters, indicating that the clusters are well-separated.
Large off-diagonal entries indicate that the corresponding pair of clusters are closely related.

```{r}
g <- clust.info$objects$graph
ratio <- pairwiseModularity(g, clusters, as.ratio=TRUE)

library(pheatmap)
pheatmap(log10(ratio+1), cluster_cols=FALSE, cluster_rows=FALSE,
    col=rev(heat.colors(100)))
```

This may be better visualized with a force-directed layout: 

```{r}
cluster.gr <- igraph::graph_from_adjacency_matrix(log2(ratio+1), 
    mode="upper", weighted=TRUE, diag=FALSE)

# Increasing the weight to increase the visibility of the lines.
set.seed(1100101)
plot(cluster.gr, edge.width=igraph::E(cluster.gr)$weight*5,
    layout=igraph::layout_with_lgl)
```

We can also tune the resolution of the clustering _post hoc_ with the `mergeCommunities()` function.
This will iteratively merge the most closely related pair of clusters together until the desired number of clusters is reached.
For example, if we wanted to whittle down the number of clusters to 10, we could do:

```{r}
merged <- mergeCommunities(g, clusters, number=10)
table(merged)
```

# Comparing two clusterings

To compare two clusterings, the `pairwiseRand()` function computes the adjusted Rand index (ARI).
High ARIs indicate that the two clusterings are similar with respect to how they partition the observations,
and an ARI of 1 means that the clusterings are identical.

```{r}
hclusters <- clusterRows(mat, HclustParam(cut.dynamic=TRUE))
pairwiseRand(clusters, hclusters, mode="index")
```

Of course, a single number is not particularly useful, 
so `clusterRand()` also provides the capability to break down the ARI into its contributions from each cluster or cluster pair.
Specifically, for each cluster or cluster pair in a "reference" clustering (here, `clusters`),
we see whether it is preserved in the "alternative" clustering (here, `hclusters`).
Large values on the diagonal indicate that the reference cluster is recapitulated;
large values off the diagonal indicate that the separation between the corresponding pair of clusters is also maintained.
Conversely, low diagonal values indicate that the corresponding cluster is fragmented in the alternative,
and low off-diagonal values can be used as a diagnostic for loss of separation.

```{r}
ratio <- pairwiseRand(clusters, hclusters, mode="ratio")

library(pheatmap)
pheatmap(ratio, cluster_cols=FALSE, cluster_rows=FALSE,
    col=viridis::viridis(100), breaks=seq(-1, 1, length=101))
```

Explicit mappings between two clusterings can be performed using `linkClusters()` (see Section \@ref(linking-clusters)).
Alternatively, we can quantify the degree of "nesting" of one clustering within another with `nestedClusters()`;
this can be useful for verifying that the higher-resolution clustering is indeed nested within its coarser counterpart.

# Bootstrapping cluster stability

We can use bootstrapping to evaluate the effect of sampling noise on the stability of a clustering procedure.
The `bootstrapStability()` function will return the ARI of the original clusters against those generated from bootstrap replicates,
averaged across multiple bootstrap iterations.
High values indicate that the clustering is robust to sample noise.

```{r}
set.seed(1001010)
ari <-bootstrapStability(mat, clusters=clusters, 
    mode="index", BLUSPARAM=NNGraphParam())
ari
```

Advanced users may also set `mode="ratio"` to obtain a more detailed breakdown of the effect of noise on each cluster (pair).

```{r}
set.seed(1001010)
ratio <-bootstrapStability(mat, clusters=clusters,
    mode="ratio", BLUSPARAM=NNGraphParam())

library(pheatmap)
pheatmap(ratio, cluster_cols=FALSE, cluster_rows=FALSE,
    col=viridis::viridis(100), breaks=seq(-1, 1, length=101))
```

# Clustering parameter sweeps

The `clusterSweep()` function provides a convenient way to test multiple combinations of parameter settings.
Given a `BlusterParam` object and a set of values for each parameter, the function will repeat the clustering ith each combination of parameters.
The example below uses graph-based clustering with a variety of `k` as well as different community detection algorithms.

```{r}
combinations <- clusterSweep(mat, BLUSPARAM=SNNGraphParam(),
    k=c(5L, 10L, 15L, 20L), cluster.fun=c("walktrap", "louvain", "infomap"))
```

This yields a list containing all clusterings and the corresponding parameter combinations used to generate them.
The function will attempt to generate some sensible name for each combination, though this may require some manual curation for large numbers of parameters.

```{r}
colnames(combinations$clusters)
combinations$parameters
```

We can combine this with some of the metrics defined above to quantify cluster separation as a function of the clustering parameters.
This allows us to quickly determine which parameters have a noticeable impact on the results. 
In principle, we could choose the clustering with the greatest separation for further analysis;
however, this tends to be disappointing as it often favors overly broad clusters.

```{r}
set.seed(10)
nclusters <- 3:25
kcombos <- clusterSweep(mat, BLUSPARAM=KmeansParam(centers=5), centers=nclusters)

sil <- vapply(as.list(kcombos$clusters), function(x) mean(approxSilhouette(mat, x)$width), 0)
plot(nclusters, sil, xlab="Number of clusters", ylab="Average silhouette width")

pur <- vapply(as.list(kcombos$clusters), function(x) mean(neighborPurity(mat, x)$purity), 0)
plot(nclusters, pur, xlab="Number of clusters", ylab="Average purity")

wcss <- vapply(as.list(kcombos$clusters), function(x) sum(clusterRMSD(mat, x, sum=TRUE)), 0)
plot(nclusters, wcss, xlab="Number of clusters", ylab="Within-cluster sum of squares")
```

# Linking clusters 

If we have many clusterings, we can identify corresponding clusters with the `linkClusters()` function.
This constructs a graph where edges are formed between pairs of clusters from different clusterings, based on the number of cells assigned to both clusters.
Re-using some of the clusterings from our previous sweep, we might do:

```{r}
linked <- linkClusters(
    list(
        walktrap=combinations$clusters$k.10_cluster.fun.walktrap,
        louvain=combinations$clusters$k.10_cluster.fun.louvain,
        infomap=combinations$clusters$k.10_cluster.fun.infomap
    )
)
linked
```

This can be used to visualize the relationships between clusters, or to identify metaclusters across clusterings with community detection algorithms:

```{r}
meta <- igraph::cluster_walktrap(linked)
plot(linked, mark.groups=meta)
```

By default, the edge weights are computed by dividing the number of shared cells with the smaller of the total number of cells in either cluster.
This favors strong edges between a large cluster in one clustering and smaller subcluster in another (finer) clustering.
Alternative weighting schemes will favour a 1:1 mapping between clusterings, which can be easier to interpret.

# Comparing multiple clusterings

The `compareClusterings()` function will return a symmetric matrix of the ARIs between pairs of different clusterings.
This is helpful for visualizing the relationships between different clusterings, e.g., to see which parameters most contribute to differences between clusterings.

```{r}
aris <- compareClusterings(combinations$clusters)
g <- igraph::graph.adjacency(aris, mode="undirected", weighted=TRUE)
meta2 <- igraph::cluster_walktrap(g)
plot(g, mark.groups=meta2)
```

We can also identify groups of clusterings, typically corresponding to parameter combinations that yield more-or-less similar results.
This allows us to prune out combinations that are largely redundant prior to downstream analyses.

# Session information {-}

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bluster-package.R
\docType{package}
\name{bluster-package}
\alias{bluster}
\alias{bluster-package}
\title{bluster: Clustering Algorithms for Bioconductor}
\description{
Wraps common clustering algorithms in an easily extended S4 framework.
    Backends are implemented for hierarchical, k-means and graph-based clustering.
    Several utilities are also provided to compare and evaluate clustering results.
}
\author{
\strong{Maintainer}: Aaron Lun \email{infinite.monkeys.with.keyboards@gmail.com}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{HierarchicalParam-class}
\alias{HierarchicalParam-class}
\alias{.defaultScalarArguments,HierarchicalParam-method}
\alias{show,HierarchicalParam-method}
\title{The HierarchicalParam class}
\description{
The HierarchicalParam is a virtual subclass of the \linkS4class{BlusterParam} class.
It causes \code{\link{clusterRows}} to dispatch to clustering algorithms that produce a dissimilarity matrix and a dendrogram.
}
\section{Available slots}{

The virtual class provides \code{metric}, the choice of distance metric.
This is conventionally passed to \code{\link{dist}} and defaults to a Euclidean distance in most subclasses.

It also provides a number of slots to manage the final tree cut:
\itemize{
\item \code{cut.fun}, a function that takes a \link{hclust} object as its first argument and returns a vector of cluster assignments.
If \code{NULL}, the choice of function is determined from \code{cut.dynamic}.
\item \code{cut.dynamic}, a logical scalar indicating whether a dynamic tree cut should be performed by \code{\link[dynamicTreeCut]{cutreeDynamic}}.
Otherwise \code{\link{cutree}} is used.
Ignored if \code{cut.fun} is not \code{NULL}.
\item \code{cut.params}, further arguments to pass to the tree cut function specified by the previous arguments.
}
}

\section{Return value}{

The contract is that, when \code{full=TRUE}, the \code{objects} field of the \code{\link{clusterRows}} return value will always contain at least the following elements:
\itemize{
\item \code{dist}, a \link{dist} object containing a dissimilarity matrix, usually a distance matrix.
\item \code{hclust}, a \link{hclust} object containing a dendrogram.
}
}

\seealso{
\linkS4class{HclustParam}, for the archetypal example of a concrete subclass.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pairwiseModularity.R
\name{pairwiseModularity}
\alias{pairwiseModularity}
\title{Compute pairwise modularity}
\usage{
pairwiseModularity(graph, clusters, get.weights = FALSE, as.ratio = FALSE)
}
\arguments{
\item{graph}{A \link{graph} object from \pkg{igraph}, usually where each node represents an observation.}

\item{clusters}{Factor specifying the cluster identity for each node.}

\item{get.weights}{Logical scalar indicating whether the observed and expected edge weights should be returned, rather than the modularity.}

\item{as.ratio}{Logical scalar indicating whether the log-ratio of observed to expected weights should be returned.}
}
\value{
By default, an upper triangular numeric matrix of order equal to the number of clusters is returned.
Each entry corresponds to a pair of clusters and is proportional to the difference between the observed and expected edge weights between those clusters.

If \code{as.ratio=TRUE}, an upper triangular numeric matrix is again returned.
Here, each entry is equal to the ratio between the observed and expected edge weights.

If \code{get.weights=TRUE}, a list is returned containing two upper triangular numeric matrices. 
The \code{observed} matrix contains the observed sum of edge weights between and within clusters,
while the \code{expected} matrix contains the expected sum of edge weights under the random model.
}
\description{
Calculate the modularity of each pair of clusters from a graph, based on a null model of random connections between nodes.
}
\details{
This function computes a modularity score in the same manner as that from \code{\link{modularity}}.
The modularity is defined as the (scaled) difference between the observed and expected number of edges between nodes in the same cluster.
The expected number of edges is defined by a null model where edges are randomly distributed among nodes.
The same logic applies for weighted graphs, replacing the number of edges with the summed weight of edges.

Whereas \code{\link{modularity}} returns a modularity score for the entire graph, \code{pairwiseModularity} provides scores for the individual clusters.
The sum of the diagonal elements of the output matrix should be equal to the output of \code{\link{modularity}} 
(after supplying weights to the latter, if necessary).
A well-separated cluster should have mostly intra-cluster edges and a high modularity score on the corresponding diagonal entry,
while two closely related clusters that are weakly separated will have many inter-cluster edges and a high off-diagonal score.

In practice, the modularity may not the most effective metric for evaluating cluster separatedness.
This is because the modularity is proportional to the number of observations, so larger clusters will naturally have a large score regardless of separation.
An alternative approach is to set \code{as.ratio=TRUE}, which returns the ratio of the observed to expected weights for each entry of the matrix.
This adjusts for differences in cluster size and improves resolution of differences between clusters.

Directed graphs are treated as undirected inputs with \code{mode="each"} in \code{\link{as.undirected}}.
In the rare case that self-loops are present, these will also be handled correctly.
}
\examples{
m <- matrix(runif(10000), ncol=10)
clust.out <- clusterRows(m, BLUSPARAM=NNGraphParam(), full=TRUE)
clusters <- clust.out$clusters
g <- clust.out$objects$graph

# Examining the modularity values directly.
out <- pairwiseModularity(g, clusters)
out

# Compute the ratio instead, for visualization
# (log-transform to improve range of colors).
out <- pairwiseModularity(g, clusters, as.ratio=TRUE)
image(log2(out+1))

# This can also be used to construct a graph of clusters,
# for use in further plotting, a.k.a. graph abstraction.
# (Fiddle with the scaling values for a nicer plot.)
g2 <- igraph::graph_from_adjacency_matrix(out, mode="upper",
    diag=FALSE, weighted=TRUE)
plot(g2, edge.width=igraph::E(g2)$weight*10,
    vertex.size=sqrt(table(clusters))*2)

# Alternatively, get the edge weights directly:
out <- pairwiseModularity(g, clusters, get.weights=TRUE)
out

}
\seealso{
\code{\link{makeSNNGraph}}, for one method to construct \code{graph}.

\code{\link{modularity}}, for the calculation of the entire graph modularity.

\code{\link{pairwiseRand}}, which applies a similar breakdown to the Rand index.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NNGraphParam.R
\docType{class}
\name{NNGraphParam-class}
\alias{NNGraphParam-class}
\alias{show,NNGraphParam-method}
\alias{NNGraphParam}
\alias{clusterRows,ANY,NNGraphParam-method}
\title{Graph-based clustering}
\usage{
NNGraphParam(
  shared = TRUE,
  ...,
  cluster.fun = "walktrap",
  cluster.args = list()
)

\S4method{clusterRows}{ANY,NNGraphParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{shared}{Logical scalar indicating whether a shared NN graph should be constructed.}

\item{...}{Further arguments to pass to \code{\link{makeSNNGraph}} (if \code{shared=TRUE}) or \code{\link{makeKNNGraph}}.}

\item{cluster.fun}{Function specifying the method to use to detect communities in the NN graph.
The first argument of this function should be the NN graph and the return value should be a \link{communities} object.

Alternatively, this may be a string containing the suffix of any \pkg{igraph} community detection algorithm.
For example, \code{cluster.fun="louvain"} will instruct \code{\link{clusterRows}} to use \code{\link{cluster_louvain}}.
Defaults to \code{\link{cluster_walktrap}}.}

\item{cluster.args}{Further arguments to pass to the chosen \code{cluster.fun}.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{NNGraphParam} object.}

\item{full}{Logical scalar indicating whether the graph-based clustering objects should be returned.}
}
\value{
The \code{NNGraphParam} constructor will return a \linkS4class{NNGraphParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects};
the latter is a list with \code{graph} (the graph) and \code{communities} (the output of \code{cluster.fun}).
}
\description{
Run community detection algorithms on a nearest-neighbor (NN) graph within \code{\link{clusterRows}}.
}
\details{
To modify an existing NNGraphParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
}
\examples{
clusterRows(iris[,1:4], NNGraphParam())
clusterRows(iris[,1:4], NNGraphParam(k=5))
clusterRows(iris[,1:4], NNGraphParam(cluster.fun="louvain"))

}
\seealso{
\code{\link{makeSNNGraph}} and related functions, to build the graph.

\code{\link{cluster_walktrap}} and related functions, to perform community detection.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PamParam.R
\docType{class}
\name{PamParam-class}
\alias{PamParam-class}
\alias{show,PamParam-method}
\alias{.defaultScalarArguments,PamParam-method}
\alias{PamParam}
\alias{clusterRows,ANY,PamParam-method}
\title{Partitioning around medoids}
\usage{
PamParam(
  centers,
  metric = NULL,
  medoids = NULL,
  nstart = NULL,
  stand = NULL,
  do.swap = NULL,
  variant = NULL
)

\S4method{clusterRows}{ANY,PamParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{centers}{An integer scalar specifying the number of centers.
Alternatively, a function that takes the number of observations and returns the number of centers.}

\item{metric, medoids, nstart, stand, do.swap, variant}{Further arguments to pass to \code{\link{pam}}.
Set to the function defaults if not supplied.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{PamParam} object.}

\item{full}{Logical scalar indicating whether the full PAM statistics should be returned.}
}
\value{
The \code{PamParam} constructor will return a \linkS4class{PamParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
(a list containing \code{pam}, the direct output of \code{\link{pam}}).
}
\description{
Partition observations into k-medoids as a more robust version of k-means.
}
\details{
This class usually requires the user to specify the number of clusters beforehand.
However, we can also allow the number of clusters to vary as a function of the number of observations.
The latter is occasionally useful, e.g., to allow the clustering to automatically become more granular for large datasets.

To modify an existing PamParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
}
\examples{
clusterRows(iris[,1:4], PamParam(centers=4))
clusterRows(iris[,1:4], PamParam(centers=4, variant="faster", do.swap=FALSE))
clusterRows(iris[,1:4], PamParam(centers=sqrt))
}
\seealso{
\code{\link{pam}}, which actually does all the heavy lifting.

\linkS4class{KmeansParam}, for the more commonly used k-means algorithm.

\linkS4class{ClaraParam}, for a scalable extension to the PAM approach.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pairwiseRand.R
\name{pairwiseRand}
\alias{pairwiseRand}
\title{Compute pairwise Rand indices}
\usage{
pairwiseRand(ref, alt, mode = c("ratio", "pairs", "index"), adjusted = TRUE)
}
\arguments{
\item{ref}{A character vector or factor containing one set of groupings, considered to be the reference.}

\item{alt}{A character vector or factor containing another set of groupings, to be compared to \code{alt}.}

\item{mode}{String indicating whether to return the ratio, the number of pairs or the Rand index.}

\item{adjusted}{Logical scalar indicating whether the adjusted Rand index should be returned.}
}
\value{
If \code{mode="ratio"}, a square numeric matrix is returned with number of rows equal to the number of unique levels in \code{ref}.
Each diagonal entry is the ratio of the per-cluster \eqn{a} to the total number of pairs of cells in that cluster.
Each off-diagonal entry is the ratio of the per-cluster-pair \eqn{b} to the total number of pairs of cells for that pair of clusters.
Lower-triangular entries are set to \code{NA}.
If \code{adjusted=TRUE}, counts and totals are both adjusted prior to computing the ratio.

If \code{mode="pairs"}, a list is returned containing \code{correct} and \code{total},
both of which are square numeric matrices of the same arrangement as described above.
However, \code{correct} contains the actual numbers \eqn{a} (diagonal) and \eqn{b} (off-diagonal) rather than the ratios,
while \code{total} contains the total number of cell pairs in each cluster or pair of clusters.
If \code{adjusted=TRUE}, both matrices are adjusted by subtracting the random expectations from the counts.

If \code{mode="index"}, a numeric scalar is returned containing the Rand index (or ARI, if \code{adjusted=TRUE}).
}
\description{
Breaks down the Rand index calculation to report values for each cluster and pair of clusters
in a reference clustering compared to an alternative clustering.
}
\details{
Recall that the Rand index calculation consists of four numbers:
\describe{
\item{\eqn{a}}{The number of pairs of cells in the same cluster in \code{ref} and the same cluster in \code{alt}.}
\item{\eqn{b}}{The number of pairs of cells in different clusters in \code{ref} and different clusters in \code{alt}.}
\item{\eqn{c}}{The number of pairs of cells in the same cluster in \code{ref} and different clusters in \code{alt}.}
\item{\eqn{d}}{The number of pairs of cells in different clusters in \code{ref} but the same cluster in \code{alt}.}
}
The Rand index is then computed as \eqn{a + b} divided by \eqn{a + b + c + d}, i.e., the total number of pairs.

We can break these numbers down into values for each cluster or pair of clusters in \code{ref}.
For each cluster, we compute its value of \eqn{a}, 
i.e., the number of pairs of cells in \emph{that} cluster that are also in the same cluster in \code{alt}.
Similarly, for each pair of clusters in \code{ref}, we compute its value of \eqn{b},
i.e., the number of pairs of cells that have one cell in each of those clusters 
and also belong in different clusters in \code{alt}.

This process provides more information about the specific similarities or differences between \code{ref} and \code{alt},
rather than coalescing all the values into a single statistic. 
For example, it is now possible to see which specific clusters from \code{ref} are not reproducible in \code{alt},
or which specific partitions between pairs of clusters are not reproducible.
In the default output, such events can be diagnosed by looking for low entries in the ratio matrix;
on the other hand, values close to 1 indicate that \code{ref} is almost perfectly recapitulated by \code{alt}.

If \code{adjusted=TRUE}, we adjust all counts by subtracting their expected values under a model of random permutations.
This accounts for differences in the number and sizes of clusters within and between \code{ref} and \code{alt},
in a manner that mimics the calculation of adjusted Rand index (ARI).
We subtract expectations on a per-cluster or per-cluster-pair basis for \eqn{a} and \eqn{b}, respectively;
we also redefine the \dQuote{total} number of cell pairs for each cluster or cluster pair based on the denominator of the ARI.
}
\examples{
m <- matrix(runif(10000), ncol=10)

clust1 <- kmeans(m,3)$cluster
clust2 <- kmeans(m,5)$cluster

ratio <- pairwiseRand(clust1, clust2)
ratio

# Getting the raw counts:
pairwiseRand(clust1, clust2, mode="pairs")

# Computing the original Rand index.
pairwiseRand(clust1, clust2, mode="index")

}
\seealso{
\code{\link{pairwiseModularity}}, which applies the same breakdown to the cluster modularity.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linkClusters.R
\name{linkClusters}
\alias{linkClusters}
\alias{linkClustersMatrix}
\title{Create a graph between different clusterings}
\usage{
linkClusters(clusters, prefix = TRUE, denominator = c("union", "min", "max"))

linkClustersMatrix(x, y, denominator = c("union", "min", "max"))
}
\arguments{
\item{clusters}{A list of factors or vectors where each entry corresponds to a clustering.
All vectors should be of the same length.
The list itself should usually be named with a suitable label for each clustering.}

\item{prefix}{Logical scalar indicating whether the cluster levels should be prefixed with its clustering.
If \code{clusters} is not named, numeric prefixes are used instead.}

\item{denominator}{String specifying how the strength of the correspondence between clusters should be computed.}

\item{x, y}{Factor or vector specifying a clustering of the same cells.}
}
\value{
For \code{linkClusters}, a \link{graph} object where each node is a cluster level in one of the clusterings in \code{clusters}.
Edges are weighted by the strength of the correspondence between two clusters in different clusterings.

For \code{linkClustersMatrix}, a matrix is returned where each row and column corresponds to a cluster in \code{x} and \code{y}, respectively.
Entries represent the strength of the correspondence between the associated clusters;
this is equivalent to a submatrix of the adjacency matrix from the graph returned by \code{linkClusters}.
}
\description{
Create a graph that links together clusters from different clusterings,
e.g., generated using different parameter settings or algorithms.
This is useful for identifying corresponding clusters between clusterings
and to create meta-clusters from multiple clusterings.
}
\details{
Links are only formed between clusters from different clusterings, e.g., between clusters \eqn{X} in clustering 1 and \eqn{Y} in clustering 2.
The edge weight of each link is set to the strength of the correspondence between the two clusters;
this is defined from the number of cells with those two labels in their respective clusterings.
A larger number of cells indicates that \eqn{X} and \eqn{Y} are corresponding clusters.

Of course, the number of cells also depends on the total number of cells in each cluster.
To account for this, we normalize the strength by a function of the total number of cells in the two clusters.
The choice of function is determined by \code{denominator} and determines how the strength is adjusted for dissimilar cluster sizes.
\itemize{
\item For \code{"min"}, the number of shared cells is divided by the smaller of the totals between the two clusters.
\item For \code{"max"}, the number of shared cells is divided by the larger of the totals.
\item For \code{"union"}, the number of shared cells is divided by the size of the union of cells in the two clusters.
The result is equivalent to the Jaccard index.
}

In situations where \eqn{X} splits into multiple smaller clusters \eqn{Y1}, \eqn{Y2}, etc. in another clustering,
\code{denominator="min"} will report strong links between \eqn{X} and its constituent subclusters while \code{"max"} and \code{"union"} will report weak links.
Conversely, \code{denominator="max"} and \code{"union"} can only form strong links when there is a 1:1 mapping between clusters in different clusterings.
This usually yields simpler correspondences between clusterings at the cost of orphaning some of the smaller subclusters.
\code{denominator="union"} is most stringent as it will penalize the presence of non-shared cells in both clusters, whereas \code{"max"} only does so for the larger cluster.

The general idea is to use the graph returned by this function in visualization routines or for community-based clustering,
to identify \dQuote{clusters of clusters} that can inform about the relationships between clusterings.
}
\examples{
clusters <- list(
    nngraph = clusterRows(iris[,1:4], NNGraphParam()),
    hclust = clusterRows(iris[,1:4], HclustParam(cut.dynamic=TRUE)),
    kmeans = clusterRows(iris[,1:4], KmeansParam(5))
)

g <- linkClusters(clusters)
plot(g)

igraph::cluster_walktrap(g)

# Results as a matrix, for two clusterings:
linkClustersMatrix(clusters[[1]], clusters[[2]], denominator="union")
}
\seealso{
The \pkg{clustree} package, which provides another method for visualizing relationships between clusterings.

\code{\link{compareClusterings}}, which computes similarities between the clusterings themselves.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DmmParam.R
\docType{class}
\name{DmmParam-class}
\alias{DmmParam-class}
\alias{show,DmmParam-method}
\alias{DmmParam}
\alias{clusterRows,ANY,DmmParam-method}
\title{Dirichlet multinomial mixture clustering}
\usage{
DmmParam(k = 1:3, type = "laplace", seed = NULL, BPPARAM = SerialParam())

\S4method{clusterRows}{ANY,DmmParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{k}{An integer vector indicating the number of clusters to create with the DMM algorithm.
A vector containing two or more values will instruct \code{\link{clusterRows}} to perform clustering on each number,
and choose the optimal number of clusters based on \code{type}.}

\item{type}{A string specifying the method to use to find the optimal number of clusters. 
Must be equal to \code{"laplace"}, \code{"AIC"} or \code{"BIC"}.
Only used when \code{k} contains multiple values.}

\item{seed}{Integer scalar specifying the seed to use.
If \code{NULL}, a random value is used on each invocation of \code{\link{clusterRows}}.}

\item{BPPARAM}{A \linkS4class{BiocParallelParam} object indicating how multiple clusterings should be parallelized.
Only relevant if \code{k} contains multiple values.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.
Values are expected to be counts.}

\item{BLUSPARAM}{A \linkS4class{BlusterParam} object specifying the algorithm to use.}

\item{full}{Logical scalar indicating whether the full clustering statistics should be returned for each method.}
}
\value{
The \code{DmmParam} constructor will return a \linkS4class{DmmParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects};
the latter is a list containing:
\itemize{
\item \code{dmm}, a list containing the output of \code{\link[DirichletMultinomial]{dmn}} for each value of \code{k}.
\item \code{best}, an integer scalar specifying the best choice of \code{k} according to the method of \code{type}.
\item \code{prob}, a matrix array of probabilities where each row is an observation and each column is a cluster.
The number of columns is set to the best number of clusters in \code{best}.
\item \code{seed}, an integer scalar specifying the seed used for clustering.
}
}
\description{
Apply the Dirichlet multinomial mixture (DMM) algorithm from the \pkg{DirichletMultinomial} package.
This is commonly used in microbial ecology and in analyses of metagenomic and 16S rRNA count data.
}
\details{
To modify an existing DmmParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
}
\examples{
# Mocking up a small example.
nfeatures <- 50
out1 <- matrix(rpois(20 * nfeatures, lambda = rgamma(nfeatures, 5)), ncol=nfeatures, byrow=TRUE)
out2 <- matrix(rpois(20 * nfeatures, lambda = rgamma(nfeatures, 5)), ncol=nfeatures, byrow=TRUE)
out <- rbind(out1, out2)
clusterRows(out, DmmParam())

}
\references{
Holmes I, Harris K and Quince C (2012).
Dirichlet multinomial mixtures: generative models for microbial metagenomics.
\emph{PLoS ONE}, 7(2), 1-15
}
\author{
Basil Courbayre
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{BlusterParam-class}
\alias{BlusterParam-class}
\alias{[[,BlusterParam-method}
\alias{[[<-,BlusterParam-method}
\alias{show,BlusterParam-method}
\title{The BlusterParam class}
\description{
The BlusterParam class is a virtual base class controlling S4 dispatch in \code{\link{clusterRows}} and friends.
Concrete subclasses specify the choice of clustering algorithm,
while the slots of an instance of such a subclass represent the parameters for that algorithm.
}
\section{Available methods}{

In the following code snippets, \code{x} is a \linkS4class{BlusterParam} object or one of its subclasses.
\itemize{
\item \code{x[[i]]} will return the value of the parameter \code{i}.
Refer to the documentation for each concrete subclass for more details on the available parameters.
\item \code{x[[i]] <- value} will set the value of the parameter \code{i} to \code{value}.
\item \code{show(x)} will print some information about the class instance.
}
}

\seealso{
\linkS4class{HclustParam}, \linkS4class{KmeansParam} and \linkS4class{NNGraphParam} 
for some examples of concrete subclasses.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/BlusterParam.R
\name{.defaultScalarArguments}
\alias{.defaultScalarArguments}
\alias{.defaultScalarArguments,BlusterParam-method}
\alias{.showScalarArguments}
\alias{.extractScalarArguments}
\title{Define the default arguments}
\usage{
.defaultScalarArguments(x)

.showScalarArguments(object)

.extractScalarArguments(x)
}
\arguments{
\item{x, object}{A \linkS4class{BlusterParam} object.}
}
\value{
For \code{.defaultScalarArguments}, a named character vector is returned.
Each entry corresponds to an argument to the clustering function - the name is the argument name, and the value is the argument type.

For \code{.extractScalarArguments}, a named list of non-default scalar arguments is returned.
Any arguments set to their default values are omitted from the list.

For \code{.showScalarArguments}, the values of the arguments are printed to screen.
Default values are marked with \code{[default]}.
}
\description{
Provide a consistent mechanism to handle specification of default arguments to the underlying clustering functions.
}
\details{
The idea is to simplify the derivation of new \linkS4class{BlusterParam} objects,
by allowing developers to indicate that the underlying function default should be used for particular arguments.
This avoids duplication of the default arguments in the object constructor;
instead, default arguments can be indicated as such by setting them to \code{NULL},
in which case they will not be explicitly passed to the underlying clustering function.
}
\examples{
.defaultScalarArguments(PamParam(10))
.extractScalarArguments(PamParam(10))
.extractScalarArguments(PamParam(10, variant="faster"))

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeSNNGraph.R
\name{makeSNNGraph}
\alias{makeSNNGraph}
\alias{makeKNNGraph}
\alias{neighborsToSNNGraph}
\alias{neighborsToKNNGraph}
\title{Build a nearest-neighbor graph}
\usage{
makeSNNGraph(
  x,
  k = 10,
  type = c("rank", "number", "jaccard"),
  BNPARAM = KmknnParam(),
  BPPARAM = SerialParam()
)

makeKNNGraph(
  x,
  k = 10,
  directed = FALSE,
  BNPARAM = KmknnParam(),
  BPPARAM = SerialParam()
)

neighborsToSNNGraph(indices, type = c("rank", "number", "jaccard"))

neighborsToKNNGraph(indices, directed = FALSE)
}
\arguments{
\item{x}{A matrix-like object containing expression values for each observation (row) and dimension (column).}

\item{k}{An integer scalar specifying the number of nearest neighbors to consider during graph construction.}

\item{type}{A string specifying the type of weighting scheme to use for shared neighbors.}

\item{BNPARAM}{A \linkS4class{BiocNeighborParam} object specifying the nearest neighbor algorithm.}

\item{BPPARAM}{A \linkS4class{BiocParallelParam} object to use for parallel processing.}

\item{directed}{A logical scalar indicating whether the output of \code{buildKNNGraph} should be a directed graph.}

\item{indices}{An integer matrix where each row corresponds to an observation
and contains the indices of the \code{k} nearest neighbors (by increasing distance and excluding self) from that observation.}
}
\value{
A \link{graph} where nodes are cells and edges represent connections between nearest neighbors.
For \code{buildSNNGraph}, these edges are weighted by the number of shared nearest neighbors.
For \code{buildKNNGraph}, edges are not weighted but may be directed if \code{directed=TRUE}.
}
\description{
Build a shared or k-nearest-neighbors graph of observations for downstream community detection.
}
\details{
The \code{makeSNNGraph} function builds a shared nearest-neighbour graph using observations as nodes.
For each observation, its \code{k} nearest neighbours are identified using the \code{\link{findKNN}} function,
based on distances between their expression profiles (Euclidean by default).
An edge is drawn between all pairs of observations that share at least one neighbour,
weighted by the characteristics of the shared nearest neighbors - see \dQuote{Weighting Schemes} below.

The aim is to use the SNN graph to perform clustering of observations via community detection algorithms in the \pkg{igraph} package.
This is faster and more memory efficient than hierarchical clustering for large numbers of observations.
In particular, it avoids the need to construct a distance matrix for all pairs of observations.
Only the identities of nearest neighbours are required, which can be obtained quickly with methods in the \pkg{BiocNeighbors} package.

The choice of \code{k} controls the connectivity of the graph and the resolution of community detection algorithms.
Smaller values of \code{k} will generally yield smaller, finer clusters, while increasing \code{k} will increase the connectivity of the graph and make it more difficult to resolve different communities.
The value of \code{k} can be roughly interpreted as the anticipated size of the smallest subpopulation.
If a subpopulation in the data has fewer than \code{k+1} observations, \code{buildSNNGraph} and \code{buildKNNGraph} will forcibly construct edges between observations in that subpopulation and observations in other subpopulations. 
This increases the risk that the subpopulation will not form its own cluster as it is more interconnected with the rest of the observations in the dataset.

The \code{makeKNNGraph} method builds a simpler k-nearest neighbour graph.
Observations are again nodes, and edges are drawn between each observation and its k-nearest neighbours.
No weighting of the edges is performed.
In theory, these graphs are directed as nearest neighour relationships may not be reciprocal.
However, by default, \code{directed=FALSE} such that an undirected graph is returned.

The \code{neighborsToSNNGraph} and \code{neighborsToKNNGraph} functions operate directly on a matrix of nearest neighbor indices,
obtained using functions like \code{\link{findKNN}}.
This may be useful for constructing a graph from precomputed nearest-neighbor search results.
Note that the user is responsible for ensuring that the indices are valid, i.e., \code{range(indices)} is positive and no greater than \code{max(indices)}.
}
\section{Shared neighbor weighting schemes}{

If \code{type="rank"}, the weighting scheme defined by Xu and Su (2015) is used.
The weight between two nodes is \eqn{k - r/2} where \eqn{r} is the smallest sum of ranks for any shared neighboring node.
For example, if one node was the closest neighbor of each of two nodes, the weight between the two latter nodes would be \eqn{k - 1}.
For the purposes of this ranking, each node has a rank of zero in its own nearest-neighbor set. 
More shared neighbors, or shared neighbors that are close to both observations, will generally yield larger weights.

If \code{type="number"}, the weight between two nodes is simply the number of shared nearest neighbors between them.
The weight can range from zero to \eqn{k + 1}, as the node itself is included in its own nearest-neighbor set.
This is a simpler scheme that is also slightly faster but does not account for the ranking of neighbors within each set.

If \code{type="jaccard"}, the weight between two nodes is the Jaccard similarity between the two sets of neighbors for those nodes.
This weight can range from zero to 1, and is a monotonic transformation of the weight used by \code{type="number"}.
It is provided for consistency with other clustering algorithms such as those in \pkg{seurat}.

Tehcnically, edges with zero weights are assigned a nominal small positive weight of the order of \code{1e-6}.
This is done only to satisfy the requirements for positive weights in many \pkg{igraph} clustering algorithms.
We do not just remove these edges as that might lead to the situation where some observations have no edges at all and thus form single-observation clusters.

Note that the behavior of \code{k} for \code{type="rank"} is slightly different from that used in the original SNN-Cliq implementation by Xu and Su.
The original implementation considers each observation to be its first nearest neighbor that contributes to \code{k}.
Here, the \code{k} nearest neighbours refers to the number of \emph{other} observations.
}

\examples{
m <- matrix(rnorm(10000), ncol=10)

g <- makeSNNGraph(m)
clusters <- igraph::cluster_fast_greedy(g)$membership
table(clusters)

# Any clustering method from igraph can be used:
clusters <- igraph::cluster_walktrap(g)$membership
table(clusters)

# Smaller 'k' usually yields finer clusters:
g <- makeSNNGraph(m, k=5)
clusters <- igraph::cluster_walktrap(g)$membership
table(clusters)

}
\references{
Xu C and Su Z (2015).
Identification of cell types from single-cell transcriptomes using a novel clustering method.
\emph{Bioinformatics} 31:1974-80
}
\seealso{
See \code{\link{make_graph}} for details on the graph output object.

See \code{\link{cluster_walktrap}}, \code{\link{cluster_louvain}} and related functions in \pkg{igraph} for clustering based on the produced graph.

Also see \code{\link{findKNN}} for specifics of the nearest-neighbor search.
}
\author{
Aaron Lun, with KNN code contributed by Jonathan Griffiths.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bootstrapStability.R
\name{bootstrapStability}
\alias{bootstrapStability}
\title{Assess cluster stability by bootstrapping}
\usage{
bootstrapStability(
  x,
  FUN = clusterRows,
  clusters = NULL,
  iterations = 20,
  average = c("median", "mean"),
  ...,
  compare = NULL,
  mode = "ratio",
  adjusted = TRUE,
  transposed = FALSE
)
}
\arguments{
\item{x}{A numeric matrix-like object containing observations in the rows and variables in the columns.
If \code{transposed=TRUE}, observations are assumed to be in the columns instead.}

\item{FUN}{A function that takes \code{x} as its first argument and returns a vector or factor of cluster identities.}

\item{clusters}{A vector or factor of cluster identities equivalent to that obtained by calling \code{FUN(x, ...)}.
This is provided as an additional argument in the case that the clusters have already been computed,
in which case we can save a single round of computation.}

\item{iterations}{A positive integer scalar specifying the number of bootstrap iterations.}

\item{average}{String specifying the method to use to average across bootstrap iterations.}

\item{...}{Further arguments to pass to \code{FUN} to control the clustering procedure.}

\item{compare}{A function that accepts the original clustering and the bootstrapped clustering,
and returns a numeric vector or matrix containing some measure of similarity between them - see Details.}

\item{mode, adjusted}{Further arguments to pass to \code{\link{pairwiseRand}} when \code{compare=NULL}.}

\item{transposed}{Logical scalar indicating that resampling should be done on the columns instead.}
}
\value{
If \code{compare=NULL} and \code{mode="ratio"}, a numeric matrix is returned with upper triangular entries set to the ratio of the adjusted observation pair counts (see \code{?\link{pairwiseRand}}) for each pair of clusters in \code{clusters}.
Each ratio is averaged across bootstrap iterations as specified by \code{average}.

If \code{compare=NULL} and \code{mode="index"}, a numeric scalar containing the average ARI between \code{clusters} and the bootstrap replicates across iterations is returned.

If \code{compare} is provided, a numeric array of the same type as the output of \code{compare} is returned, containing the average statistic(s) across bootstrap replicates.
}
\description{
Generate bootstrap replicates and recluster on them to determine the stability of clusters with respect to sampling noise.
}
\details{
Bootstrapping is conventionally used to evaluate the precision of an estimator by applying it to an \emph{in silico}-generated replicate dataset.
We can (ab)use this framework to determine the stability of the clusters given the original dataset.
We sample observations with replacement from \code{x}, perform clustering with \code{FUN} and compare the new clusters to \code{clusters}.

For comparing clusters, we compute the ratio matrix from \code{\link{pairwiseRand}} and average its values across bootstrap iterations.
High on-diagonal values indicate that the corresponding cluster remains coherent in the bootstrap replicates,
while high off-diagonal values indicate that the corresponding pair of clusters are still separated in the replicates.
If a single value is necessary, we can instead average the adjusted Rand indices across iterations with \code{mode="index"}.

We use the ratio matrix by default as it is more interpretable than a single value like the ARI or the Jaccard index (see the \pkg{fpc} package).
It focuses on the relevant differences between clusters, allowing us to determine which aspects of a clustering are stable.
For example, A and B may be well separated but A and C may not be, which is difficult to represent in a single stability measure for A.
If our main interest lies in the A/B separation, we do not want to be overly pessimistic about the stability of A, even though it might not be well-separated from all other clusters.
}
\section{Using another comparison function}{

We can use a different method for comparing clusterings by setting \code{compare}.
This is expected to be a function that takes two arguments - 
the original clustering first, and the bootstrapped clustering second - 
and returns some kind of numeric scalar, vector or matrix containing 
statistics for the similarity or difference between the original and bootstrapped clustering.
These statistics are then averaged across all bootstrap iterations.

Any numeric output of \code{compare} is acceptable as long as the dimensions are only dependent on the \emph{levels} of the original clustering - including levels that have no observations, due to resampling! - and thus do not change across bootstrap iterations.
}

\section{Statistical note on bootstrap comparisons}{

Technically speaking, some mental gymnastics are required to compare the original and bootstrap clusters in this manner.
After bootstrapping, the sampled observations represent distinct entities from the original dataset (otherwise it would be difficult to treat them as independent replicates) for which the original clusters do not immediately apply.
Instead, we assume that we perform label transfer using a nearest-neighbors approach - which, in this case, is the same as using the original label for each observation, as the nearest neighbor of each resampled observation to the original dataset is itself.

Needless to say, bootstrapping will only generate replicates that differ by sampling noise.
Real replicates will differ due to composition differences, variability in expression across individuals, etc.
Thus, any stability inferences from bootstrapping are likely to be overly optimistic.
}

\examples{
m <- matrix(runif(10000), ncol=10)

# BLUSPARAM just gets passed to the default FUN=clusterRows:
bootstrapStability(m, BLUSPARAM=KmeansParam(4), iterations=10)

# Defining your own clustering function:
kFUN <- function(x) kmeans(x, 2)$cluster  
bootstrapStability(m, FUN=kFUN)

# Using an alternative comparison, in this case the Rand index:
bootstrapStability(m, FUN=kFUN, compare=pairwiseRand)

}
\seealso{
\code{\link{clusterRows}}, for the default clustering function.

\code{\link{pairwiseRand}}, for the calculation of the ARI.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ClaraParam.R
\docType{class}
\name{ClaraParam-class}
\alias{ClaraParam-class}
\alias{show,ClaraParam-method}
\alias{.defaultScalarArguments,ClaraParam-method}
\alias{ClaraParam}
\alias{clusterRows,ANY,ClaraParam-method}
\title{Clustering Large Applications}
\usage{
ClaraParam(
  centers,
  metric = NULL,
  stand = NULL,
  samples = NULL,
  sampsize = NULL
)

\S4method{clusterRows}{ANY,ClaraParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{centers}{An integer scalar specifying the number of centers.
Alternatively, a function that takes the number of observations and returns the number of centers.}

\item{metric, stand, samples, sampsize}{Further arguments to pass to \code{\link{clara}}.
Set to the function defaults if not supplied.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{ClaraParam} object.}

\item{full}{Logical scalar indicating whether the full PAM statistics should be returned.}
}
\value{
The \code{ClaraParam} constructor will return a \linkS4class{ClaraParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
(a list containing \code{clara}, the direct output of \code{\link{clara}}).
}
\description{
Run the CLARA algorithm, an extension of the PAM method for large datasets.
}
\details{
This class usually requires the user to specify the number of clusters beforehand.
However, we can also allow the number of clusters to vary as a function of the number of observations.
The latter is occasionally useful, e.g., to allow the clustering to automatically become more granular for large datasets.

To modify an existing ClaraParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.

Note that \code{clusterRows} will always use \code{rngR=TRUE}, for greater consistency with other algorithms of the \linkS4class{FixedNumberParam} class;
and \code{pamLike=TRUE}, for consistency with the PAM implementation from which it was derived.
}
\examples{
clusterRows(iris[,1:4], ClaraParam(centers=4))
clusterRows(iris[,1:4], ClaraParam(centers=4, sampsize=50))
clusterRows(iris[,1:4], ClaraParam(centers=sqrt))
}
\seealso{
\code{\link{clara}}, which actually does all the heavy lifting.

\linkS4class{PamParam}, for the original PAM algorithm.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HclustParam.R
\docType{class}
\name{HclustParam-class}
\alias{HclustParam-class}
\alias{show,HclustParam-method}
\alias{HclustParam}
\alias{clusterRows,ANY,HclustParam-method}
\title{Hierarchical clustering}
\usage{
HclustParam(
  metric = "euclidean",
  method = "complete",
  cut.fun = NULL,
  cut.dynamic = FALSE,
  cut.height = NULL,
  cut.number = NULL,
  ...
)

\S4method{clusterRows}{ANY,HclustParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{metric}{String specifying the distance metric to use in \code{\link{dist}}.}

\item{method}{String specifying the agglomeration method to use in \code{\link{hclust}}.}

\item{cut.fun}{Function specifying the method to use to cut the dendrogram.
The first argument of this function should be the output of \code{\link{hclust}},
and the return value should be an atomic vector specifying the cluster assignment for each observation.
Defaults to \code{\link{cutree}} if \code{cut.dynamic=FALSE} and \code{\link[dynamicTreeCut]{cutreeDynamic}} otherwise.}

\item{cut.dynamic}{Logical scalar indicating whether a dynamic tree cut should be performed using the \pkg{dynamicTreeCut} package.}

\item{cut.height}{Numeric scalar specifying the cut height to use for the tree cut when \code{cut.fun=NULL}.
If \code{NULL}, defaults to half the tree height.
Ignored if \code{cut.number} is set.}

\item{cut.number}{Integer scalar specifying the number of clusters to generate from the tree cut when \code{cut.fun=NULL}.}

\item{...}{Further arguments to pass to \code{cut.fun}, when \code{cut.dynamic=TRUE} or \code{cut.fun} is non-\code{NULL}.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{HclustParam} object.}

\item{full}{Logical scalar indicating whether the hierarchical clustering statistics should be returned.}
}
\value{
The \code{HclustParam} constructor will return a \linkS4class{HclustParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects} (the \code{\link{hclust}} output).
}
\description{
Run the base \code{\link{hclust}} function on a distance matrix within \code{\link{clusterRows}}.
}
\details{
To modify an existing HclustParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
}
\examples{
clusterRows(iris[,1:4], HclustParam())
clusterRows(iris[,1:4], HclustParam(method="ward.D2"))

}
\seealso{
\code{\link{dist}}, \code{\link{hclust}} and \code{\link{cutree}}, which actually do all the heavy lifting.

\code{\link[dynamicTreeCut]{cutreeDynamic}}, for an alternative tree cutting method to use in \code{cut.fun}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TwoStepParam.R
\docType{class}
\name{TwoStepParam-class}
\alias{TwoStepParam-class}
\alias{show,TwoStepParam-method}
\alias{TwoStepParam}
\alias{clusterRows,ANY,TwoStepParam-method}
\title{Two step clustering with vector quantization}
\usage{
TwoStepParam(first = KmeansParam(centers = sqrt), second = NNGraphParam())

\S4method{clusterRows}{ANY,TwoStepParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{first}{A \linkS4class{BlusterParam} object specifying a fast vector quantization technique.}

\item{second}{A \linkS4class{BlusterParam} object specifying the second clustering technique on the centroids.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{KmeansParam} object.}

\item{full}{Logical scalar indicating whether the clustering statistics from both steps should be returned.}
}
\value{
The \code{TwoStepParam} constructor will return a \linkS4class{TwoStepParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with a \code{clusters} factor and an \code{objects} list containing:
\itemize{
\item \code{first}, a list of objects from the first clustering step.
This is equal to the \code{objects} list in the output of \code{\link{clusterRows}} with the \code{first} BlusterParam.
\item \code{centroids}, a numeric matrix of centroids generated from the first clustering step.
\item \code{second}, a list of objects from the second clustering step on the centroids.
This is equal to the \code{objects} list in the output of \code{\link{clusterRows}} with the \code{second} BlusterParam.
}
}
\description{
For large datasets, we can perform vector quantization (e.g., with k-means clustering) to create centroids.
These centroids are then subjected to a slower clustering technique such as graph-based community detection.
The label for each cell is set to the label of the centroid to which it was assigned.
}
\details{
Here, the idea is to use a fast clustering algorithm to perform vector quantization and reduce the size of the dataset,
followed by a slower algorithm that aggregates the centroids for easier interpretation.
The exact choice of the number of clusters is less relevant to the first clustering step
as long as not too many centroids are generated but the clusters are still sufficiently granular.
The second step can take more care (and computational time) summarizing the centroids into meaningful \dQuote{meta-clusters}.

The default choice is to use k-means for the first step, with number of clusters set to the root of the number of observations;
and graph-based clustering for the second step, which automatically detects a suitable number of clusters.
K-means also eliminates density differences in the data that can introduce variable resolution from graph-based methods.

To modify an existing TwoStepParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
}
\examples{
m <- matrix(runif(100000), ncol=10)
stuff <- clusterRows(m, TwoStepParam())
table(stuff)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mergeCommunities.R
\name{mergeCommunities}
\alias{mergeCommunities}
\title{Merge communities from graph-based clustering}
\usage{
mergeCommunities(graph, clusters, number = NULL, steps = NULL)
}
\arguments{
\item{graph}{A \link{graph} object from \pkg{igraph}, usually where each node represents an observation.}

\item{clusters}{Factor specifying the cluster identity for each node.}

\item{number}{Integer scalar specifying the number of clusters to obtain.
Ignored if \code{steps} is specified.}

\item{steps}{Integer scalar specifying the number of merge steps.}
}
\value{
A vector or factor of the same length as \code{clusters},
containing the desired number of merged clusters.
}
\description{
Adjust the resolution of a graph-based community detection algorithm by greedily merging clusters together.
At each step, the pair of clusters that yield the highest modularity are merged.
}
\details{
This function is similar to the \code{\link{cut_at}} function from the \pkg{igraph} package,
but works on clusters that were not generated by a hierarchical algorithm.
The aim is to facilitate rapid adjustment of the number of clusters without having to repeat the clustering -
or, even worse, repeating the graph construction, e.g., in \code{\link{makeSNNGraph}}.
}
\examples{
output <- clusterRows(iris[,1:4], NNGraphParam(k=5), full=TRUE)
table(output$clusters)

merged <- mergeCommunities(output$objects$graph, output$clusters, number=3)
table(merged)

}
\seealso{
\code{\link{cut_at}}, for a faster and more natural adjustment when using a hierarchical community detection algorithm.

\linkS4class{NNGraphParam}, for a one-liner to generate graph-based clusters.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AgnesParam.R
\docType{class}
\name{AgnesParam-class}
\alias{AgnesParam-class}
\alias{show,AgnesParam-method}
\alias{.defaultScalarArguments,AgnesParam-method}
\alias{AgnesParam}
\alias{clusterRows,ANY,AgnesParam-method}
\title{Agglomerative nesting}
\usage{
AgnesParam(
  metric = NULL,
  stand = NULL,
  method = NULL,
  par.method = NULL,
  cut.fun = NULL,
  cut.dynamic = FALSE,
  cut.params = list()
)

\S4method{clusterRows}{ANY,AgnesParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{metric, stand, method, par.method}{Further arguments to pass to \code{\link{agnes}}.}

\item{cut.fun}{Function specifying the method to use to cut the dendrogram.
The first argument of this function should be the output of \code{\link{hclust}},
and the return value should be an atomic vector specifying the cluster assignment for each observation.
Defaults to \code{\link{cutree}} if \code{cut.dynamic=FALSE} and \code{\link[dynamicTreeCut]{cutreeDynamic}} otherwise.}

\item{cut.dynamic}{Logical scalar indicating whether a dynamic tree cut should be performed using the \pkg{dynamicTreeCut} package.}

\item{cut.params}{Further arguments to pass to \code{cut.fun}.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{HclustParam} object.}

\item{full}{Logical scalar indicating whether the hierarchical clustering statistics should be returned.}
}
\value{
The \code{AgnesParam} constructor will return a \linkS4class{AgnesParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
(a list containing \code{agnes}, the function output; \code{dist}, the dissimilarity matrix; and \code{hclust}, a \link{hclust} object created from \code{agnes}).
}
\description{
Run the \code{\link{agnes}} function on a distance matrix within \code{\link{clusterRows}}.
}
\details{
To modify an existing AgnesParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.

If \code{cut.fun=NULL}, \code{cut.dynamic=FALSE} and \code{cut.params} does not have \code{h} or \code{k},
\code{\link{clusterRows}} will automatically set \code{h} to half the tree height when calling \code{\link{cutree}}.
}
\examples{
clusterRows(iris[,1:4], AgnesParam())
clusterRows(iris[,1:4], AgnesParam(method="ward"))

}
\seealso{
\code{\link{agnes}}, which actually does all the heavy lifting.

\linkS4class{HclustParam}, for the more commonly used implementation of hierarchical clustering.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R
\name{clusterRows}
\alias{clusterRows}
\title{Cluster rows of a matrix}
\usage{
clusterRows(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{BlusterParam} object specifying the algorithm to use.}

\item{full}{Logical scalar indicating whether the full clustering statistics should be returned for each method.}
}
\value{
By default, a factor of length equal to \code{nrow(x)} containing cluster assignments for each row of \code{x}.

If \code{full=TRUE}, a list is returned containing \code{clusters}, a factor as described above;
and \code{objects}, an arbitrary object containing algorithm-specific statistics or intermediate objects.
}
\description{
Cluster rows of a matrix-like object with a variety of algorithms.
}
\details{
This generic allows users to write agile code that can use a variety of clustering algorithms.
By simply changing \code{BLUSPARAM}, we can tune the clustering procedure in analysis workflows and package functions.
}
\examples{
m <- matrix(runif(10000), ncol=10)

clusterRows(m, KmeansParam(10L))
clusterRows(m, HclustParam())
clusterRows(m, NNGraphParam())
}
\seealso{
\linkS4class{HclustParam}, \linkS4class{KmeansParam} and \linkS4class{NNGraphParam} 
for some examples of values for \code{BLUSPARAM}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{FixedNumberParam-class}
\alias{FixedNumberParam-class}
\alias{centers}
\alias{centers,FixedNumberParam-method}
\alias{centers<-}
\alias{centers<-,FixedNumberParam-method}
\alias{show,FixedNumberParam-method}
\title{The FixedNumberParam class}
\description{
The FixedNumberParam is a virtual subclass of the \linkS4class{BlusterParam} class.
It causes \code{\link{clusterRows}} to dispatch to clustering algorithms that rely on a pre-specified number of clusters, e.g., \linkS4class{KmeansParam}.
}
\section{Available methods}{

\code{centers(x, n=NULL)} will return the specified number of centers in a FixedNumberParam \code{x}.
This can be an positive integer, or a function that accepts the number of observations and returns a positive number.
If a function and \code{n} is supplied, the function is called on \code{n} and the result is rounded to obtain an integer.

\code{centers(x) <- value} will replace the specified number of centers in \code{x} with an integer scalar or function \code{value}.
The function should accept a single argument and return a positive integer.
}

\seealso{
\linkS4class{KmeansParam}, for the archetypal example of a concrete subclass.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MbkmeansParam.R
\docType{class}
\name{MbkmeansParam-class}
\alias{MbkmeansParam-class}
\alias{show,MbkmeansParam-method}
\alias{MbkmeansParam}
\alias{clusterRows,ANY,MbkmeansParam-method}
\title{Mini-batch k-means clustering}
\usage{
MbkmeansParam(
  centers,
  batch_size = NULL,
  max_iters = 100,
  num_init = 1,
  init_fraction = NULL,
  initializer = "kmeans++",
  calc_wcss = FALSE,
  early_stop_iter = 10,
  tol = 1e-04,
  BPPARAM = SerialParam()
)

\S4method{clusterRows}{ANY,MbkmeansParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{centers}{An integer scalar specifying the number of centers.
Alternatively, a function that takes the number of observations and returns the number of centers.}

\item{batch_size, max_iters, num_init, init_fraction, initializer, calc_wcss, early_stop_iter, tol, BPPARAM}{Further arguments to pass to \code{\link[mbkmeans]{mbkmeans}}.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{MbkmeansParam} object.}

\item{full}{Logical scalar indicating whether the full mini-batch k-means statistics should be returned.}
}
\value{
The \code{MbkmeansParam} constructor will return a \linkS4class{MbkmeansParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
(a list containing \code{mbkmeans}, the direct output of \code{\link[mbkmeans]{mbkmeans}}).
}
\description{
Run the mini-batch k-means \code{\link[mbkmeans]{mbkmeans}} function with the specified number of centers within \code{\link{clusterRows}}.
This sacrifices some accuracy for speed compared to the standard k-means algorithm.
Note that this requires installation of the \pkg{mbkmeans} package.
}
\details{
This class usually requires the user to specify the number of clusters beforehand.
However, we can also allow the number of clusters to vary as a function of the number of observations.
The latter is occasionally useful, e.g., to allow the clustering to automatically become more granular for large datasets.

To modify an existing MbkmeansParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.

For \code{batch_size} and \code{init_fraction}, a value of \code{NULL} means that the default arguments in the \code{\link[mbkmeans]{mbkmeans}} function signature are used.
These defaults are data-dependent and so cannot be specified during construction of the MbkmeansParam object, but instead are defined within the \code{clusterRows} method.
}
\examples{
clusterRows(iris[,1:4], MbkmeansParam(centers=3))
clusterRows(iris[,1:4], MbkmeansParam(centers=3, batch_size=10))
clusterRows(iris[,1:4], MbkmeansParam(centers=3, init_fraction=0.5))
}
\seealso{
\code{\link[mbkmeans]{mbkmeans}} from the \pkg{mbkmeans} package, which actually does all the heavy lifting.

\linkS4class{KmeansParam}, for dispatch to the standard k-means algorithm.
}
\author{
Stephanie Hicks
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nestedClusters.R
\name{nestedClusters}
\alias{nestedClusters}
\title{Map nested clusterings}
\usage{
nestedClusters(ref, alt)
}
\arguments{
\item{ref}{A character vector or factor containing one set of groupings, considered to be the reference.}

\item{alt}{A character vector or factor containing another set of groupings, to be compared to \code{alt}.}
}
\value{
A list containing:
\itemize{
\item \code{proportions}, a matrix where each row corresponds to one of the \code{alt} clusters and each column corresponds to one of the \code{ref} clusters.
Each matrix entry represents the proportion of cells in \code{alt} that are assigned to each cluster in \code{ref}.
(That is, the proportions across all \code{ref} clusters should sum to unity for each \code{alt} cluster.)
\item \code{alt.mapping}, a \linkS4class{DataFrame} with one row per cluster in \code{alt}.
This contains the columns \code{max}, a numeric vector specifying the maximum value of \code{statistic} for that \code{alt} cluster;
and \code{which}, a character vector specifying the \code{ref} cluster in which the maximum value occurs.
\item \code{ref.score}, a numeric vector of length equal to the number of \code{ref} clusters.
This represents the degree of nesting of \code{alt} clusters within each \code{ref} cluster, see Details.
}
}
\description{
Map an alternative clustering to a reference clustering,
where the latter is expected to be nested within the former.
}
\details{
This function identifies mappings between two clusterings on the same set of cells where \code{alt} is potentially nested within \code{ref} (e.g., as it is computed at higher resolution).
To do so, we take each \code{alt} cluster and compute the the proportion of its cells that are derived from each \code{ref} cluster.
The corresponding \code{ref} cluster is identified as that with the highest proportion, as reported by the \code{which} field in the \code{mapping} DataFrame. 

The quality of the mapping is determined by \code{max} in the output \code{mapping} DataFrame.
A low value indicates that \code{alt} does not have a clear counterpart in \code{ref}, representing loss of heterogeneity.
Note that this is not a symmetrical inference; multiple \code{alt} clusters can map to the same \code{ref} cluster without manifesting as a low \code{max}.
This implicitly assumes that an increase in resolution in \code{alt} is not problematic.

The \code{ref.score} value for each cluster \code{ref} is formally defined as the probability of randomly picking a cell that belongs to \code{ref},
conditional on the event that the chosen cell belongs to the same \code{alt} cluster as a randomly chosen cell from \code{ref}.
This probability is equal to unity when \code{ref} is an exact superset of all \code{alt} clusters that contain its cells, corresponding to perfect 1:many nesting.
In contrast, if the \code{alt} clusters contain a mix of cells from different \code{ref}, this probability will be low and can be used as a diagnostic for imperfect nesting.
}
\examples{
m <- matrix(runif(10000), ncol=10)
clust1 <- kmeans(m,10)$cluster
clust2 <- kmeans(m,20)$cluster
nestedClusters(clust1, clust2)

# The ref.score is 1 in cases of perfect nesting.
nestedClusters(clust1, clust1)$ref.score

nest.clust <- paste0(clust1, sample(letters, length(clust1), replace=TRUE))
nestedClusters(clust1, nest.clust)$ref.score

# In contrast, it is much lower when nesting is bad.
nestedClusters(clust1, sample(clust1))$ref.score

}
\seealso{
\code{\link{linkClusters}}, to do this in a symmetric manner (i.e., without nesting).

\code{\link{pairwiseRand}}, for another way of comparing two sets of clusterings.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SomParam.R
\docType{class}
\name{SomParam-class}
\alias{SomParam-class}
\alias{show,SomParam-method}
\alias{SomParam}
\alias{clusterRows,ANY,SomParam-method}
\title{Clustering with self-organizing maps}
\usage{
SomParam(
  centers,
  dim.ratio = 1,
  topo = "rectangular",
  neighbourhood.fct = "bubble",
  toroidal = FALSE,
  rlen = 100,
  alpha = c(0.05, 0.01),
  radius = NULL,
  dist.fct = "sumofsquares"
)

\S4method{clusterRows}{ANY,SomParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{centers}{An integer scalar specifying the number of centers.
Alternatively, a function that takes the number of observations and returns the number of centers.}

\item{dim.ratio}{A positive numeric scalar in specifying how \code{centers} should be distributed between the \code{x} and \code{y} dimensions.
Defaults to equal distribution, i.e., both dimensions will be of length equal to the square root of \code{centers}.
Values above 1 will distribute more nodes to \code{x} while values below 1 will distribute mode nodes to \code{y}.}

\item{topo, neighbourhood.fct, toroidal}{Further arguments to pass to the \code{\link[kohonen]{somgrid}} function in the \pkg{kohonen} package.}

\item{rlen, alpha, radius, dist.fct}{Further arguments to pass to the \code{\link[kohonen]{som}} function in the \pkg{kohonen} package.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{SomParam} object.}

\item{full}{Logical scalar indicating whether the full SOM statistics should be returned.}
}
\value{
The \code{SomParam} constructor will return a \linkS4class{SomParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
(a list containing \code{som}, the direct output of \code{\link[kohonen]{som}}).
}
\description{
Use the self-organizing map implementation in the \pkg{kohonen} package to cluster observations into the specified number of nodes.
Note that this requires the installation of the \pkg{kohonen} package.
}
\details{
This class usually requires the user to specify the number of clusters beforehand.
However, we can also allow the number of clusters to vary as a function of the number of observations.
The latter is occasionally useful, e.g., to allow the clustering to automatically become more granular for large datasets.

Note that the final number of clusters may not be exactly equal to \code{centers}, depending on how \code{dim.ratio} is specified.
For example, if \code{centers} is a perfect square and \code{dim.ratio=1}, we will get exactly the requested number of points.

To modify an existing SomParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.

For \code{radius}, a value of \code{NULL} means that the default argument in the \code{\link[kohonen]{som}} function signature is used.
This is are data-dependent and so cannot be specified during construction of the SomParam object.

For \code{dist.fct}, users can specify any string that can be used in the \code{dist.fcts} arguments in \code{\link[kohonen]{som}}.
In practice, the only real alternative is \code{"manhattan"}.
}
\examples{
clusterRows(iris[,1:4], SomParam(centers=16))
clusterRows(iris[,1:4], SomParam(centers=12, dim.ratio=3/4))

}
\seealso{
\code{\link[kohonen]{som}} from the \pkg{kohonen} package, which does all of the heavy lifting.

\linkS4class{FixedNumberParam}, the parent of the SomParam class.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/neighborPurity.R
\name{neighborPurity}
\alias{neighborPurity}
\title{Compute neighborhood purity}
\usage{
neighborPurity(
  x,
  clusters,
  k = 50,
  weighted = TRUE,
  BNPARAM = KmknnParam(),
  BPPARAM = SerialParam()
)
}
\arguments{
\item{x}{A numeric matrix-like object containing observations in rows and variables in columns.}

\item{clusters}{Vector of length equal to \code{ncol(x)}, specifying the cluster assigned to each observation.}

\item{k}{Integer scalar specifying the number of nearest neighbors to use to determine the radius of the hyperspheres.}

\item{weighted}{A logical scalar indicating whether to weight each observation in inverse proportion to the size of its cluster.
Alternatively, a numeric vector of length equal to \code{clusters} containing the weight to use for each observation.}

\item{BNPARAM}{A \linkS4class{BiocNeighborParam} object specifying the nearest neighbor algorithm.
This should be an algorithm supported by \code{\link{findNeighbors}}.}

\item{BPPARAM}{A \linkS4class{BiocParallelParam} object indicating whether and how parallelization should be performed across genes.}
}
\value{
A \linkS4class{DataFrame} with one row per observation in \code{x} and the columns:
\itemize{
\item \code{purity}, a numeric field containing the purity value for the current observation.
\item \code{maximum}, the cluster with the highest proportion of observations neighboring the current observation.
}
Row names are defined as the row names of \code{x}.
}
\description{
Use a hypersphere-based approach to compute the \dQuote{purity} of each cluster based on the number of contaminating observations from different clusters in its neighborhood.
}
\details{
The purity of a cluster is quantified by creating a hypersphere around each observation in the cluster
and computing the proportion of observations in that hypersphere from the same cluster.
If all observations in a cluster have proportions close to 1, this indicates that the cluster is highly pure,
i.e., there are few observations from other clusters in its region of the coordinate space.
The distribution of purities for each cluster can be used as a measure of separation from other clusters.

In most cases, the majority of observations of a cluster will have high purities, corresponding to observations close to the cluster center.
A fraction of observations will have low values as these lie at the boundaries of two adjacent clusters.
A high degree of over-clustering will manifest as a majority of observations with purities close to zero.
The \code{maximum} field in the output can be used to determine the identity of the cluster 
with the greatest presence in a observation's neighborhood, usually an adjacent cluster for observations lying on the boundary.

The choice of \code{k} is used only to determine an appropriate value for the hypersphere radius.
We use hyperspheres as this is robust to changes in density throughout the coordinate space,
in contrast to computing purity based on the proportion of k-nearest neighbors in the same cluster.
For example, the latter will fail most obviously when the size of the cluster is less than \code{k}.
}
\section{Weighting by frequency}{

By default, purity values are computed after weighting each observation by the reciprocal of the number of observations in the same cluster.
Otherwise, clusters with more observations will have higher purities as any contamination is offset by the bulk of observations,
which would compromise comparisons of purities between clusters.
One can interpret the weighted purities as the expected value after downsampling all clusters to the same size.

Advanced users can achieve greater control by manually supplying a numeric vector of weights to \code{weighted}.
For example, we may wish to check the purity of batches after batch correction in single-cell RNA-seq.
In this application, \code{clusters} should be set to the \emph{batch blocking factor} (not the cluster identities!)
and \code{weighted} should be set to 1 over the frequency of each combination of cell type and batch.
This accounts for differences in cell type composition between batches when computing purities.

If \code{weighted=FALSE}, no weighting is performed.
}

\examples{
m <- matrix(runif(1000), ncol=10)
clusters <- clusterRows(m, BLUSPARAM=NNGraphParam())
out <- neighborPurity(m, clusters)
boxplot(split(out$purity, clusters))

# Mocking up a stronger example:
centers <- matrix(rnorm(30), nrow=3)
clusters <- sample(1:3, 1000, replace=TRUE)
y <- centers[clusters,,drop=FALSE]
y <- y + rnorm(length(y))

out2 <- neighborPurity(y, clusters)
boxplot(split(out2$purity, clusters))

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compareClusterings.R
\name{compareClusterings}
\alias{compareClusterings}
\title{Compare pairs of clusterings}
\usage{
compareClusterings(clusters, adjusted = TRUE)
}
\arguments{
\item{clusters}{A list of factors or vectors where each entry corresponds to a clustering.
All vectors should be of the same length.
The list itself should usually be named with a suitable label for each clustering.}

\item{adjusted}{Logical scalar indicating whether the adjusted Rand index should be returned.}
}
\value{
A symmetric square matrix of pairwise (adjusted) Rand indices between all pairs of clusters.

Aaron Lun
}
\description{
Compute the adjusted Rand index between all pairs of clusterings,
where larger values indicate a greater similarity between clusterings.
}
\details{
The aim of this function is to allow us to easily determine the relationships between clusterings.
For example, we might use this to determine which parameter settings have the greatest effect in a sweep by \code{\link{clusterSweep}}.
Alternatively, we could use this to obtain an \dQuote{ordering} of clusterings for visualization, e.g., with \pkg{clustree}.

This function does not provide any insight into the relationships between individual clusters.
A large Rand index only means that two clusterings are similar but does not specify the corresponding set of clusters across clusterings.
For that task, we suggest using the \code{\link{linkClusters}} function instead.
}
\examples{
clusters <- list(
    nngraph = clusterRows(iris[,1:4], NNGraphParam()),
    hclust = clusterRows(iris[,1:4], HclustParam(cut.dynamic=TRUE)),
    kmeans = clusterRows(iris[,1:4], KmeansParam(20))
)

aris <- compareClusterings(clusters)

# Visualizing the relationships between clusterings.
# Here, k-means is forced to be least similar to the others.
ari.as.graph <- igraph::graph.adjacency(aris, mode="undirected", weighted=TRUE)
plot(ari.as.graph)

# Obtain an ordering of clusterings, using the eigenvector 
# as a 1-dimensional summary of the matrix:
ev1 <- eigen(aris)$vectors[,1]
o <- order(ev1)
rownames(aris)[o]

}
\seealso{
\code{\link{linkClusters}}, which identifies relationships between individual clusters across clusterings.

\code{\link{pairwiseRand}}, for calculation of the pairwise Rand index.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/approxSilhouette.R
\name{approxSilhouette}
\alias{approxSilhouette}
\title{Approximate silhouette width}
\usage{
approxSilhouette(x, clusters)
}
\arguments{
\item{x}{A numeric matrix-like object containing observations in rows and variables in columns.}

\item{clusters}{Vector of length equal to \code{ncol(x)}, specifying the cluster assigned to each observation.}
}
\value{
A \linkS4class{DataFrame} with one row per observation in \code{x} and the columns:
\itemize{
\item \code{cluster}, the assigned cluster for each observation in \code{x}.
\item \code{other}, the closest cluster other than the one to which the current observation is assigned.
\item \code{width}, a numeric field containing the approximate silhouette width of the current cell.
}
Row names are defined as the row names of \code{x}.
}
\description{
Given a clustering, quickly compute an approximate silhouette width for each observation.
}
\details{
The silhouette width is a general-purpose method for evaluating the separation between clusters 
but requires calculating the average distance between pairs of observations within or between clusters.
This function instead approximates the average distance with the root-mean-squared-distance, which can be computed very efficiently for large datasets.
The approximated averages are then used to compute the silhouette width using the usual definition.

% In case you're wondering, here it is. Let O be the current observation.
% - The squared distance to the centroid is (O - Y)^2 where Y = mean(X) for the vector of observations X.
% - The mean of squares within X is mean[(X - Y)^2].
% - Their sum is:
%   (O - Y)^2 + mean[(X - Y)^2] = O^2 - 2OY + Y^2 + mean(X^2) - 2 mean(X) Y + Y^2
%                               = O^2 - 2OY + mean(X^2)
%                               = O^2 - 2OE(X) + mean(X^2)
%                               = mean[(O - X)^2]
%   After which we can just take the square root.
}
\examples{
m <- matrix(rnorm(10000), ncol=10)
clusters <- clusterRows(m, BLUSPARAM=KmeansParam(5))
out <- approxSilhouette(m, clusters)
boxplot(split(out$width, clusters))

# Mocking up a stronger example:
centers <- matrix(rnorm(30), nrow=3)
clusters <- sample(1:3, 1000, replace=TRUE)

y <- centers[clusters,]
y <- y + rnorm(length(y), sd=0.1)

out2 <- approxSilhouette(y, clusters)
boxplot(split(out2$width, clusters))

}
\seealso{
\code{silhouette} from the \pkg{cluster} package, for the exact calculation.

\code{\link{neighborPurity}}, for another method of evaluating cluster separation.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DbscanParam.R
\docType{class}
\name{DbscanParam-class}
\alias{DbscanParam-class}
\alias{show,DbscanParam-method}
\alias{DbscanParam}
\alias{clusterRows,ANY,DbscanParam-method}
\title{Density-based clustering with DBSCAN}
\usage{
DbscanParam(
  eps = NULL,
  min.pts = 5,
  core.prop = 0.5,
  chunk.size = 1000,
  BNPARAM = KmknnParam(),
  BPPARAM = SerialParam()
)

\S4method{clusterRows}{ANY,DbscanParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{eps}{Numeric scalar specifying the distance to use to define neighborhoods.
If \code{NULL}, this is determined from \code{min.pts} and \code{core.prop}.}

\item{min.pts}{Integer scalar specifying the minimum number of neighboring observations required for an observation to be a core point.}

\item{core.prop}{Numeric scalar specifying the proportion of observations to treat as core points.
This is only used when \code{eps=NULL}, see Details.}

\item{chunk.size}{Integer scalar specifying the number of points to process per chunk.}

\item{BNPARAM}{A \linkS4class{BiocNeighborParam} object specifying the algorithm to use for the neighbor searches.
This should be able to support both nearest-neighbor and range queries.}

\item{BPPARAM}{A \linkS4class{BiocParallelParam} object specifying the parallelization to use for the neighbor searches.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{BlusterParam} object specifying the algorithm to use.}

\item{full}{Logical scalar indicating whether additional statistics should be returned.}
}
\value{
The \code{DbscanParam} constructor will return a \linkS4class{DbscanParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
Note that this may contain \code{NA} values corresponding to noise points.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects} 
(a list containing the \code{eps} and \code{min.pts} used in the analysis).
}
\description{
Perform density-based clustering with a fast re-implementation of the DBSCAN algorithm.
}
\details{
DBSCAN operates by identifying core points, i.e., observations with at least \code{min.pts} neighbors within a distance of \code{eps}.
It then identifies which core points are neighbors of each other, forming components of connected core points.
All non-core points are then connected to the closest core point within \code{eps}.
All groups of points that are connected in this manner are considered to be part of the same cluster.
Any unconnected non-core points are treated as noise and reported as \code{NA}.

As a suitable value of \code{eps} may not be known beforehand, we can automatically determine it from the data.
For all observations, we compute the distance to the \eqn{k}th neighbor where \eqn{k} is defined as \code{round(min.pts * core.prop)}.
We then define \code{eps} as the \code{core.prop} quantile of the distances across all observations.
The default of \code{core.prop=0.5} means that around half of the observations will be treated as core points.

Larger values of \code{eps} will generally result in fewer observations classified as noise, as they are more likely to connect to a core point.
It may also promote agglomeration of existing clusters into larger entities if they are connected by regions of (relatively) low density.
Conversely, larger values of \code{min.pts} will generally increase the number of noise points and may fragment larger clusters into subclusters.

To modify an existing DbscanParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.
}
\examples{
clusterRows(iris[,1:4], DbscanParam())
clusterRows(iris[,1:4], DbscanParam(core.prop=0.8))

}
\references{
Ester M et al. (1996).
A density-based algorithm for discovering clusters in large spatial databases with noise.
\emph{Proceedings of the Second International Conference on Knowledge Discovery and Data Mining}, 226-231.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DianaParam.R
\docType{class}
\name{DianaParam-class}
\alias{DianaParam-class}
\alias{show,DianaParam-method}
\alias{.defaultScalarArguments,DianaParam-method}
\alias{DianaParam}
\alias{clusterRows,ANY,DianaParam-method}
\title{Divisive analysis clustering}
\usage{
DianaParam(
  metric = NULL,
  dist.fun = NULL,
  stand = NULL,
  cut.fun = NULL,
  cut.dynamic = FALSE,
  cut.params = list()
)

\S4method{clusterRows}{ANY,DianaParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{metric}{String specifying the distance metric to use in \code{\link{diana}}.
If \code{NULL}, the default metric is used.
If \code{dist.fun} is supplied, \code{metric} is passed to that function instead.}

\item{dist.fun}{Function specifying the function to use to compute the distance matrix. 
The function should accept a data matrix and a \code{method=} string (used to accept \code{metric}) and return a dissimilarity matrix of type \link{dist}.
If \code{NULL}, the \code{stats::\link{dist}} function is used by default.}

\item{stand}{Further arguments to pass to \code{\link{diana}}.}

\item{cut.fun}{Function specifying the method to use to cut the dendrogram.
The first argument of this function should be the output of \code{\link{hclust}},
and the return value should be an atomic vector specifying the cluster assignment for each observation.
Defaults to \code{\link{cutree}} if \code{cut.dynamic=FALSE} and \code{\link[dynamicTreeCut]{cutreeDynamic}} otherwise.}

\item{cut.dynamic}{Logical scalar indicating whether a dynamic tree cut should be performed using the \pkg{dynamicTreeCut} package.}

\item{cut.params}{Further arguments to pass to \code{cut.fun}.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{HclustParam} object.}

\item{full}{Logical scalar indicating whether the hierarchical clustering statistics should be returned.}
}
\value{
The \code{DianaParam} constructor will return a \linkS4class{DianaParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
(a list containing \code{diana}, the function output; \code{dist}, the dissimilarity matrix; and \code{hclust}, a \link{hclust} object created from \code{diana}).
}
\description{
Use the \code{\link{diana}} function to perform divisive analysis clustering.
}
\details{
To modify an existing DianaParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.

If \code{cut.fun=NULL}, \code{cut.dynamic=FALSE} and \code{cut.params} does not have \code{h} or \code{k},
\code{\link{clusterRows}} will automatically set \code{h} to half the tree height when calling \code{\link{cutree}}.
}
\examples{
clusterRows(iris[,1:4], DianaParam())
clusterRows(iris[,1:4], DianaParam(metric="manhattan"))

}
\seealso{
\code{\link{diana}}, which actually does all the heavy lifting.

\linkS4class{HclustParam}, for the more commonly used implementation of hierarchical clustering.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusterSweep.R
\name{clusterSweep}
\alias{clusterSweep}
\title{Clustering parameter sweeps}
\usage{
clusterSweep(
  x,
  BLUSPARAM,
  ...,
  full = FALSE,
  BPPARAM = SerialParam(),
  args = list()
)
}
\arguments{
\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{BlusterParam} object specifying the algorithm to use.}

\item{...}{Named vectors or lists specifying the parameters to sweep over.}

\item{full}{Logical scalar indicating whether the full clustering statistics should be returned for each method.}

\item{BPPARAM}{A \linkS4class{BiocParallelParam} specifying how the sweep should be parallelized.}

\item{args}{A named list of additional arguments to use with \code{...}.
This is provided in case there is a name conflict with the existing arguments in this function signature.}
}
\value{
A \linkS4class{List} containing:
\itemize{
\item \code{clusters}, a \linkS4class{DataFrame} with number of rows equal to that of \code{x},
where each column corresponds to (and is named after) a specific combination of clustering parameters.
\item \code{parameters}, another DataFrame with number of rows equal to the number of columns in the previous \code{clusters} DataFrame.
Each row contains the specific parameter combination for each column of \code{clusters}.
\item If \code{full=TRUE}, \code{objects} is an additional list of length equal to the number of rows in \code{clusters}.
This contains the \code{objects} produced by each run.
}
}
\description{
Perform a sweep across combinations of parameters to obtain different clusterings from the same algorithm.
}
\details{
This function allows users to conveniently test out a range of clustering parameters in a single call.
The name of each argument in \code{...} should be a legitimate argument to \code{x[[i]]},
and will be used to modify any existing values in \code{BLUSPARAM} to obtain a new set of parameters.
(For all other parameters, the existing values in \code{BLUSPARAM} are used.)
If multiple arguments are provided, all combinations are tested.

We attempt to create a unique name for each column based on its parameter combination.
This has the format of \code{<NAME1>.<VALUE1>_<NAME2>.<VALUE2>_...} based on the parameter names and values.
Note that any non-atomic values are simply represented by the name of their class;
no attempt is made to convert these into a compact string.

If an entry of \code{...} is a \emph{named} list of vectors, we expand those to generate all possible combinations of values.
For example, if we passed:
\preformatted{    blah.args = list(a = 1:5, b = LETTERS[1:3])}
This would be equivalent to manually specifying:
\preformatted{    blah.args = list(list(a = 1, b = "A"), list(a = 1, b = "B"), ...)}
The auto-expansion mechanism allows us to conveniently test parameter combinations when those parameters are stored inside \code{x} as a list.
The algorithm is recursive so any internal named lists containing vectors are similarly expanded.
Expansion can be disabled by wrapping vectors in \code{\link{I}}, in which case they are passed verbatim.
No expansion is performed for non-vector arguments.
}
\examples{
out <- clusterSweep(iris[,1:4], KmeansParam(10), 
    centers=4:10, algorithm=c("Lloyd", "Hartigan-Wong"))
out$clusters[,1:5]
out$parameters

out <- clusterSweep(iris[,1:4], NNGraphParam(), k=c(5L, 10L, 15L, 20L),
    cluster.fun=c("louvain", "walktrap"))
out$clusters[,1:5]
out$parameters

# Combinations are automatically expanded inside named lists:
out <- clusterSweep(iris[,1:4], NNGraphParam(), k=c(5L, 10L, 15L, 20L), 
    cluster.args=list(steps=3:4))
out$clusters[,1:5]
out$parameters

}
\seealso{
\code{\link{clusterRows}}, which manages the dispatch to specific methods based on \code{BLUSPARAM}.

\linkS4class{BlusterParam}, which determines which algorithm is actually used.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AffinityParam.R
\docType{class}
\name{AffinityParam-class}
\alias{AffinityParam-class}
\alias{show,AffinityParam-method}
\alias{AffinityParam}
\alias{clusterRows,ANY,AffinityParam-method}
\title{Affinity propogation}
\usage{
AffinityParam(
  s = NULL,
  p = NA,
  q = NA,
  maxits = 1000,
  convits = 100,
  lam = 0.9,
  nonoise = FALSE
)

\S4method{clusterRows}{ANY,AffinityParam}(x, BLUSPARAM, full = FALSE)
}
\arguments{
\item{s}{A function that accepts a matrix of observations by dimensions and returns a similarity matrix.
If \code{NULL}, defaults to the output of \code{\link[apcluster]{negDistMat}} with \code{r=2}.}

\item{p, q}{Numeric scalars controlling the input preference, i.e., the resolution of the clustering.
These are passed to the \code{\link[apcluster]{apcluster}} function, where values of \code{NA} are the default.}

\item{maxits, convits, lam, nonoise}{Further arguments to pass to the \code{\link[apcluster]{apcluster}} function.}

\item{x}{A numeric matrix-like object where rows represent observations and columns represent variables.}

\item{BLUSPARAM}{A \linkS4class{AffinityParam} object.}

\item{full}{Logical scalar indicating whether the full affinity propagation statistics should be returned.}
}
\value{
The \code{AffinityParam} constructor will return a \linkS4class{AffinityParam} object with the specified parameters.

The \code{clusterRows} method will return a factor of length equal to \code{nrow(x)} containing the cluster assignments.
If \code{full=TRUE}, a list is returned with \code{clusters} (the factor, as above) and \code{objects}
(a list containing \code{similarity}, the similarity matrix; and \code{apcluster}, the direct output of \code{\link[apcluster]{apcluster}}).
}
\description{
Use affinity propagation from the \pkg{apcluster} package to cluster observations.
Note that this requires the installation of the \pkg{apcluster} package.
}
\details{
To modify an existing AffinityParam object \code{x},
users can simply call \code{x[[i]]} or \code{x[[i]] <- value} where \code{i} is any argument used in the constructor.

Setting \code{q} (and less typically, \code{p}) allows us to tune the resolution of the clustering.
In particular, when \code{p=NA}, it is computed based on the setting of \code{q}:
\itemize{
\item If the specified \code{q} lies in [0, 1], \code{p} is defined as the \code{q}-quantile of the finite similarities across all pairs of observations.
When \code{q=NA}, it defaults to 0.5.
\item If \code{q} is negative, \code{p} is defined as the \code{M + abs(M) * q} where \code{M} is the smallest finite similarity across all pairs.
This yields smaller \code{p} values while still responding to the scale of the similarities. 
}
The resulting value is used as the self-preference, i.e., the diagonal of the availability matrix.
Larger values yield more clusters as each data point is more inclined to form its own cluster.
}
\examples{
clusterRows(iris[,1:4], AffinityParam())
clusterRows(iris[,1:4], AffinityParam(q=0.9))
clusterRows(iris[,1:4], AffinityParam(s=apcluster::expSimMat()))

}
\seealso{
\code{\link[apcluster]{apcluster}} from the \pkg{apcluster} package, which does all of the heavy lifting.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusterRMSD.R
\name{clusterRMSD}
\alias{clusterRMSD}
\title{Compute the RMSD per cluster}
\usage{
clusterRMSD(x, clusters, sum = FALSE)
}
\arguments{
\item{x}{Numeric matrix containing observations in rows and variables in columns.}

\item{clusters}{Vector containing the assigned cluster for each observation.}

\item{sum}{Logical scalar indicating whether to compute the sum of squares.}
}
\value{
Numeric vector of RMSD values per cluster.
If \code{sum=TRUE}, a numeric vector of the sum of squares per cluster is returned instead.
}
\description{
Compute the root mean-squared deviation (RMSD) for each cluster.
}
\details{
The RMSD for each cluster is a measure of its dispersion;
clusters with large internal heterogeneity will have high RMSDs and are good candidates for further subclustering.
}
\examples{
x <- matrix(rnorm(10000), ncol=10)
kout <- kmeans(x, 5)
clusterRMSD(x, kout$cluster)

}
\author{
Aaron Lun
}
